/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "math.h"
#include "string.h"
#include "fix_insert.h"
#include "atom.h"
#include "force.h"
#include "update.h"
#include "modify.h"
#include "fix_gravity.h"
#include "fix_shear_history.h"
#include "neighbor.h"
#include "domain.h"
#include "region.h"
#include "region_block.h"
#include "region_cylinder.h"
#include "random_park.h"
#include "memory.h"
#include "error.h"

#define EPSILON 0.001

/* ---------------------------------------------------------------------- */

FixInsert::FixInsert(int narg, char **arg) : Fix(narg, arg)
{
  if (narg < 6) error->all("Illegal fix insert command");

  if (atom->check_style("granular") == 0)
    error->all("Must use fix insert with atom style granular");

  // required args

  ninsert = atoi(arg[3]);
  ntype = atoi(arg[4]);
  seed = atoi(arg[5]);

  PI = 4.0*atan(1.0);

  // option defaults

  int iregion = -1;
  radius_lo = radius_hi = 0.5;
  density_lo = density_hi = 1.0;
  volfrac = 0.25;
  maxattempt = 50;
  zrate = 0.0;
  vxlo = vxhi = vylo = vyhi = vz = 0.0;

  // optional args

  int iarg = 6;
  while (iarg < narg) {
    if (strcmp(arg[iarg],"region") == 0) {
      if (iarg+2 > narg) error->all("Incorrect args in fix insert options");
      for (iregion = 0; iregion < domain->nregion; iregion++)
	if (strcmp(arg[iarg+1],domain->regions[iregion]->id) == 0) break;
      if (iregion == domain->nregion) 
	error->all("Fix insert region ID does not exist");
      iarg += 2;
    } else if (strcmp(arg[iarg],"diam") == 0) {
      if (iarg+3 > narg) error->all("Incorrect args in fix insert options");
      radius_lo = 0.5 * atof(arg[iarg+1]);
      radius_hi = 0.5 * atof(arg[iarg+2]);
      iarg += 3;
    } else if (strcmp(arg[iarg],"dens") == 0) {
      if (iarg+3 > narg) error->all("Incorrect args in fix insert options");
      density_lo = atof(arg[iarg+1]);
      density_hi = atof(arg[iarg+2]);
      iarg += 3;
    } else if (strcmp(arg[iarg],"vol") == 0) {
      if (iarg+3 > narg) error->all("Incorrect args in fix insert options");
      volfrac = atof(arg[iarg+1]);
      maxattempt = atoi(arg[iarg+2]);
      iarg += 3;
    } else if (strcmp(arg[iarg],"zrate") == 0) {
      if (iarg+2 > narg) error->all("Incorrect args in fix insert options");
      zrate = atof(arg[iarg+1]);
      iarg += 2;
    } else if (strcmp(arg[iarg],"vel") == 0) {
      if (iarg+6 > narg) error->all("Incorrect args in fix insert options");
      vxlo = atof(arg[iarg+1]);
      vxhi = atof(arg[iarg+2]);
      vylo = atof(arg[iarg+3]);
      vyhi = atof(arg[iarg+4]);
      vz = atof(arg[iarg+5]);
      iarg += 6;
    } else error->all("Incorrect args in fix insert options");
  }

  // error check that a valid region was specified

  if (iregion == -1) error->all("Must specify a region in fix insert");

  // error checks on region

  if (domain->regions[iregion]->interior == 0)
    error->all("Must use region with side = in with fix insert");

  if (strcmp(domain->regions[iregion]->style,"block") == 0) {
    region_style = 1;
    xlo = ((RegBlock *) domain->regions[iregion])->xlo;
    xhi = ((RegBlock *) domain->regions[iregion])->xhi;
    ylo = ((RegBlock *) domain->regions[iregion])->ylo;
    yhi = ((RegBlock *) domain->regions[iregion])->yhi;
    zlo = ((RegBlock *) domain->regions[iregion])->zlo;
    zhi = ((RegBlock *) domain->regions[iregion])->zhi;
    if (xlo < domain->boxxlo || xhi > domain->boxxhi || 
	ylo < domain->boxylo || yhi > domain->boxyhi || 
	zlo < domain->boxzlo || zhi > domain->boxzhi)
      error->all("Insertion region extends outside simulation box");
  } else if (strcmp(domain->regions[iregion]->style,"cylinder") == 0) {
    region_style = 2;
    char axis = ((RegCylinder *) domain->regions[iregion])->axis;
    xc = ((RegCylinder *) domain->regions[iregion])->c1;
    yc = ((RegCylinder *) domain->regions[iregion])->c2;
    rc = ((RegCylinder *) domain->regions[iregion])->radius;
    zlo = ((RegCylinder *) domain->regions[iregion])->lo;
    zhi = ((RegCylinder *) domain->regions[iregion])->hi;
    if (axis != 'z')
      error->all("Must use a z-axis cylinder with fix insert");
    if (xc-rc < domain->boxxlo || xc+rc > domain->boxxhi || 
	yc-rc < domain->boxylo || yc+rc > domain->boxyhi || 
	zlo < domain->boxzlo || zhi > domain->boxzhi)
      error->all("Insertion region extends outside simulation box");
  } else error->all("Must use a block or cylinder region with fix insert");

  // random number generator, same for all procs

  random = new RanPark(seed);

  // allgather arrays

  MPI_Comm_rank(world,&me);
  MPI_Comm_size(world,&nprocs);
  recvcounts = new int[nprocs];
  displs = new int[nprocs];

  // nevery = timesteps between insertions
  // should be time for a particle to fall from top of insertion region
  //   to bottom, taking into account that the region may be moving
  // 1st insertion on next timestep

  double g = 1.0;
  double vz_relative = vz - zrate;
  double delta = vz_relative + 
    sqrt(vz_relative*vz_relative + 2.0*g*(zhi-zlo)) / g;
  nevery = static_cast<int> (delta/update->dt + 0.5);

  next_reneighbor = update->ntimestep + 1;
  nfirst = next_reneighbor;
  ninserted = 0;

  // nper = # to insert each time
  // depends on specified volume fraction
  // volume = volume of insertion region
  // volume_one = volume of inserted particle (with max possible radius)
  // dy = min dimension of block region in y, for quasi-2d simulations

  double volume;
  if (region_style == 1) {
    double dy = yhi - ylo;
    if (dy < 1.0) dy = 1.0;
    volume = (xhi-xlo) * dy * (zhi-zlo);
  } else if (region_style == 2)
    volume = PI*rc*rc * (zhi-zlo);

  double volume_one = 4.0/3.0 * PI * radius_hi*radius_hi*radius_hi;
  nper = static_cast<int> (volfrac*volume/volume_one);
  int nfinal = update->ntimestep + 1 + (ninsert-1)/nper * nevery;

  // print stats

  if (me == 0) {
    if (screen)
      fprintf(screen,
	      "Particle insertion: %d every %d steps, %d by step %d\n",
	      nper,nevery,ninsert,nfinal);
    if (logfile)
      fprintf(logfile,
	      "Particle insertion: %d every %d steps, %d by step %d\n",
	      nper,nevery,ninsert,nfinal);
  }
}

/* ---------------------------------------------------------------------- */

FixInsert::~FixInsert()
{
  delete random;
  delete [] recvcounts;
  delete [] displs;
}

/* ---------------------------------------------------------------------- */

int FixInsert::setmask()
{
  int mask = 0;
  mask |= PRE_EXCHANGE;
  return mask;
}

/* ---------------------------------------------------------------------- */

void FixInsert::init()
{
  // insure gravity fix exists and points in -z
  // else insertion cannot work

  int ifix;
  for (ifix = 0; ifix < modify->nfix; ifix++)
    if (strcmp(modify->fix[ifix]->style,"gravity") == 0) break;
  if (ifix == modify->nfix) 
    error->all("Must use fix gravity with fix insert");
  if (fabs(((FixGravity *) modify->fix[ifix])->theta+180.0) > EPSILON)
    error->all("Gravity must point in -z to use with fix insert");

  // check if a shear history fix exists

  ifix_history = -1;
  if (force->pair_match("gran/history") || force->pair_match("gran/hertzian"))
    for (int i = 0; i < modify->nfix; i++)
      if (strcmp(modify->fix[i]->style,"SHEAR_HISTORY") == 0) ifix_history = i;
}

/* ----------------------------------------------------------------------
   perform particle insertion
------------------------------------------------------------------------- */

void FixInsert::pre_exchange()
{
  int i;

  // just return if should not be called on this timestep

  if (next_reneighbor != update->ntimestep) return;

  // nnew = # to insert this timestep

  int nnew = nper;
  if (ninserted + nnew > ninsert) nnew = ninsert - ninserted;

  // zlo/hi current = z bounds of insertion region this timestep

  zlo_current = zlo + (update->ntimestep - nfirst) * update->dt * zrate;
  zhi_current = zhi + (update->ntimestep - nfirst) * update->dt * zrate;
  
  // ncount = # of my atoms that overlap the insertion region
  // nprevious = total of ncount across all procs
  
  int ncount = 0;
  for (i = 0; i < atom->nlocal; i++)
    if (overlap(i)) ncount++;

  int nprevious;
  MPI_Allreduce(&ncount,&nprevious,1,MPI_INT,MPI_SUM,world);

  // xmine is for my atoms
  // xnear is for atoms from all procs + atoms to be inserted

  double **xmine = 
    memory->create_2d_double_array(ncount,4,"fix_insert:xmine");
  double **xnear = 
    memory->create_2d_double_array(nprevious+nnew,4,"fix_insert:xnear");
  int nnear = nprevious;

  // setup for allgatherv

  int n = 4*ncount;
  MPI_Allgather(&n,1,MPI_INT,recvcounts,1,MPI_INT,world);

  displs[0] = 0;
  for (int iproc = 1; iproc < nprocs; iproc++)
    displs[iproc] = displs[iproc-1] + recvcounts[iproc-1];

  // load up xmine array
  
  double **x = atom->x;
  double *radius = atom->radius;

  ncount = 0;
  for (i = 0; i < atom->nlocal; i++)
    if (overlap(i)) {
      xmine[ncount][0] = x[i][0];
      xmine[ncount][1] = x[i][1];
      xmine[ncount][2] = x[i][2];
      xmine[ncount][3] = radius[i];
      ncount++;
    }

  // perform allgatherv to acquire list of nearby particles on all procs

  double *ptr = NULL;
  if (ncount) ptr = xmine[0];
  MPI_Allgatherv(ptr,4*ncount,MPI_DOUBLE,
		 xnear[0],recvcounts,displs,MPI_DOUBLE,world);

  // insert new atoms into xnear list, one by one
  // check against all nearby atoms and previously inserted ones
  // if there is an overlap then just try again at same z position
  // else insert by adding to xnear list
  // max = maximum # of insertion attempts for all particles
  // ztmp is biased to give uniform distribution in time of insertion

  int ok;
  double xtmp,ytmp,ztmp,radtmp,delx,dely,delz,rsq,radsum,rn;

  int attempt = 0;
  int max = nnew * maxattempt;
  int ntotal = nprevious+nnew;

  while (nnear < ntotal && attempt < max) {
    rn = random->uniform();
    ztmp = zhi_current - rn*rn * (zhi_current-zlo_current);
    radtmp = radius_lo + random->uniform() * (radius_hi-radius_lo);
    while (attempt < max) {
      attempt++;
      xy_random(xtmp,ytmp);
      ok = 1;
      for (i = 0; i < nnear; i++) {
	delx = xtmp - xnear[i][0];
	dely = ytmp - xnear[i][1];
	delz = ztmp - xnear[i][2];
	rsq = delx*delx + dely*dely + delz*delz;
	radsum = radtmp + xnear[i][3];
	if (rsq <= radsum*radsum) {
	  ok = 0;
	  break;
	}
      }
      if (ok) break;
    }
    xnear[nnear][0] = xtmp;
    xnear[nnear][1] = ytmp;
    xnear[nnear][2] = ztmp;
    xnear[nnear][3] = radtmp;
    nnear++;
  }

  // warn if not all insertions were performed

  ninserted += nnear-nprevious;
  if (nnear - nprevious < nnew && me == 0)
    error->warning("Less insertions than requested");

  // add new atoms in my sub-box to my arrays
  // initialize info about the atoms
  // type, diameter, density set from fix parameters
  // group mask set to "all" plus fix group
  // z velocity set to what velocity would be if particle
  //   had fallen from top of insertion region
  // this gives continuous stream of atoms
  // set npartner for new atoms to 0 (assume not touching any others)

  int m;
  double denstmp,vxtmp,vytmp;
  double g = 1.0;

  for (i = nprevious; i < nnear; i++) {
    xtmp = xnear[i][0];
    ytmp = xnear[i][1];
    ztmp = xnear[i][2];
    radtmp = xnear[i][3];
    denstmp = density_lo + random->uniform() * (density_hi-density_lo);
    vxtmp = vxlo + random->uniform() * (vxhi-vxlo);
    vytmp = vylo + random->uniform() * (vyhi-vylo);

    if (xtmp >= domain->subxlo && xtmp < domain->subxhi &&
	ytmp >= domain->subylo && ytmp < domain->subyhi &&
	ztmp >= domain->subzlo && ztmp < domain->subzhi) {
      atom->create_one(ntype,xtmp,ytmp,ztmp);
      m = atom->nlocal - 1;
      atom->type[m] = ntype;
      atom->radius[m] = radtmp;
      atom->density[m] = denstmp;
      atom->rmass[m] = 4.0*PI/3.0 * radtmp*radtmp*radtmp * denstmp;
      atom->mask[m] = 1 | groupbit;
      atom->v[m][0] = vxtmp;
      atom->v[m][1] = vytmp;
      atom->v[m][2] = vz - sqrt(2.0*g*(zhi_current-ztmp));
      if (ifix_history >= 0)
	((FixShearHistory *) modify->fix[ifix_history])->npartner[m] = 0;
    }
  }

  // tag # of new particles grow beyond all previous atoms

  atom->tag_extend();

  // free local memory

  memory->destroy_2d_double_array(xmine);
  memory->destroy_2d_double_array(xnear);

  // next timestep to insert

  if (ninserted < ninsert) next_reneighbor += nevery;
  else next_reneighbor = 0;
}

/* ----------------------------------------------------------------------
   check if particle i could overlap with a particle inserted into region
   return 1 if yes, 0 if no
   use maximum diameter for inserted particle
------------------------------------------------------------------------- */

int FixInsert::overlap(int i)
{
  double delta = radius_hi + atom->radius[i];
  double **x = atom->x;

  if (region_style == 1) {
    if (x[i][2] < zlo_current-delta || x[i][2] > zhi_current+delta ||
	x[i][0] < xlo-delta || x[i][0] > xhi+delta ||
	x[i][1] < ylo-delta || x[i][1] > xhi+delta) return 0;
  } else {
    if (x[i][2] < zlo_current-delta || x[i][2] > zhi_current+delta) return 0;
    double delx = x[i][0] - xc;
    double dely = x[i][1] - yc;
    double rsq = delx*delx + dely*dely;
    double r = rc + delta;
    if (rsq > r*r) return 0;
  }

  return 1;
}

/* ---------------------------------------------------------------------- */

void FixInsert::xy_random(double &x, double &y)
{
  if (region_style == 1) {
    x = xlo + random->uniform() * (xhi-xlo);
    y = ylo + random->uniform() * (yhi-ylo);
  } else {
    double r1,r2;
    while (1) {
      r1 = random->uniform() - 0.5;
      r2 = random->uniform() - 0.5;
      if (r1*r1 + r2*r2 < 0.25) break;
    }
    x = xc + 2.0*r1*rc;
    y = yc + 2.0*r2*rc;
  }
}
