/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "mpi.h"
#include "string.h"
#include "stdio.h"
#include "thermo_gran.h"
#include "atom.h"
#include "force.h"
#include "modify.h"
#include "fix.h"
#include "domain.h"
#include "update.h"
#include "error.h"

// moment of inertia for a sphere

#define INERTIA 0.4

/* ---------------------------------------------------------------------- */

ThermoGran::ThermoGran(char *style_in) : Thermo(style_in)
{
  int n;
  char *str;

  // insure atom style granular

  if (atom->check_style("granular") == 0)
    error->all("Must use atom style granular with thermo style gran");

  MPI_Comm_rank(world,&me);

  str = "%8d %.15g %12.8g %12.8g %12.8g\n";
  n = strlen(str) + 1;
  format = new char[n];
  strcpy(format,str);

  str = "%8d %.15g %12.8g %12.8g %12.8g %12.8g\n";
  n = strlen(str) + 1;
  format_volume = new char[n];
  strcpy(format_volume,str);
}

/* ---------------------------------------------------------------------- */

ThermoGran::~ThermoGran()
{
  delete [] format;
  delete [] format_volume;
}

/* ---------------------------------------------------------------------- */

void ThermoGran::init_style()
{
  // check for freeze Fix and set freeze_group_bit

  int i;
  for (i = 0; i < modify->nfix; i++)
    if (strcmp(modify->fix[i]->style,"freeze") == 0) break;
  if (i < modify->nfix) freeze_group_bit = modify->fix[i]->groupbit;
  else freeze_group_bit = 0;
}

/* ---------------------------------------------------------------------- */

void ThermoGran::header()
{
  if (me == 0) {
    if (volflag == 0) {
      if (screen)
	fprintf(screen,"Step Atoms Trans-KE Rot-KE TransKE/Atom\n");
      if (logfile)
	fprintf(logfile,"Step Atoms Trans-KE Rot-KE TransKE/Atom\n");
    } else {
      if (screen)
	fprintf(screen,"Step Atoms Trans-KE Rot-KE TransKE/Atom Volume\n");
      if (logfile)
	fprintf(logfile,"Step Atoms Trans-KE Rot-KE TransKE/Atom Volume\n");
    }
  }
}

/* ---------------------------------------------------------------------- */

void ThermoGran::compute(int flag)
{
  double natoms = lost_check();

  // volume

  double volume;
  if (volflag) volume = domain->xprd * domain->yprd * domain->zprd;

  // kinetic and rotational energy for non-frozen atoms

  double *rmass = atom->rmass;
  double *radius = atom->radius;
  double **v = atom->v;
  double **phiv = atom->phiv;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  double nactive = 0;
  double ke = 0.0;
  double ephi = 0.0;

  for (int i = 0; i < nlocal; i++) {
    if (!(mask[i] & freeze_group_bit)) {
      nactive++;
      ke += (v[i][0]*v[i][0] + v[i][1]*v[i][1] + v[i][2]*v[i][2]) *
	rmass[i];
      ephi += (phiv[i][0]*phiv[i][0] + phiv[i][1]*phiv[i][1] + 
	       phiv[i][2]*phiv[i][2]) * radius[i]*radius[i]*rmass[i];
    }
  }

  ke *= 0.5 * force->mvv2e;
  ephi *= 0.5 * force->mvv2e * INERTIA;

  double nactive_tot;
  double ke_tot,ephi_tot;

  MPI_Allreduce(&nactive,&nactive_tot,1,MPI_DOUBLE,MPI_SUM,world);
  MPI_Allreduce(&ke,&ke_tot,1,MPI_DOUBLE,MPI_SUM,world);
  MPI_Allreduce(&ephi,&ephi_tot,1,MPI_DOUBLE,MPI_SUM,world);

  // normalization (turn off for now to agree with GranFlow)

  // if (normflag) {
  //  double n = nactive_tot;
  //  if (n == 0) n = 1;
  //  ke_tot /= n;
  //  ephi_tot /= n;
  // }

  double ke_per = 0.0;
  if (nactive_tot > 0) ke_per = ke_tot/nactive_tot;

  // output

  if (me == 0) {
    if (volflag == 0) {
      if (screen)
	fprintf(screen,format,update->ntimestep,
		natoms,ke_tot,ephi_tot,ke_per);
      if (logfile) {
	fprintf(logfile,format,update->ntimestep,
		natoms,ke_tot,ephi_tot,ke_per);
	if (flush_flag) fflush(logfile);
      }
    } else {
      if (screen)
	fprintf(screen,format_volume,update->ntimestep,natoms,
		ke_tot,ephi_tot,ke_per,volume);
      if (logfile) {
	fprintf(logfile,format_volume,update->ntimestep,natoms,
		ke_tot,ephi_tot,ke_per,volume);
	if (flush_flag) fflush(logfile);
      }
    }
  }
}
