/* -----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------ */

#include "math.h"
#include "string.h"
#include "bond_hybrid.h"
#include "atom.h"
#include "neighbor.h"
#include "domain.h"
#include "comm.h"
#include "force.h"
#include "memory.h"
#include "error.h"

/* ----------------------------------------------------------------------
   set all global defaults 
------------------------------------------------------------------------- */

BondHybrid::BondHybrid()
{
  styles = NULL;
  keywords = NULL;
}

/* ----------------------------------------------------------------------
   free all arrays
------------------------------------------------------------------------- */

BondHybrid::~BondHybrid()
{
  for (int i = 0; i < nstyles; i++) delete styles[i];
  delete [] styles;
  for (int i = 0; i < nstyles; i++) delete [] keywords[i];
  delete [] keywords;

  if (allocated) {
    memory->sfree(setflag);
    memory->sfree(map);
  }
}

/* ---------------------------------------------------------------------- */

void BondHybrid::compute(int eflag, int vflag)
{
  int i1,i2,n,type,factor;
  double delx,dely,delz,rsq,fforce,rfactor,eng;

  energy = 0.0;
  if (vflag) for (n = 0; n < 6; n++) virial[n] = 0.0;

  double **x = atom->x;
  double **f = atom->f;
  int **bondlist = neighbor->bondlist;
  int nbondlist = neighbor->nbondlist;
  int nlocal = atom->nlocal;
  int newton_bond = force->newton_bond;

  for (n = 0; n < nbondlist; n++) {
    
    i1 = bondlist[n][0];
    i2 = bondlist[n][1];
    type = bondlist[n][2];

    if (newton_bond) factor = 2;
    else {
      factor = 0;
      if (i1 < nlocal) factor++;
      if (i2 < nlocal) factor++;
    }
    rfactor = 0.5 * factor;

    delx = x[i1][0] - x[i2][0];
    dely = x[i1][1] - x[i2][1];
    delz = x[i1][2] - x[i2][2];
    domain->minimum_image(&delx,&dely,&delz);
    rsq = delx*delx + dely*dely + delz*delz;

    // call to low-level bond style's compute_one function

    styles[map[type]]->single(type,rsq,rfactor,eflag,fforce,eng);

    if (eflag) energy += eng;

    // apply force to each of 2 atoms

    if (newton_bond || i1 < nlocal) {
      f[i1][0] += delx*fforce;
      f[i1][1] += dely*fforce;
      f[i1][2] += delz*fforce;
    }

    if (newton_bond || i2 < nlocal) {
      f[i2][0] -= delx*fforce;
      f[i2][1] -= dely*fforce;
      f[i2][2] -= delz*fforce;
    }

    // virial contribution

    if (vflag) {
      virial[0] += rfactor*delx*delx*fforce;
      virial[1] += rfactor*dely*dely*fforce;
      virial[2] += rfactor*delz*delz*fforce;
      virial[3] += rfactor*delx*dely*fforce;
      virial[4] += rfactor*delx*delz*fforce;
      virial[5] += rfactor*dely*delz*fforce;
    }
  }
}

/* ---------------------------------------------------------------------- */

void BondHybrid::allocate()
{
  allocated = 1;
  int n = atom->nbondtypes;

  map = (int *) memory->smalloc((n+1)*sizeof(int),"bond:map");
  setflag = (int *) memory->smalloc((n+1)*sizeof(int),"bond:setflag");
  for (int i = 1; i < n; i++) setflag[i] = 0;
}

/* ----------------------------------------------------------------------
   create one bond style for each arg in list
------------------------------------------------------------------------- */

void BondHybrid::settings(int narg, char **arg)
{
  nstyles = narg;
  styles = new Bond*[nstyles];
  keywords = new char*[nstyles];

  for (int i = 0; i < nstyles; i++) {
    for (int j = 0; j < i; j++)
      if (strcmp(arg[i],arg[j]) == 0) 
	error->all("Bond style hybrid cannot use same bond style twice");
    if (strcmp(arg[i],"hybrid") == 0) 
      error->all("Bond style hybrid cannot have hybrid as an argument");
    styles[i] = force->new_bond(arg[i]);
    keywords[i] = new char[strlen(arg[i])+1];
    strcpy(keywords[i],arg[i]);
  }
}

/* ----------------------------------------------------------------------
   set coeffs for one type
---------------------------------------------------------------------- */

void BondHybrid::coeff(int narg, char **arg)
{
  if (!allocated) allocate();

  int ilo,ihi;
  force->bounds(arg[0],atom->nbondtypes,ilo,ihi);

  // 2nd arg = bond style name (harmonic, fene, etc)

  int j;
  for (j = 0; j < nstyles; j++)
    if (strcmp(arg[1],keywords[j]) == 0) break;
  if (j == nstyles) error->all("Bond coeff for hybrid has invalid style");

  // set low-level coefficients for each bondtype
  // replace 2nd arg with i, call coeff() with no 1st arg

  for (int i = ilo; i <= ihi; i++) {
    sprintf(arg[1],"%d",i);
    map[i] = j;
    styles[map[i]]->coeff(narg-1,&arg[1]);
    setflag[i] = 1;
  }
}

/* ----------------------------------------------------------------------
   return an equilbrium bond length 
------------------------------------------------------------------------- */

double BondHybrid::equilibrium_distance(int i)
{
  return styles[map[i]]->equilibrium_distance(i);
}

/* ----------------------------------------------------------------------
   proc 0 writes to restart file
------------------------------------------------------------------------- */

void BondHybrid::write_restart(FILE *fp)
{
  fwrite(&nstyles,sizeof(int),1,fp);

  int n;
  for (int i = 0; i < nstyles; i++) {
    n = strlen(keywords[i]) + 1;
    fwrite(&n,sizeof(int),1,fp);
    fwrite(keywords[i],sizeof(char),n,fp);
  }
}

/* ----------------------------------------------------------------------
   proc 0 reads from restart file, bcasts
------------------------------------------------------------------------- */

void BondHybrid::read_restart(FILE *fp)
{
  allocate();

  int me = comm->me;
  if (me == 0) fread(&nstyles,sizeof(int),1,fp);
  MPI_Bcast(&nstyles,1,MPI_INT,0,world);
  styles = new Bond*[nstyles];
  keywords = new char*[nstyles];
  
  int n;
  for (int i = 0; i < nstyles; i++) {
    if (me == 0) fread(&n,sizeof(int),1,fp);
    MPI_Bcast(&n,1,MPI_INT,0,world);
    keywords[i] = new char[n];
    if (me == 0) fread(keywords[i],sizeof(char),n,fp);
    MPI_Bcast(keywords[i],n,MPI_CHAR,0,world);
    styles[i] = force->new_bond(keywords[i]);
  }
}

/* ---------------------------------------------------------------------- */

void BondHybrid::single(int type, double rsq, double rfactor, int eflag,
			double &fforce, double &eng)
{
  styles[map[type]]->single(type,rsq,rfactor,eflag,fforce,eng);
}
