/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "string.h"
#include "delete_atoms.h"
#include "atom.h"
#include "comm.h"
#include "domain.h"
#include "group.h"
#include "region.h"
#include "error.h"

#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

#define MAXATOMS 0x7FFFFFFF

/* ---------------------------------------------------------------------- */

void DeleteAtoms::command(int narg, char **arg)
{
  if (domain->box_exist == 0) 
    error->all("Delete_atoms command before simulation box is defined");

  if (narg < 1) error->all("Illegal delete_atoms command");

  // store state before delete

  if (atom->natoms >= MAXATOMS)
    error->all("Too many atoms to use delete atoms command");

  int natoms_previous = static_cast<int> (atom->natoms);
  int tag_lowest = natoms_previous + 1;

  // delete group: delete all atoms in group, group still exists

  if (strcmp(arg[0],"group") == 0) {

    if (narg != 2) error->all("Illegal delete_atoms group command");

    int igroup = group->find(arg[1]);
    if (igroup == -1) error->all("Could not find delete_atoms group ID");

    int *mask = atom->mask;
    int *tag = atom->tag;
    int nlocal = atom->nlocal;
    int groupbit = group->bitmask[igroup];

    int i = 0;
    while (i < nlocal) {
      if (mask[i] & groupbit) {
	tag_lowest = MIN(tag_lowest,tag[i]);
	atom->copy(nlocal-1,i);
	nlocal--;
      } else i++;
    }
    atom->nlocal = nlocal;

    // delete region: delete an atom if it is in all regions

  } else if (strcmp(arg[0],"region") == 0) {

    if (narg < 2) error->all("Illegal delete_atoms region command");

    int nregion = narg-1;
    int *list = new int[nregion];

    int iregion;
    for (int iarg = 1; iarg < narg; iarg++) {
      for (iregion = 0; iregion < domain->nregion; iregion++)
	if (strcmp(arg[iarg],domain->regions[iregion]->id) == 0) break;
      if (iregion == domain->nregion)
	error->all("Delete_atoms region ID does not exist");
      list[iarg-1] = iregion;
    }

    double **x = atom->x;
    int *tag = atom->tag;
    int nlocal = atom->nlocal;

    int i = 0;
    while (i < nlocal) {
      for (iregion = 0; iregion < nregion; iregion++)
	if (!domain->regions[list[iregion]]->match(x[i][0],x[i][1],x[i][2]))
	  break;
      if (iregion == nregion) {
	tag_lowest = MIN(tag_lowest,tag[i]);
	atom->copy(nlocal-1,i);
	nlocal--;
      } else i++;
    }
    atom->nlocal = nlocal;

    delete [] list;

  // delete overlap: delete one of any 2 atoms within overlap

  } else if (strcmp(arg[0],"overlap") == 0) {

    error->all("Delete_atoms overlap not yet implemented");

  } else error->all("Illegal delete_atoms command");

  // if non-molecular system, reset atom tags to be contiguous
  // tag_lowest_all = lowest atom ID of a deleted atom
  // set all atom IDs >= tag_lowest_all to 0
  // tag_extend to reset tags for all those atoms

  if (atom->molecular == 0) {
    int tag_lowest_all;
    MPI_Allreduce(&tag_lowest,&tag_lowest_all,1,MPI_INT,MPI_MIN,world);

    int *tag = atom->tag;
    int nlocal = atom->nlocal;
 
    for (int i = 0; i < nlocal; i++)
      if (tag[i] >= tag_lowest_all) tag[i] = 0;

    atom->tag_extend();
  }

  // reset atom->natoms
  // reset atom->map if a molecular system
  // set nghost to 0 so old ghosts of deleted atoms won't be mapped

  double rlocal = atom->nlocal;
  MPI_Allreduce(&rlocal,&atom->natoms,1,MPI_DOUBLE,MPI_SUM,world);
  if (atom->molecular) {
    atom->nghost = 0;
    atom->map_init();
    atom->map_set();
  }

  // print atoms before and after

  int ndelete = static_cast<int> (natoms_previous - atom->natoms);

  if (comm->me == 0) {
    if (screen) fprintf(screen,"Deleted %d atoms, new total = %.15g\n",
			ndelete,atom->natoms);
    if (logfile) fprintf(logfile,"Deleted %d atoms, new total = %.15g\n",
			 ndelete,atom->natoms);
  }
}
