/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "mpi.h"
#include "string.h"
#include "dump.h"
#include "atom.h"
#include "group.h"
#include "error.h"
#include "memory.h"

/* ---------------------------------------------------------------------- */

Dump::Dump(int narg, char **arg)
{
  MPI_Comm_rank(world,&me);
  MPI_Comm_size(world,&nprocs);

  int n = strlen(arg[0]) + 1;
  id = new char[n];
  strcpy(id,arg[0]);

  igroup = group->find(arg[1]);
  groupbit = group->bitmask[igroup];

  n = strlen(arg[2]) + 1;
  style = new char[n];
  strcpy(style,arg[2]);

  scale_flag = 1;
  image_flag = 0;
  header_flag = 0;
  flush_flag = 0;

  format = NULL;
  format_user = NULL;

  maxbuf = 0;
  buf = NULL;

  if (me == 0) {
    fp = fopen(arg[4],"w");
    if (fp == NULL) error->one("Could not open dump file");
  }
}

/* ---------------------------------------------------------------------- */

Dump::~Dump()
{
  delete [] id;
  delete [] style;

  delete [] format;
  delete [] format_default;
  delete [] format_user;

  memory->sfree(buf);
  if (me == 0) fclose(fp);
}

/* ---------------------------------------------------------------------- */

void Dump::write()
{
  int iproc,send_size,recv_size;

  // n = # of dump lines this proc will contribute to dump
  // ndump = total # of dump lines
  // nmax = max # on any proc

  int n,ndump,nmax;
  n = count();
  MPI_Allreduce(&n,&ndump,1,MPI_INT,MPI_SUM,world);
  MPI_Allreduce(&n,&nmax,1,MPI_INT,MPI_MAX,world);

  // proc 0 writes header to file

  if (me == 0) write_header(ndump);

  // grow communication buffer if necessary

  if (nmax*size_one > maxbuf) {
    maxbuf = nmax*size_one;
    memory->sfree(buf);
    buf = (double *) memory->smalloc(maxbuf*sizeof(double),"dump:buf");
  }

  // pack my data into buf
  // send_size = # of quantities in buf

  send_size = pack();

  // proc 0 pings each proc, receives it's data, writes to file
  // all other procs wait for ping, send their data to proc 0

  int tmp;
  MPI_Status status;
  MPI_Request request;

  if (me == 0) {
    for (iproc = 0; iproc < nprocs; iproc++) {
      if (iproc) {
	MPI_Irecv(buf,maxbuf,MPI_DOUBLE,iproc,0,world,&request);
	MPI_Send(&tmp,0,MPI_INT,iproc,0,world);
	MPI_Wait(&request,&status);
	MPI_Get_count(&status,MPI_DOUBLE,&recv_size);
	recv_size /= size_one;
      } else recv_size = send_size/size_one;

      write_data(recv_size,buf);
    }
    if (flush_flag) fflush(fp);

  } else {
    MPI_Recv(&tmp,0,MPI_INT,0,0,world,&status);
    MPI_Rsend(buf,send_size,MPI_DOUBLE,0,0,world);
  }
}

/* ----------------------------------------------------------------------
   modify parameters of a Dump 
------------------------------------------------------------------------- */

void Dump::modify_params(int narg, char **arg)
{
  int iarg = 0;
  while (iarg < narg) {
    if (strcmp(arg[iarg],"format") == 0) {
      if (iarg+2 > narg) error->all("Illegal dump_modify command");
      delete [] format_user;
      format_user = NULL;
      if (strcmp(arg[iarg+1],"none")) {
	int n = strlen(arg[iarg+1]) + 1;
	format_user = new char[n];
	strcpy(format_user,arg[iarg+1]);
      }
      iarg += 2;
    } else if (strcmp(arg[iarg],"scale") == 0) {
      if (iarg+2 > narg) error->all("Illegal dump_modify command");
      if (strcmp(arg[iarg+1],"yes") == 0) scale_flag = 1;
      else if (strcmp(arg[iarg+1],"no") == 0) scale_flag = 0;
      else error->all("Illegal dump_modify command");
      iarg += 2;
    } else if (strcmp(arg[iarg],"image") == 0) {
      if (iarg+2 > narg) error->all("Illegal dump_modify command");
      if (strcmp(arg[iarg+1],"yes") == 0) image_flag = 1;
      else if (strcmp(arg[iarg+1],"no") == 0) image_flag = 0;
      else error->all("Illegal dump_modify command");
      iarg += 2;
    } else if (strcmp(arg[iarg],"header") == 0) {
      if (iarg+2 > narg) error->all("Illegal dump_modify command");
      if (strcmp(arg[iarg+1],"item") == 0) header_flag = 0;
      else if (strcmp(arg[iarg+1],"self") == 0) header_flag = 1;
      else if (strcmp(arg[iarg+1],"xyz") == 0) header_flag = 2;
      else error->all("Illegal dump_modify command");
      iarg += 2;
    } else if (strcmp(arg[iarg],"flush") == 0) {
      if (iarg+2 > narg) error->all("Illegal dump_modify command");
      if (strcmp(arg[iarg+1],"yes") == 0) flush_flag = 1;
      else if (strcmp(arg[iarg+1],"no") == 0) flush_flag = 0;
      else error->all("Illegal dump_modify command");
      iarg += 2;
    } else error->all("Incorrect args in dump_modify command");
  }
}
