/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "string.h"
#include "dump_custom.h"
#include "atom.h"
#include "domain.h"
#include "update.h"
#include "modify.h"
#include "fix.h"
#include "fix_centro.h"
#include "fix_energy.h"
#include "fix_stress.h"
#include "error.h"

#define INT    1
#define DOUBLE 2

#define TAG     1
#define TYPE    2
#define X       3
#define Y       4
#define Z       5
#define XS      6
#define YS      7
#define ZS      8
#define IX      9
#define IY     10
#define IZ     11
#define VX     12
#define VY     13
#define VZ     14
#define FX     15
#define FY     16
#define FZ     17
#define Q      18
#define MUX    19
#define MUY    20
#define MUZ    21
#define TQX    22
#define TQY    23
#define TQZ    24
#define CENTRO 25
#define ENG    26
#define SXX    27
#define SYY    28
#define SZZ    29
#define SXY    30
#define SXZ    31
#define SYZ    32

/* ---------------------------------------------------------------------- */

DumpCustom::DumpCustom(int narg, char **arg) : Dump(narg, arg)
{
  if (narg == 5) error->all("No dump custom arguments specified");

  size_one = narg-5;
  vname = new int[size_one];
  vtype = new int[size_one];
  pack_choice = new FnPtrPack[size_one];

  ifix_centro = -1;
  ifix_energy = -1;
  ifix_stress = -1;

  int i;
  for (int iarg = 5; iarg < narg; iarg++) {
    i = iarg-5;
    if (strcmp(arg[iarg],"tag") == 0) {
      vname[i] = TAG;
      vtype[i] = INT;
      pack_choice[i] = &DumpCustom::pack_tag;
    } else if (strcmp(arg[iarg],"type") == 0) {
      vname[i] = TYPE;
      vtype[i] = INT;
      pack_choice[i] = &DumpCustom::pack_type;
    } else if (strcmp(arg[iarg],"x") == 0) {
      vname[i] = X;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_x;
    } else if (strcmp(arg[iarg],"y") == 0) {
      vname[i] = Y;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_y;
    } else if (strcmp(arg[iarg],"z") == 0) {
      vname[i] = Z;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_z;
    } else if (strcmp(arg[iarg],"xs") == 0) {
      vname[i] = XS;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_xs;
    } else if (strcmp(arg[iarg],"ys") == 0) {
      vname[i] = YS;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_ys;
    } else if (strcmp(arg[iarg],"zs") == 0) {
      vname[i] = ZS;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_zs;
    } else if (strcmp(arg[iarg],"ix") == 0) {
      vname[i] = IX;
      vtype[i] = INT;
      pack_choice[i] = &DumpCustom::pack_ix;
    } else if (strcmp(arg[iarg],"iy") == 0) {
      vname[i] = IY;
      vtype[i] = INT;
      pack_choice[i] = &DumpCustom::pack_iy;
    } else if (strcmp(arg[iarg],"iz") == 0) {
      vname[i] = IZ;
      vtype[i] = INT;
      pack_choice[i] = &DumpCustom::pack_iz;
    } else if (strcmp(arg[iarg],"vx") == 0) {
      vname[i] = VX;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_vx;
    } else if (strcmp(arg[iarg],"vy") == 0) {
      vname[i] = VY;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_vy;
    } else if (strcmp(arg[iarg],"vz") == 0) {
      vname[i] = VZ;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_vz;
    } else if (strcmp(arg[iarg],"fx") == 0) {
      vname[i] = FX;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_fx;
    } else if (strcmp(arg[iarg],"fy") == 0) {
      vname[i] = FY;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_fy;
    } else if (strcmp(arg[iarg],"fz") == 0) {
      vname[i] = FZ;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_fz;
    } else if (strcmp(arg[iarg],"q") == 0) {
      vname[i] = Q;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_q;
    } else if (strcmp(arg[iarg],"mux") == 0) {
      vname[i] = MUX;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_mux;
    } else if (strcmp(arg[iarg],"muy") == 0) {
      vname[i] = MUY;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_muy;
    } else if (strcmp(arg[iarg],"muz") == 0) {
      vname[i] = MUZ;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_muz;
    } else if (strcmp(arg[iarg],"tqx") == 0) {
      vname[i] = TQX;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_tqx;
    } else if (strcmp(arg[iarg],"tqy") == 0) {
      vname[i] = TQY;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_tqy;
    } else if (strcmp(arg[iarg],"tqz") == 0) {
      vname[i] = TQZ;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_tqz;
    } else if (strcmp(arg[iarg],"centro") == 0) {
      vname[i] = CENTRO;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_centro;
      if (ifix_centro < 0) ifix_centro = fix_create("CENTRO",arg);
    } else if (strcmp(arg[iarg],"eng") == 0) {
      vname[i] = ENG;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_energy;
      if (ifix_energy < 0) ifix_energy = fix_create("ENERGY",arg);
    } else if (strcmp(arg[iarg],"sxx") == 0) {
      vname[i] = SXX;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_sxx;
      if (ifix_stress < 0) ifix_stress = fix_create("STRESS",arg);
    } else if (strcmp(arg[iarg],"syy") == 0) {
      vname[i] = SYY;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_syy;
      if (ifix_stress < 0) ifix_stress = fix_create("STRESS",arg);
    } else if (strcmp(arg[iarg],"szz") == 0) {
      vname[i] = SZZ;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_szz;
      if (ifix_stress < 0) ifix_stress = fix_create("STRESS",arg);
    } else if (strcmp(arg[iarg],"sxy") == 0) {
      vname[i] = SXY;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_sxy;
      if (ifix_stress < 0) ifix_stress = fix_create("STRESS",arg);
    } else if (strcmp(arg[iarg],"sxz") == 0) {
      vname[i] = SXZ;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_sxz;
      if (ifix_stress < 0) ifix_stress = fix_create("STRESS",arg);
    } else if (strcmp(arg[iarg],"syz") == 0) {
      vname[i] = SYZ;
      vtype[i] = DOUBLE;
      pack_choice[i] = &DumpCustom::pack_syz;
      if (ifix_stress < 0) ifix_stress = fix_create("STRESS",arg);

    } else error->all("Illegal vector in dump custom command");
  }

  vformat = new char*[size_one];

  format_default = new char[3*size_one+1];
  format_default[0] = '\0';

  for (i = 0; i < size_one; i++) {
    if (vtype[i] == INT) format_default = strcat(format_default,"%d ");
    else format_default = strcat(format_default,"%g ");
    vformat[i] = NULL;
  }
}

/* ----------------------------------------------------------------------
   free memory
------------------------------------------------------------------------- */

DumpCustom::~DumpCustom()
{
  delete [] vname;
  delete [] vtype;
  for (int i = 0; i < size_one; i++) delete [] vformat[i];
  delete [] vformat;
  delete [] pack_choice;

  // delete fixes for computing per-atom custom quantities if they exist

  if (ifix_centro >= 0) fix_delete("CENTRO");
  if (ifix_energy >= 0) fix_delete("ENERGY");
  if (ifix_stress >= 0) fix_delete("STRESS");
}

/* ---------------------------------------------------------------------- */

void DumpCustom::init()
{
  delete [] format;
  char *str;
  if (format_user) str = format_user;
  else str = format_default;

  int n = strlen(str) + 1;
  format = new char[n];
  strcpy(format,str);

  // tokenize the format string and add space at end of each format element

  char *ptr;
  for (int i = 0; i < size_one; i++) {
    if (i == 0) ptr = strtok(format," \0");
    else ptr = strtok(NULL," \0");
    vformat[i] = new char[strlen(ptr) + 2];
    strcpy(vformat[i],ptr);
    vformat[i] = strcat(vformat[i]," ");
  }

  // setup function ptrs

  if (header_flag == 0) header_choice = &DumpCustom::header_item;
  else if (header_flag == 1) header_choice = &DumpCustom::header_self;
  else if (header_flag == 2) header_choice = &DumpCustom::header_xyz;
}

/* ---------------------------------------------------------------------- */

void DumpCustom::write_header(int ndump)
{
  (this->*header_choice)(ndump);
}

/* ---------------------------------------------------------------------- */

int DumpCustom::count()
{
  if (igroup == 0) {
    nmine = atom->nlocal;
    return nmine;
  }

  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  nmine = 0;
  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) nmine++;
  return nmine;
}

/* ---------------------------------------------------------------------- */

int DumpCustom::pack()
{
  for (int n = 0; n < size_one; n++) (this->*pack_choice[n])(n);
  return nmine*size_one;
}

/* ---------------------------------------------------------------------- */

void DumpCustom::write_data(int n, double *buf)
{
  int i,j;

  int m = 0;
  for (i = 0; i < n; i++) {
    for (j = 0; j < size_one; j++) {
      if (vtype[j] == INT) fprintf(fp,vformat[j],static_cast<int> (buf[m]));
      else fprintf(fp,vformat[j],buf[m]);
      m++;
    }
    fprintf(fp,"\n");
  }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::header_item(int ndump)
{
  fprintf(fp,"ITEM: TIMESTEP\n");
  fprintf(fp,"%d\n",update->ntimestep);
  fprintf(fp,"ITEM: NUMBER OF ATOMS\n");
  fprintf(fp,"%d\n",ndump);
  fprintf(fp,"ITEM: BOX BOUNDS\n");
  fprintf(fp,"%g %g\n",domain->boxxlo,domain->boxxhi);
  fprintf(fp,"%g %g\n",domain->boxylo,domain->boxyhi);
  fprintf(fp,"%g %g\n",domain->boxzlo,domain->boxzhi);
  //fprintf(fp,"%16.12g %16.12g\n",domain->boxxlo,domain->boxxhi);
  //fprintf(fp,"%16.12g %16.12g\n",domain->boxylo,domain->boxyhi);
  //fprintf(fp,"%16.12g %16.12g\n",domain->boxzlo,domain->boxzhi);
  fprintf(fp,"ITEM: ATOMS\n");
}

/* ---------------------------------------------------------------------- */

void DumpCustom::header_self(int ndump)
{
  fprintf(fp,"Timestep 1 1\n");
  fprintf(fp,"%d\n",update->ntimestep);
  fprintf(fp,"BoxBounds 3 2\n");
  fprintf(fp,"%g %g\n",domain->boxxlo,domain->boxxhi);
  fprintf(fp,"%g %g\n",domain->boxylo,domain->boxyhi);
  fprintf(fp,"%g %g\n",domain->boxzlo,domain->boxzhi);
  fprintf(fp,"Atoms %d %d\n",ndump,size_one);
}

/* ---------------------------------------------------------------------- */

void DumpCustom::header_xyz(int ndump)
{
  fprintf(fp,"%d\n",ndump);
  fprintf(fp,"Atoms\n");
}

/* ----------------------------------------------------------------------
   create a fix of style keyword for computing a per-atom custom quantity 
------------------------------------------------------------------------- */

int DumpCustom::fix_create(char *keyword, char **arg)
{
  char **fixarg = new char*[4];
  fixarg[0] = new char[strlen(arg[0]) + strlen(keyword) + 1];
  strcpy(fixarg[0],arg[0]);
  strcat(fixarg[0],keyword);
  fixarg[1] = arg[1];
  fixarg[2] = keyword;
  fixarg[3] = arg[3];
  modify->add_fix(4,fixarg);
  delete [] fixarg[0];
  delete [] fixarg;
  return (modify->nfix - 1);
}

/* ----------------------------------------------------------------------
   delete a fix of style keyword for computing a per-atom custom quantity 
------------------------------------------------------------------------- */

void DumpCustom::fix_delete(char *keyword)
{
  char *name = new char[strlen(id) + strlen(keyword) + 1];
  strcpy(name,id);
  strcat(name,keyword);
  modify->delete_fix(name);
  delete [] name;
}

/* ----------------------------------------------------------------------
   one routine for every kind of quantity dump custom can output
   the atom quantity is packed into buf starting at n with stride size_one
------------------------------------------------------------------------- */

void DumpCustom::pack_tag(int n)
{
  int *tag = atom->tag;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = tag[i];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_type(int n)
{
  int *type = atom->type;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = type[i];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_x(int n)
{
  double **x = atom->x;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = x[i][0];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_y(int n)
{
  double **x = atom->x;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = x[i][1];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_z(int n)
{
  double **x = atom->x;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = x[i][2];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_xs(int n)
{
  double **x = atom->x;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  double boxxlo = domain->boxxlo;
  double invxprd = 1.0/domain->xprd;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = (x[i][0] - boxxlo) * invxprd;
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_ys(int n)
{
  double **x = atom->x;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  double boxylo = domain->boxylo;
  double invyprd = 1.0/domain->yprd;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = (x[i][1] - boxylo) * invyprd;
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_zs(int n)
{
  double **x = atom->x;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  double boxzlo = domain->boxzlo;
  double invzprd = 1.0/domain->zprd;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = (x[i][2] - boxzlo) * invzprd;
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_ix(int n)
{
  int *image = atom->image;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = image[i]%1000 - 500;
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_iy(int n)
{
  int *image = atom->image;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = (image[i]/1000)%1000 - 500;
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_iz(int n)
{
  int *image = atom->image;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = image[i]/1000000 - 500;
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_vx(int n)
{
  double **v = atom->v;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = v[i][0];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_vy(int n)
{
  double **v = atom->v;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = v[i][1];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_vz(int n)
{
  double **v = atom->v;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = v[i][2];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_fx(int n)
{
  double **f = atom->f;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = f[i][0];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_fy(int n)
{
  double **f = atom->f;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = f[i][1];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_fz(int n)
{
  double **f = atom->f;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = f[i][2];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_q(int n)
{
  double *q = atom->q;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = q[i];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_mux(int n)
{
  double **mu = atom->mu;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = mu[i][0];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_muy(int n)
{
  double **mu = atom->mu;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = mu[i][1];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_muz(int n)
{
  double **mu = atom->mu;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = mu[i][2];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_tqx(int n)
{
  double **torque = atom->torque;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = torque[i][0];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_tqy(int n)
{
  double **torque = atom->torque;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = torque[i][1];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_tqz(int n)
{
  double **torque = atom->torque;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = torque[i][2];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_centro(int n)
{
  double *centro = ((FixCentro *) modify->fix[ifix_centro])->centro;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = centro[i];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_energy(int n)
{
  double *energy = ((FixEnergy *) modify->fix[ifix_energy])->energy;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = energy[i];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_sxx(int n)
{
  double **stress = ((FixStress *) modify->fix[ifix_stress])->stress;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = stress[i][0];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_syy(int n)
{
  double **stress = ((FixStress *) modify->fix[ifix_stress])->stress;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = stress[i][1];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_szz(int n)
{
  double **stress = ((FixStress *) modify->fix[ifix_stress])->stress;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = stress[i][2];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_sxy(int n)
{
  double **stress = ((FixStress *) modify->fix[ifix_stress])->stress;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = stress[i][3];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_sxz(int n)
{
  double **stress = ((FixStress *) modify->fix[ifix_stress])->stress;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = stress[i][4];
      n += size_one;
    }
}

/* ---------------------------------------------------------------------- */

void DumpCustom::pack_syz(int n)
{
  double **stress = ((FixStress *) modify->fix[ifix_stress])->stress;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[n] = stress[i][5];
      n += size_one;
    }
}
