/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "string.h"
#include "dump_velocity.h"
#include "atom.h"
#include "domain.h"
#include "update.h"
#include "error.h"

DumpVelocity::DumpVelocity(int narg, char **arg) : Dump(narg, arg)
{
  if (narg != 5) error->all("Illegal dump velocity command");

  size_one = 5;

  char *str = "%d %d %g %g %g";
  int n = strlen(str) + 1;
  format_default = new char[n];
  strcpy(format_default,str);
}

/* ---------------------------------------------------------------------- */

void DumpVelocity::init()
{
  delete [] format;
  char *str;
  if (format_user) str = format_user;
  else str = format_default;

  int n = strlen(str) + 2;
  format = new char[n];
  strcpy(format,str);
  strcat(format,"\n");
}

/* ---------------------------------------------------------------------- */

void DumpVelocity::write_header(int ndump)
{
  if (header_flag == 0) {
    fprintf(fp,"ITEM: TIMESTEP\n");
    fprintf(fp,"%d\n",update->ntimestep);
    fprintf(fp,"ITEM: NUMBER OF ATOMS\n");
    fprintf(fp,"%d\n",ndump);
    fprintf(fp,"ITEM: BOX BOUNDS\n");
    fprintf(fp,"%g %g\n",domain->boxxlo,domain->boxxhi);
    fprintf(fp,"%g %g\n",domain->boxylo,domain->boxyhi);
    fprintf(fp,"%g %g\n",domain->boxzlo,domain->boxzhi);
    fprintf(fp,"ITEM: VELOCITIES\n");
  } else if (header_flag == 1) {
    fprintf(fp,"Timestep 1 1\n");
    fprintf(fp,"%d\n",update->ntimestep);
    fprintf(fp,"BoxBounds 3 2\n");
    fprintf(fp,"%g %g\n",domain->boxxlo,domain->boxxhi);
    fprintf(fp,"%g %g\n",domain->boxylo,domain->boxyhi);
    fprintf(fp,"%g %g\n",domain->boxzlo,domain->boxzhi);
    fprintf(fp,"Velocities %d %d\n",ndump,size_one);
  } else if (header_flag == 2) {
    fprintf(fp,"%d\n",ndump);
    fprintf(fp,"Velocities\n");
  }
}

/* ---------------------------------------------------------------------- */

int DumpVelocity::count()
{
  if (igroup == 0) return atom->nlocal;

  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  int m = 0;
  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) m++;
  return m;
}

/* ---------------------------------------------------------------------- */

int DumpVelocity::pack()
{
  int *tag = atom->tag;
  int *type = atom->type;
  int *mask = atom->mask;
  double **v = atom->v;
  int nlocal = atom->nlocal;

  int m = 0;
  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      buf[m++] = tag[i];
      buf[m++] = type[i];
      buf[m++] = v[i][0];
      buf[m++] = v[i][1];
      buf[m++] = v[i][2];
    }
  return m;
}

/* ---------------------------------------------------------------------- */

void DumpVelocity::write_data(int n, double *buf)
{
  int m = 0;
  for (int i = 0; i < n; i++) {
    fprintf(fp,format,
	    static_cast<int> (buf[m]), static_cast<int> (buf[m+1]),
	    buf[m+2],buf[m+3],buf[m+4]);
    m += size_one;
  }
}
