/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "string.h"
#include "fix_com.h"
#include "atom.h"
#include "update.h"
#include "domain.h"
#include "group.h"
#include "error.h"

/* ---------------------------------------------------------------------- */

FixCOM::FixCOM(int narg, char **arg) : Fix(narg, arg)
{
  if (narg != 5) error->all("Illegal fix com command");

  MPI_Comm_rank(world,&me);

  nevery = atoi(arg[3]);
  if (nevery <= 0) error->all("Illegal fix com command");
  first = 1;

  if (me == 0) {
    fp = fopen(arg[4],"w");
    if (fp == NULL) {
      char str[128];
      sprintf(str,"Cannot open fix com file %s",arg[4]);
      error->one(str);
    }
  }

  if (me == 0) {
    fprintf(fp,"Center-of-mass for group %s\n",group->names[igroup]);
    fprintf(fp,"TimeStep x y z\n");
  }
}

/* ---------------------------------------------------------------------- */

FixCOM::~FixCOM()
{
  if (me == 0) fclose(fp);
}

/* ---------------------------------------------------------------------- */

int FixCOM::setmask()
{
  int mask = 0;
  mask |= END_OF_STEP;
  return mask;
}

/* ---------------------------------------------------------------------- */

void FixCOM::init()
{
  // masstotal = total mass of atoms in group

  int *mask = atom->mask;
  int *type = atom->type;
  double *mass = atom->mass;
  int nlocal = atom->nlocal;

  double rmass = 0.0;
  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) rmass += mass[type[i]];
  MPI_Allreduce(&rmass,&masstotal,1,MPI_DOUBLE,MPI_SUM,world);
}

/* ---------------------------------------------------------------------- */

void FixCOM::setup()
{
  if (first) end_of_step();
  first = 0;
}

/* ---------------------------------------------------------------------- */

void FixCOM::end_of_step()
{
  if (update->ntimestep % nevery) return;

  double **x = atom->x;
  int *mask = atom->mask;
  int *type = atom->type;
  int *image = atom->image;
  double *mass = atom->mass;
  int nlocal = atom->nlocal;

  double com[3];
  com[0] = com[1] = com[2] = 0.0;
  double xprd = domain->xprd;
  double yprd = domain->yprd;
  double zprd = domain->zprd;

  int xbox,ybox,zbox;
  double massone;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      xbox = image[i]%1000 - 500;
      ybox = (image[i]/1000) % 1000 - 500;
      zbox = image[i]/1000000 - 500;
      massone = mass[type[i]];
      com[0] += (x[i][0] + xbox*xprd) * massone;
      com[1] += (x[i][1] + ybox*yprd) * massone;
      com[2] += (x[i][2] + zbox*zprd) * massone;
    }

  double com_all[3];
  MPI_Allreduce(com,com_all,3,MPI_DOUBLE,MPI_SUM,world);
  com_all[0] /= masstotal;
  com_all[1] /= masstotal;
  com_all[2] /= masstotal;

  if (me == 0) fprintf(fp,"%d %g %g %g\n",
		       update->ntimestep,com_all[0],com_all[1],com_all[2]);
}
