/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "math.h"
#include "string.h"
#include "fix_indent.h"
#include "atom.h"
#include "domain.h"
#include "update.h"
#include "respa.h"
#include "error.h"

/* ---------------------------------------------------------------------- */

FixIndent::FixIndent(int narg, char **arg) : Fix(narg, arg)
{
  if (narg != 11) error->all("Illegal fix indent command");

  if (strcmp(domain->lattice_style,"none") == 0)
    error->all("Use of fix indent with undefined lattice");

  double xscale = domain->xlattice;
  double yscale = domain->ylattice;
  double zscale = domain->zlattice;

  k = atof(arg[3]);
  r0 = xscale*atof(arg[4]);
  x0 = xscale*atof(arg[5]);
  y0 = yscale*atof(arg[6]);
  z0 = zscale*atof(arg[7]);
  vx = xscale * atof(arg[8]);
  vy = yscale * atof(arg[9]);
  vz = zscale * atof(arg[10]);

  ntimestep_initial = update->ntimestep;
}

/* ---------------------------------------------------------------------- */

int FixIndent::setmask()
{
  int mask = 0;
  mask |= POST_FORCE;
  mask |= POST_FORCE_RESPA;
  return mask;
}

/* ---------------------------------------------------------------------- */

void FixIndent::init()
{
  if (strcmp(update->integrate_style,"respa") == 0)
    nlevels_respa = ((Respa *) update->integrate)->nlevels;
}

/* ---------------------------------------------------------------------- */

void FixIndent::setup()
{
  if (strcmp(update->integrate_style,"verlet") == 0)
    post_force(1);
  else {
    ((Respa *) update->integrate)->copy_flevel_f(nlevels_respa-1);
    post_force_respa(1,nlevels_respa-1,0);
    ((Respa *) update->integrate)->copy_f_flevel(nlevels_respa-1);
  }
}

/* ---------------------------------------------------------------------- */

void FixIndent::post_force(int vflag)
{
  // x1,y1,z1 = current position of indenter

  double delta = (update->ntimestep - ntimestep_initial) * update->dt;
  double x1 = x0 + delta*vx;
  double y1 = y0 + delta*vy;
  double z1 = z0 + delta*vz;

  double **x = atom->x;
  double **f = atom->f;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  double delx,dely,delz,r,dr,fmag;

  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      delx = x[i][0] - x1;
      dely = x[i][1] - y1;
      delz = x[i][2] - z1;
      r = sqrt(delx*delx + dely*dely + delz*delz);
      dr = r - r0;
      if (dr < 0.0) {
	fmag = k*dr*dr;
	f[i][0] += delx*fmag/r;
	f[i][1] += dely*fmag/r;
	f[i][2] += delz*fmag/r;
      }
    }
}

/* ---------------------------------------------------------------------- */

void FixIndent::post_force_respa(int vflag, int ilevel, int iloop)
{
  if (ilevel == nlevels_respa-1) post_force(vflag);
}
