/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "mpi.h"
#include "math.h"
#include "string.h"
#include "stdlib.h"
#include "fix_langevin.h"
#include "atom.h"
#include "force.h"
#include "update.h"
#include "respa.h"
#include "random_mars.h"
#include "error.h"

/* ---------------------------------------------------------------------- */

FixLangevin::FixLangevin(int narg, char **arg) : Fix(narg, arg)
{
  if (narg != 7 && narg != 10) error->all("Illegal fix langevin command");

  t_start = atof(arg[3]);
  t_stop = atof(arg[4]);
  double t_period = atof(arg[5]);

  int seed = atoi(arg[6]);
  int me;
  MPI_Comm_rank(world,&me);
  ranmars = new RanMars(seed+me);

  if (narg == 7) flagx = flagy = flagz = 1;
  else {
    flagx = atoi(arg[7]);
    flagy = atoi(arg[8]);
    flagz = atoi(arg[9]);
  }

  // convert input period to frequency

  if (t_period <= 0.0) error->all("Fix langevin period must be > 0.0");
  t_freq = 1.0 / t_period;
}

/* ---------------------------------------------------------------------- */

FixLangevin::~FixLangevin()
{
  delete ranmars;
}

/* ---------------------------------------------------------------------- */

int FixLangevin::setmask()
{
  int mask = 0;
  mask |= POST_FORCE;
  mask |= POST_FORCE_RESPA;
  return mask;
}

/* ---------------------------------------------------------------------- */

void FixLangevin::init()
{
  dthalf = 0.5 * update->dt;
  boltz = force->boltz;
  ftm2v = force->ftm2v;
  mvv2e = force->mvv2e;

  if (strcmp(update->integrate_style,"respa") == 0)
    nlevels_respa = ((Respa *) update->integrate)->nlevels;
}

/* ---------------------------------------------------------------------- */

void FixLangevin::setup()
{
  if (strcmp(update->integrate_style,"verlet") == 0)
    post_force(1);
  else {
    ((Respa *) update->integrate)->copy_flevel_f(nlevels_respa-1);
    post_force_respa(1,nlevels_respa-1,0);
    ((Respa *) update->integrate)->copy_f_flevel(nlevels_respa-1);
  }
}

/* ---------------------------------------------------------------------- */

void FixLangevin::post_force(int vflag)
{
  double rmass,gamma1,gamma2;

  double **v = atom->v;
  double **f = atom->f;
  double *mass = atom->mass;
  int *type = atom->type;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  double delta = update->ntimestep - update->firststep;
  delta /= update->nsteps;
  double t_target = t_start + delta * (t_stop-t_start);
  double gconst1 = -t_freq / ftm2v;
  double gconst2 = sqrt(12.0*boltz*t_target*t_freq/dthalf/mvv2e) / ftm2v;

  for (int i = 0; i < nlocal; i++) {
    if (mask[i] & groupbit) {
      rmass = mass[type[i]];
      gamma1 = gconst1 * rmass;
      gamma2 = gconst2 * sqrt(rmass);
      if (flagx) f[i][0] += gamma1*v[i][0] + gamma2*(ranmars->uniform()-0.5);
      if (flagy) f[i][1] += gamma1*v[i][1] + gamma2*(ranmars->uniform()-0.5);
      if (flagz) f[i][2] += gamma1*v[i][2] + gamma2*(ranmars->uniform()-0.5);
    }
  }
}

/* ---------------------------------------------------------------------- */

void FixLangevin::post_force_respa(int vflag, int ilevel, int iloop)
{
  if (ilevel == nlevels_respa-1) post_force(vflag);
}

/* ---------------------------------------------------------------------- */

void FixLangevin::reset_target(double t_new)
{
  t_start = t_stop = t_new;
}
