/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
   Contributing author: Mark Stevens (SNL)
------------------------------------------------------------------------- */

#include "string.h"
#include "stdlib.h"
#include "math.h"
#include "fix_npt.h"
#include "atom.h"
#include "force.h"
#include "comm.h"
#include "output.h"
#include "modify.h"
#include "fix_rigid.h"
#include "kspace.h"
#include "update.h"
#include "respa.h"
#include "temperature.h"
#include "pressure.h"
#include "domain.h"
#include "error.h"

/* ---------------------------------------------------------------------- */

FixNPT::FixNPT(int narg, char **arg) : Fix(narg, arg)
{
  if (narg < 7) error->all("Illegal fix npt command");

  restart_global = 1;

  t_start = atof(arg[3]);
  t_stop = atof(arg[4]);
  double t_period = atof(arg[5]);

  double p_period[3];
  if (strcmp(arg[6],"xyz") == 0) {
    press_couple = 0;
    p_start[0] = p_start[1] = p_start[2] = atof(arg[7]);
    p_stop[0] = p_stop[1] = p_stop[2] = atof(arg[8]);
    p_period[0] = p_period[1] = p_period[2] = atof(arg[9]);
    p_flag[0] = p_flag[1] = p_flag[2] = 1;
  } else {
    if (strcmp(arg[6],"xy") == 0) press_couple = 1;
    else if (strcmp(arg[6],"yz") == 0) press_couple = 2;
    else if (strcmp(arg[6],"xz") == 0) press_couple = 3;
    else if (strcmp(arg[6],"aniso") == 0) press_couple = 4;
    else error->all("Illegal fix npt command");

    if (strcmp(arg[7],"NULL") == 0) {
      p_start[0] = p_stop[0] = p_period[0] = 0.0;
      p_flag[0] = 0;
    } else {
      p_start[0] = atof(arg[7]);
      p_stop[0] = atof(arg[8]);
      p_flag[0] = 1;
    }
    if (strcmp(arg[9],"NULL") == 0) {
      p_start[1] = p_stop[1] = p_period[1] = 0.0;
      p_flag[1] = 0;
    } else {
      p_start[1] = atof(arg[9]);
      p_stop[1] = atof(arg[10]);
      p_flag[1] = 1;
    }
    if (strcmp(arg[11],"NULL") == 0) {
      p_start[2] = p_stop[2] = p_period[2] = 0.0;
      p_flag[2] = 0;
    } else {
      p_start[2] = atof(arg[11]);
      p_stop[2] = atof(arg[12]);
      p_flag[2] = 1;
    }

    double period = atof(arg[13]);
    if (p_flag[0]) p_period[0] = period;
    if (p_flag[1]) p_period[1] = period;
    if (p_flag[2]) p_period[2] = period;
  }

  // check for periodicity in controlled dimensions

  if (p_flag[0] && domain->xperiodic == 0)
    error->all("Cannot fix npt on a non-periodic dimension");
  if (p_flag[1] && domain->yperiodic == 0)
    error->all("Cannot fix npt on a non-periodic dimension");
  if (p_flag[2] && domain->zperiodic == 0)
    error->all("Cannot fix npt on a non-periodic dimension");

  // temperature init
  // no need to create a new temperature (like for NVT)
  // this is because pressure is always global (all) and thus its
  //   KE/temperature contribution should use group all

  temp_chosen = 0;
  eta = eta_dot = 0.0;

  // convert input periods to frequencies

  if (t_period <= 0.0 || (p_flag[0] && p_period[0] <= 0.0) || 
      (p_flag[1] && p_period[1] <= 0.0) || (p_flag[2] && p_period[2] <= 0.0))
    error->all("Fix npt periods must be > 0.0");

  t_freq = 1.0 / t_period;
  p_freq[0] = p_freq[1] = p_freq[2] = 0.0;
  if (p_flag[0]) p_freq[0] = 1.0 / p_period[0];
  if (p_flag[1]) p_freq[1] = 1.0 / p_period[1];
  if (p_flag[2]) p_freq[2] = 1.0 / p_period[2];

  // pressure init

  pressure = force->pressure;
  omega[0] = omega[1] = omega[2] = 0.0;
  omega_dot[0] = omega_dot[1] = omega_dot[2] = 0.0;

  nrigid = 0;
  rfix = NULL;
}

/* ---------------------------------------------------------------------- */

FixNPT::~FixNPT()
{
  delete [] rfix;
}

/* ---------------------------------------------------------------------- */

int FixNPT::setmask()
{
  int mask = 0;
  mask |= INITIAL_INTEGRATE;
  mask |= FINAL_INTEGRATE;
  mask |= INITIAL_INTEGRATE_RESPA;
  mask |= FINAL_INTEGRATE_RESPA;
  return mask;
}

/* ---------------------------------------------------------------------- */

void FixNPT::init()
{
  // warn if temperature computed for global pressure is not for group all
  // this can happen if fix_modify used to change temperature assigned to NPT

  temperature = force->templist[temp_chosen];
  if (temperature->igroup != 0 && comm->me == 0)
    error->warning("Temperature for NPT pressure is not for group all");

  dtv = update->dt;
  dtf = 0.5 * update->dt * force->ftm2v;
  dthalf = 0.5 * update->dt;

  boltz = force->boltz;
  nktv2p = force->nktv2p;

  double mass = 0.0;
  for (int i = 0; i < atom->nlocal; i++) mass += atom->mass[atom->type[i]];
  MPI_Allreduce(&mass,&total_mass,1,MPI_DOUBLE,MPI_SUM,world);

  if (force->kspace) kspace_flag = 1;
  else kspace_flag = 0;

  if (strcmp(update->integrate_style,"respa") == 0) {
    nlevels_respa = ((Respa *) update->integrate)->nlevels;
    step_respa = ((Respa *) update->integrate)->step;
  }

  // detect if any fix rigid exist so rigid bodies move when box is dilated
  // rfix[] = indices to each fix rigid

  delete [] rfix;
  nrigid = 0;
  rfix = NULL;

  for (int i = 0; i < modify->nfix; i++)
    if (strcmp(modify->fix[i]->style,"rigid") == 0) nrigid++;
  if (nrigid) {
    rfix = new int[nrigid];
    nrigid = 0;
    for (int i = 0; i < modify->nfix; i++)
      if (strcmp(modify->fix[i]->style,"rigid") == 0) rfix[nrigid++] = i;
  }
}

/* ----------------------------------------------------------------------
   compute temp and press before integrator starts 
------------------------------------------------------------------------- */

void FixNPT::setup()
{
  t_current = temperature->compute();
  pressure->compute(t_current);
  couple();
}

/* ----------------------------------------------------------------------
   1st half of Verlet update 
------------------------------------------------------------------------- */

void FixNPT::initial_integrate()
{
  int i;

  double delta = update->ntimestep - update->firststep;
  delta /= update->nsteps;

  t_target = t_start + delta * (t_stop-t_start);
  f_eta = t_freq*t_freq * (t_current/t_target - 1.0);
  eta_dot += f_eta*dthalf;
  eta += 2.0*dthalf*eta_dot;

  // update omega & omega_dot
  // for non-varying dims, p_freq is 0.0, so omega doesn't change
  // update omega by entire timestep since isn't used at end-of-step

  double p_target,f_omega,omega_half;
  double factor[3];

  double denskt = (atom->natoms*boltz*t_target) / 
    (domain->xprd*domain->yprd*domain->zprd) * nktv2p;

  for (i = 0; i < 3; i++) {
    p_target = p_start[i] + delta * (p_stop[i]-p_start[i]);
    f_omega = p_freq[i]*p_freq[i] * (p_current[i]-p_target)/denskt;
    omega_dot[i] += f_omega*dthalf;
    omega_half = dthalf*omega_dot[i];
    factor[i] = exp(-dthalf*(eta_dot+omega_dot[i]));
    dilation[i] = exp(omega_half);
    omega[i] += omega_half;
  }

  // velocity update only for atoms in NPT group

  double **x = atom->x;
  double **v = atom->v;
  double **f = atom->f;
  double *mass = atom->mass;
  int *type = atom->type;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  double dtfm;
  for (i = 0; i < nlocal; i++) {
    if (mask[i] & groupbit) {
      dtfm = dtf / mass[type[i]];
      v[i][0] = v[i][0]*factor[0] + dtfm*f[i][0];
      v[i][1] = v[i][1]*factor[1] + dtfm*f[i][1];
      v[i][2] = v[i][2]*factor[2] + dtfm*f[i][2];
    }
  }

  // rescale simulation box and all atoms

  box_dilate();

  // redo KSpace coeffs since volume has changed

  if (kspace_flag) force->kspace->setup();

  // position update only for atoms in NPT group

  for (i = 0; i < nlocal; i++) {
    if (mask[i] & groupbit) {
      x[i][0] += dtv * v[i][0];
      x[i][1] += dtv * v[i][1];
      x[i][2] += dtv * v[i][2];
    }
  }
}

/* ----------------------------------------------------------------------
   2nd half of Verlet update 
------------------------------------------------------------------------- */

void FixNPT::final_integrate()
{
  int i;

  double factor[3];
  for (i = 0; i < 3; i++)
    factor[i] = exp(-dthalf*(eta_dot+omega_dot[i]));

  // velocity update only for atoms in NPT group

  double **v = atom->v;
  double **f = atom->f;
  double *mass = atom->mass;
  int *type = atom->type;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  double dtfm;
  for (i = 0; i < nlocal; i++) {
    if (mask[i] & groupbit) {
      dtfm = dtf / mass[type[i]];
      v[i][0] = v[i][0]*factor[0] + dtfm*f[i][0]*factor[0];
      v[i][1] = v[i][1]*factor[1] + dtfm*f[i][1]*factor[1];
      v[i][2] = v[i][2]*factor[2] + dtfm*f[i][2]*factor[2];
    }
  }

  // rescale simulation box and all atoms

  box_dilate();

  // compute temperature

  t_current = temperature->compute();

  // update eta_dot

  double delta = update->ntimestep - update->firststep;
  delta /= update->nsteps;

  t_target = t_start + delta * (t_stop-t_start);
  f_eta = t_freq*t_freq * (t_current/t_target - 1.0);
  eta_dot += f_eta*dthalf;

  // compute pressure

  pressure->compute(t_current);
  couple();

  // update omega_dot

  double p_target,f_omega;

  double denskt = (atom->natoms*boltz*t_target) / 
    (domain->xprd*domain->yprd*domain->zprd) * nktv2p;

  for (i = 0; i < 3; i++) {
    p_target = p_start[i] + delta * (p_stop[i]-p_start[i]);
    f_omega = p_freq[i]*p_freq[i] * (p_current[i]-p_target)/denskt;
    omega_dot[i] += f_omega*dthalf;
  }
}

/* ---------------------------------------------------------------------- */

void FixNPT::initial_integrate_respa(int ilevel)
{
  double dtfm;

  // set timesteps by level

  dtv = step_respa[ilevel];
  dtf = 0.5 * step_respa[ilevel] * force->ftm2v;
  dthalf = 0.5 * step_respa[ilevel];

  // atom quantities

  double **x = atom->x;
  double **v = atom->v;
  double **f = atom->f;
  double *mass = atom->mass;
  int *type = atom->type;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;

  // outermost level - update eta and omega, apply to v, dilate box
  // all other levels - NVE update of v
  // x,v updates only performed for atoms in NPT group

  if (ilevel == nlevels_respa-1) {

    double delta = update->ntimestep - update->firststep;
    delta /= update->nsteps;

    t_target = t_start + delta * (t_stop-t_start);
    f_eta = t_freq*t_freq * (t_current/t_target - 1.0);
    eta_dot += f_eta*dthalf;
    eta += 2.0*dthalf*eta_dot;

    // update omega & omega_dot
    // for non-varying dims, p_freq is 0.0, so omega doesn't change
    // update omega by entire timestep since isn't used at end-of-step

    double p_target,f_omega,omega_half;
    double factor[3];

    double denskt = (atom->natoms*boltz*t_target) / 
      (domain->xprd*domain->yprd*domain->zprd) * nktv2p;

    for (int i = 0; i < 3; i++) {
      p_target = p_start[i] + delta * (p_stop[i]-p_start[i]);
      f_omega = p_freq[i]*p_freq[i] * (p_current[i]-p_target)/denskt;
      omega_dot[i] += f_omega*dthalf;
      omega_half = dthalf*omega_dot[i];
      factor[i] = exp(-dthalf*(eta_dot+omega_dot[i]));
      dilation[i] = exp(omega_half);
      omega[i] += omega_half;
    }

    for (int i = 0; i < nlocal; i++) {
      if (mask[i] & groupbit) {
	dtfm = dtf / mass[type[i]];
	v[i][0] = v[i][0]*factor[0] + dtfm*f[i][0];
	v[i][1] = v[i][1]*factor[1] + dtfm*f[i][1];
	v[i][2] = v[i][2]*factor[2] + dtfm*f[i][2];
      }
    }

    // rescale simulation box and all atoms

    box_dilate();

    // redo KSpace coeffs since volume has changed

    if (kspace_flag) force->kspace->setup();

  } else {
    for (int i = 0; i < nlocal; i++) {
      if (mask[i] & groupbit) {
	dtfm = dtf / mass[type[i]];
	v[i][0] += dtfm*f[i][0];
	v[i][1] += dtfm*f[i][1];
	v[i][2] += dtfm*f[i][2];
      }
    }
  }

  // innermost level - also update x

  if (ilevel == 0) {
    for (int i = 0; i < nlocal; i++) {
      if (mask[i] & groupbit) {
	x[i][0] += dtv * v[i][0];
	x[i][1] += dtv * v[i][1];
	x[i][2] += dtv * v[i][2];
      }
    }
  }
}

/* ---------------------------------------------------------------------- */

void FixNPT::final_integrate_respa(int ilevel)
{
  double dtfm;

  // set timesteps by level

  dtf = 0.5 * step_respa[ilevel] * force->ftm2v;
  dthalf = 0.5 * step_respa[ilevel];

  // outermost level - update eta and apply to v via final_integrate()
  // all other levels - NVE update of v
  // v update only performed for atoms in NPT group

  if (ilevel == nlevels_respa-1) final_integrate();
  else {
    double **v = atom->v;
    double **f = atom->f;
    double *mass = atom->mass;
    int *type = atom->type;
    int *mask = atom->mask;
    int nlocal = atom->nlocal;

    for (int i = 0; i < nlocal; i++) {
      if (mask[i] & groupbit) {
	dtfm = dtf / mass[type[i]];
	v[i][0] += dtfm*f[i][0];
	v[i][1] += dtfm*f[i][1];
	v[i][2] += dtfm*f[i][2];
      }
    }
  }
}

/* ---------------------------------------------------------------------- */

void FixNPT::couple()
{
  double *p_component = pressure->p_component;

  if (press_couple == 0)
    p_current[0] = p_current[1] = p_current[2] = pressure->p_total;
  else if (press_couple == 1) {
    double ave = 0.5 * (p_component[0] + p_component[1]);
    p_current[0] = p_current[1] = ave;
    p_current[2] = p_component[2];
  } else if (press_couple == 2) {
    double ave = 0.5 * (p_component[1] + p_component[2]);
    p_current[1] = p_current[2] = ave;
    p_current[0] = p_component[0];
  } else if (press_couple == 3) {
    double ave = 0.5 * (p_component[0] + p_component[2]);
    p_current[0] = p_current[2] = ave;
    p_current[1] = p_component[1];
  } if (press_couple == 4) {
    p_current[0] = p_component[0];
    p_current[1] = p_component[1];
    p_current[2] = p_component[2];
  }
}

/* ----------------------------------------------------------------------
   dilate the box after computing center-of-mass 
------------------------------------------------------------------------- */

void FixNPT::box_dilate()
{
  int i;

  double **x = atom->x;
  double *mass = atom->mass;
  int *type = atom->type;
  int *image = atom->image;
  int nlocal = atom->nlocal;

  // center-of-mass using true positions

  double cm[3],cmall[3];
  cm[0] = cm[1] = cm[2] = 0.0;
  int ix,iy,iz;
  double rmass;

  double xprd = domain->xprd;
  double yprd = domain->yprd;
  double zprd = domain->zprd;

  for (i = 0; i < nlocal; i++) {
    ix = image[i] % 1000 - 500;
    iy = (image[i]/1000) % 1000 - 500;
    iz = image[i]/1000000 - 500;
    rmass = mass[type[i]];
    cm[0] += rmass * (x[i][0] + ix*xprd);
    cm[1] += rmass * (x[i][1] + iy*yprd);
    cm[2] += rmass * (x[i][2] + iz*zprd);
  }

  MPI_Allreduce(cm,cmall,3,MPI_DOUBLE,MPI_SUM,world);
  cm[0] = cmall[0]/total_mass;
  cm[1] = cmall[1]/total_mass;
  cm[2] = cmall[2]/total_mass;

  // dilate the simulation box about the center of mass 
  // re-define simulation box via xprd/yprd/zprd
  // scale atom coords
  // if fix rigid exists, scale rigid body centers-of-mass
  // don't do anything if non-periodic or press style is constant volume

  double oldlo,oldhi;

  if (domain->xperiodic && p_flag[0]) {
    oldlo = domain->boxxlo;
    oldhi = domain->boxxhi;
    domain->boxxlo = (oldlo-cm[0])*dilation[0] + cm[0];
    domain->boxxhi = (oldhi-cm[0])*dilation[0] + cm[0];
    domain->xprd = domain->boxxhi - domain->boxxlo;
    for (i = 0; i < nlocal; i++)
      x[i][0] = cm[0] + (x[i][0]-cm[0])*dilation[0];
    if (nrigid)
      for (i = 0; i < nrigid; i++)
	((FixRigid *) modify->fix[rfix[i]])->
	  dilate(0,oldlo,oldhi,domain->boxxlo,domain->boxxhi);
  }

  if (domain->yperiodic && p_flag[1]) {
    oldlo = domain->boxylo;
    oldhi = domain->boxyhi;
    domain->boxylo = (oldlo-cm[1])*dilation[1] + cm[1];
    domain->boxyhi = (oldhi-cm[1])*dilation[1] + cm[1];
    domain->yprd = domain->boxyhi - domain->boxylo;
    for (i = 0; i < nlocal; i++)
      x[i][1] = cm[1] + (x[i][1]-cm[1])*dilation[1];
    if (nrigid)
      for (i = 0; i < nrigid; i++)
	((FixRigid *) modify->fix[rfix[i]])->
	  dilate(1,oldlo,oldhi,domain->boxylo,domain->boxyhi);
  }

  if (domain->zperiodic && p_flag[2]) {
    oldlo = domain->boxzlo;
    oldhi = domain->boxzhi;
    domain->boxzlo = (oldlo-cm[2])*dilation[2] + cm[2];
    domain->boxzhi = (oldhi-cm[2])*dilation[2] + cm[2];
    domain->zprd = domain->boxzhi - domain->boxzlo;
    for (i = 0; i < nlocal; i++)
      x[i][2] = cm[2] + (x[i][2]-cm[2])*dilation[2];
    if (nrigid)
      for (i = 0; i < nrigid; i++)
	((FixRigid *) modify->fix[rfix[i]])->
	  dilate(2,oldlo,oldhi,domain->boxzlo,domain->boxzhi);
  }
}

/* ----------------------------------------------------------------------
   pack entire state of Fix into one write 
------------------------------------------------------------------------- */

void FixNPT::write_restart(FILE *fp)
{
  int n = 0;
  double list[8];
  list[n++] = eta;
  list[n++] = eta_dot;
  list[n++] = omega[0];
  list[n++] = omega[1];
  list[n++] = omega[2];
  list[n++] = omega_dot[0];
  list[n++] = omega_dot[1];
  list[n++] = omega_dot[2];

  if (comm->me == 0) {
    int size = n * sizeof(double);
    fwrite(&size,sizeof(int),1,fp);
    fwrite(&list,sizeof(double),n,fp);
  }
}

/* ----------------------------------------------------------------------
   use state info from restart file to restart the Fix 
------------------------------------------------------------------------- */

void FixNPT::restart(char *buf)
{
  int n = 0;
  double *list = (double *) buf;
  eta = list[n++];
  eta_dot = list[n++];
  omega[0] = list[n++];
  omega[1] = list[n++];
  omega[2] = list[n++];
  omega_dot[0] = list[n++];
  omega_dot[1] = list[n++];
  omega_dot[2] = list[n++];
}

/* ---------------------------------------------------------------------- */

void FixNPT::set_temp_style(int itemp)
{
  temp_chosen = itemp;
}
