/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "string.h"
#include "fix_spring.h"
#include "atom.h"
#include "update.h"
#include "respa.h"
#include "domain.h"
#include "error.h"

/* ---------------------------------------------------------------------- */

FixSpring::FixSpring(int narg, char **arg) : Fix(narg, arg)
{
  if (narg != 7) error->all("Illegal fix spring command");

  xflag = yflag = zflag = 1;

  if (strcmp(arg[3],"NULL") == 0) xflag = 0;
  else xc = atof(arg[3]);
  if (strcmp(arg[4],"NULL") == 0) yflag = 0;
  else yc = atof(arg[4]);
  if (strcmp(arg[5],"NULL") == 0) zflag = 0;
  else zc = atof(arg[5]);

  k_spring = atof(arg[6]);
}

/* ---------------------------------------------------------------------- */

int FixSpring::setmask()
{
  int mask = 0;
  mask |= POST_FORCE;
  mask |= POST_FORCE_RESPA;
  return mask;
}

/* ---------------------------------------------------------------------- */

void FixSpring::init()
{
  // masstotal = total mass of atoms in group

  int *mask = atom->mask;
  int *type = atom->type;
  double *mass = atom->mass;
  int nlocal = atom->nlocal;

  double rmass = 0.0;
  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) rmass += mass[type[i]];
  MPI_Allreduce(&rmass,&masstotal,1,MPI_DOUBLE,MPI_SUM,world);

  if (strcmp(update->integrate_style,"respa") == 0)
    nlevels_respa = ((Respa *) update->integrate)->nlevels;
}

/* ---------------------------------------------------------------------- */

void FixSpring::setup()
{
  if (strcmp(update->integrate_style,"verlet") == 0)
    post_force(1);
  else {
    ((Respa *) update->integrate)->copy_flevel_f(nlevels_respa-1);
    post_force_respa(1,nlevels_respa-1,0);
    ((Respa *) update->integrate)->copy_f_flevel(nlevels_respa-1);
  }
}

/* ---------------------------------------------------------------------- */

void FixSpring::post_force(int vflag)
{
  double **x = atom->x;
  double **f = atom->f;
  int *mask = atom->mask;
  int *type = atom->type;
  int *image = atom->image;
  double *mass = atom->mass;
  int nlocal = atom->nlocal;

  // compute center-of-mass of group
  // use unwrapped coordinates via image
  //   else answer is bogus if group straddles a periodic boundary

  double com[3];
  com[0] = com[1] = com[2] = 0.0;
  double xprd = domain->xprd;
  double yprd = domain->yprd;
  double zprd = domain->zprd;

  int xbox,ybox,zbox;
  double massone;
  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      xbox = image[i]%1000 - 500;
      ybox = (image[i]/1000) % 1000 - 500;
      zbox = image[i]/1000000 - 500;
      massone = mass[type[i]];
      com[0] += (x[i][0] + xbox*xprd) * massone;
      com[1] += (x[i][1] + ybox*yprd) * massone;
      com[2] += (x[i][2] + zbox*zprd) * massone;
    }

  double com_all[3];
  MPI_Allreduce(com,com_all,3,MPI_DOUBLE,MPI_SUM,world);
  com_all[0] /= masstotal;
  com_all[1] /= masstotal;
  com_all[2] /= masstotal;

  // fx,fy,fz = components of k * (r-r0)

  double dx,dy,dz,fx,fy,fz;

  dx = com_all[0] - xc;
  dy = com_all[1] - yc;
  dz = com_all[2] - zc;
  if (!xflag) dx = 0.0;
  if (!yflag) dy = 0.0;
  if (!zflag) dz = 0.0;
  fx = k_spring * dx;
  fy = k_spring * dy;
  fz = k_spring * dz;

  // apply restoring force to atoms in group
  // f = -k*(r-r0)*mass/masstotal

  double massfrac;
  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      massfrac = mass[type[i]]/masstotal;
      f[i][0] -= fx*massfrac;
      f[i][1] -= fy*massfrac;
      f[i][2] -= fz*massfrac;
    }
}

/* ---------------------------------------------------------------------- */

void FixSpring::post_force_respa(int vflag, int ilevel, int iloop)
{
  if (ilevel == nlevels_respa-1) post_force(vflag);
}
