/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "string.h"
#include "stdlib.h"
#include "math.h"
#include "fix_temp_rescale.h"
#include "atom.h"
#include "force.h"
#include "update.h"
#include "temperature.h"
#include "temp_full.h"
#include "error.h"

/* ---------------------------------------------------------------------- */

FixTempRescale::FixTempRescale(int narg, char **arg) : Fix(narg, arg)
{
  if (narg != 8) error->all("Illegal fix temp/rescale command");
  t_start = atof(arg[3]);
  t_end = atof(arg[4]);
  nevery = atoi(arg[5]);
  t_window = atof(arg[6]);
  fraction = atof(arg[7]);

  if (nevery <= 0) error->all("Illegal fix temp/rescale command");

  // if group is "all", use default temperature
  // else create a temperature with style "full" for the correct group
  //   name of new temperature = "FixTempRescale" + fix-id + group-name
  //   if name already exists, just point to it

  if (igroup == 0) temp_chosen = 0;
  else {
    char **targ = new char*[3];
    targ[0] =
      new char[strlen("FixTempRescale") + strlen(id) + strlen(arg[1]) + 1];
    sprintf(targ[0],"%s%s%s","FixTempRescale",id,arg[1]);
    int itemp;
    for (itemp = 0; itemp < force->ntemp; itemp++)
      if (strcmp(targ[0],force->templist[itemp]->id) == 0) break;
    if (itemp < force->ntemp) temp_chosen = itemp;
    else {
      targ[1] = arg[1];
      targ[2] = "full";
      force->add_temp(3,targ);
      temp_chosen = force->ntemp - 1;
    }
    delete [] targ[0];
    delete [] targ;
  }
}

/* ---------------------------------------------------------------------- */

int FixTempRescale::setmask()
{
  int mask = 0;
  mask |= END_OF_STEP;
  return mask;
}

/* ---------------------------------------------------------------------- */

void FixTempRescale::init()
{
  temperature = force->templist[temp_chosen];
}

/* ---------------------------------------------------------------------- */

void FixTempRescale::end_of_step()
{
  if (update->ntimestep % nevery) return;

  double t_current = temperature->compute();
  double delta = update->ntimestep - update->firststep;
  delta /= update->nsteps;
  double t_target = t_start + delta * (t_end-t_start);

  if (fabs(t_current-t_target) > t_window) {
    t_target = t_current - fraction*(t_current-t_target);
    double factor = sqrt(t_target/t_current);
    double **v = atom->v;
    int *mask = atom->mask;
    int nlocal = atom->nlocal;
    for (int i = 0; i < nlocal; i++) {
      if (mask[i] & groupbit) {
	v[i][0] *= factor;
	v[i][1] *= factor;
	v[i][2] *= factor;
      }
    }
  }
}

/* ---------------------------------------------------------------------- */

void FixTempRescale::set_temp_style(int itemp)
{
  temp_chosen = itemp;
}
