/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "mpi.h"
#include "stdio.h"
#include "string.h"
#include "stdlib.h"
#include "group.h"
#include "domain.h"
#include "region.h"
#include "memory.h"
#include "error.h"
#include "atom.h"

#define MAX_GROUP 32

#define TYPE      1
#define MOLECULE  2
#define ID        3

#define LESS_THAN      1
#define GREATER_THAN   2
#define LESS_EQUAL     3
#define GREATER_EQUAL  4
#define BETWEEN        5

#define BIG 1.0e20

/* ----------------------------------------------------------------------
   initialize group memory
------------------------------------------------------------------------- */

Group::Group()
{
  MPI_Comm_rank(world,&me);

  names = (char **) memory->smalloc(MAX_GROUP*sizeof(char *),"group:names");
  bitmask = new int[MAX_GROUP];
  inversemask = new int[MAX_GROUP];
  for (int i = 0; i < MAX_GROUP; i++) bitmask[i] = 1 << i;
  for (int i = 0; i < MAX_GROUP; i++) inversemask[i] = bitmask[i] ^ ~0;

  // create "all" group

  char *str = "all";
  int n = strlen(str) + 1;
  names[0] = (char *) memory->smalloc(n*sizeof(char),"group:names[]");
  strcpy(names[0],str);
  ngroup = 1;
}

/* ----------------------------------------------------------------------
   free all memory
------------------------------------------------------------------------- */

Group::~Group()
{
  for (int i = 0; i < ngroup; i++) memory->sfree(names[i]);
  memory->sfree(names);
  delete [] bitmask;
  delete [] inversemask;
}

/* ----------------------------------------------------------------------
   assign atoms to a new or existing group
------------------------------------------------------------------------- */

void Group::assign(int narg, char **arg)
{
  int i;

  if (domain->box_exist == 0) 
    error->all("Group command before simulation box is defined");
  if (narg < 2) error->all("Illegal group command");

  // find group in existing list
  // igroup = -1 is a new group name, add it

  int igroup = find(arg[0]);

  if (igroup == -1) {
    if (ngroup == MAX_GROUP) error->all("Too many groups");
    igroup = ngroup;
    ngroup++;
    int n = strlen(arg[0]) + 1;
    names[igroup] = (char *) memory->smalloc(n*sizeof(char),"group:names[]");
    strcpy(names[igroup],arg[0]);
  }

  double **x = atom->x;
  int *mask = atom->mask;
  int nlocal = atom->nlocal;
  int bit = bitmask[igroup];
    
  // style = region

  if (strcmp(arg[1],"region") == 0) {

    if (narg < 3) error->all("Illegal group command");
    
    int length = narg-2;
    int *list = new int[length];

    int iregion;
    for (int iarg = 2; iarg < narg; iarg++) {
      for (iregion = 0; iregion < domain->nregion; iregion++)
	if (strcmp(arg[iarg],domain->regions[iregion]->id) == 0) break;
      if (iregion == domain->nregion)
	error->all("Group region ID does not exist");
      list[iarg-2] = iregion;
    }

    // add to group if atom is in all regions in list

    int ilist;
    for (i = 0; i < nlocal; i++) {
      for (ilist = 0; ilist < length; ilist++)
	if (!domain->regions[list[ilist]]->match(x[i][0],x[i][1],x[i][2]))
	  break;
      if (ilist == length) mask[i] |= bit;
    }

    delete [] list;

    // style = logical condition

  } else if (strcmp(arg[2],"<") == 0 || strcmp(arg[2],">") == 0 || 
	     strcmp(arg[2],"<=") == 0 || strcmp(arg[2],">=") == 0 || 
	     strcmp(arg[2],"<>") == 0) {

    if (narg < 4 || narg > 5) error->all("Illegal group command");
    int category,condition,bound1,bound2;

    if (strcmp(arg[1],"type") == 0) category = TYPE;
    else if (strcmp(arg[1],"molecule") == 0) category = MOLECULE;
    else if (strcmp(arg[1],"id") == 0) category = ID;
    else error->all("Illegal group command");
    
    if (strcmp(arg[2],"<") == 0) condition = LESS_THAN;
    else if (strcmp(arg[2],">") == 0) condition = GREATER_THAN;
    else if (strcmp(arg[2],"<=") == 0) condition = LESS_EQUAL;
    else if (strcmp(arg[2],">=") == 0) condition = GREATER_EQUAL;
    else if (strcmp(arg[2],"<>") == 0) condition = BETWEEN;
    else error->all("Illegal group command");

    bound1 = atoi(arg[3]);
    bound2 = -1;

    if (condition == BETWEEN) {
      if (narg != 5) error->all("Illegal group command");
      bound2 = atoi(arg[4]);
    }

    int *attribute;
    if (category == TYPE) attribute = atom->type;
    else if (category == MOLECULE) attribute = atom->molecule;
    else if (category == ID) attribute = atom->tag;

    // add to group if meets condition

    if (condition == LESS_THAN) {
      for (i = 0; i < nlocal; i++) if (attribute[i] < bound1) mask[i] |= bit;
    } else if (condition == GREATER_THAN) {
      for (i = 0; i < nlocal; i++) if (attribute[i] > bound1) mask[i] |= bit;
    } else if (condition == LESS_EQUAL) {
      for (i = 0; i < nlocal; i++) if (attribute[i] <= bound1) mask[i] |= bit;
    } else if (condition == GREATER_EQUAL) {
      for (i = 0; i < nlocal; i++) if (attribute[i] >= bound1) mask[i] |= bit;
    } else if (condition == BETWEEN) {
      for (i = 0; i < nlocal; i++)
	if (attribute[i] >= bound1 && attribute[i] <= bound2) mask[i] |= bit;
    }

    // style = list of values

  } else if (strcmp(arg[1],"type") == 0 || strcmp(arg[1],"molecule") == 0 ||
	     strcmp(arg[1],"id") == 0) {

    if (narg < 3) error->all("Illegal group command");

    int length = narg-2;
    int *list = new int[length];

    int category;
    if (strcmp(arg[1],"type") == 0) category = TYPE;
    else if (strcmp(arg[1],"molecule") == 0) category = MOLECULE;
    else if (strcmp(arg[1],"id") == 0) category = ID;
    else error->all("Illegal group command");
    
    length = narg - 2;
    for (int iarg = 2; iarg < narg; iarg++) list[iarg-2] = atoi(arg[iarg]);

    int *attribute;
    if (category == TYPE) attribute = atom->type;
    else if (category == MOLECULE) attribute = atom->molecule;
    else if (category == ID) attribute = atom->tag;

    // add to group if attribute is any in list

    for (int ilist = 0; ilist < length; ilist++)
      for (i = 0; i < nlocal; i++)
	if (attribute[i] == list[ilist]) mask[i] |= bit;

    delete [] list;
    
    // style = join

  } else if (strcmp(arg[1],"join") == 0) {

    if (narg < 3) error->all("Illegal group command");
    
    int length = narg-2;
    int *list = new int[length];

    int jgroup;
    for (int iarg = 2; iarg < narg; iarg++) {
      jgroup = find(arg[iarg]);
      if (jgroup == -1) error->all("Group ID does not exist");
      list[iarg-2] = jgroup;
    }

    // add to group if in any other group in list

    int otherbit;

    for (int ilist = 0; ilist < length; ilist++) {
      otherbit = bitmask[list[ilist]];
      for (i = 0; i < nlocal; i++) 
	if (mask[i] & otherbit) mask[i] |= bit;
    }

    delete [] list;

    // style = subtract

  } else if (strcmp(arg[1],"subtract") == 0) {

    if (narg < 4) error->all("Illegal group command");
    
    int length = narg-2;
    int *list = new int[length];

    int jgroup;
    for (int iarg = 2; iarg < narg; iarg++) {
      jgroup = find(arg[iarg]);
      if (jgroup == -1) error->all("Group ID does not exist");
      list[iarg-2] = jgroup;
    }

    // add to group if in 1st group in list

    int otherbit = bitmask[list[0]];

    for (i = 0; i < nlocal; i++) 
      if (mask[i] & otherbit) mask[i] |= bit;

    // remove atoms if they are in any of the other groups
    // AND with inverse mask removes the atom from group

    int inverse = inversemask[igroup];

    for (int ilist = 1; ilist < length; ilist++) {
      otherbit = bitmask[list[ilist]];
      for (i = 0; i < nlocal; i++) 
	if (mask[i] & otherbit) mask[i] &= inverse;
    }

    delete [] list;

    // style = intersect

  } else if (strcmp(arg[1],"intersect") == 0) {

    if (narg < 4) error->all("Illegal group command");
    
    int length = narg-2;
    int *list = new int[length];

    int jgroup;
    for (int iarg = 2; iarg < narg; iarg++) {
      jgroup = find(arg[iarg]);
      if (jgroup == -1) error->all("Group ID does not exist");
      list[iarg-2] = jgroup;
    }

    // add to group if in all groups in list

    int otherbit,ok,ilist;

    for (i = 0; i < nlocal; i++) {
      ok = 1;
      for (ilist = 0; ilist < length; ilist++) {
	otherbit = bitmask[list[ilist]];
	if ((mask[i] & otherbit) == 0) ok = 0;
      }
      if (ok) mask[i] |= bit;
    }

    delete [] list;

    // not a valid group style

  } else error->all("Illegal group command");

  int n,all;
  n = 0;
  for (i = 0; i < nlocal; i++) if (mask[i] & bit) n++;
  MPI_Allreduce(&n,&all,1,MPI_INT,MPI_SUM,world);

  if (me == 0) {
    if (screen) fprintf(screen,"%d atoms in group %s\n",all,names[igroup]);
    if (logfile) fprintf(logfile,"%d atoms in group %s\n",all,names[igroup]);
  }
}

/* ----------------------------------------------------------------------
   return group index if name matches existing group, -1 if no such group
------------------------------------------------------------------------- */

int Group::find(char *name)
{
  for (int igroup = 0; igroup < ngroup; igroup++)
    if (strcmp(name,names[igroup]) == 0) return igroup;
  return -1;
}

/* ----------------------------------------------------------------------
   write group info to a restart file
   only called by proc 0
------------------------------------------------------------------------- */

void Group::write_restart(FILE *fp)
{
  fwrite(&ngroup,sizeof(int),1,fp);

  int n;
  for (int i = 0; i < ngroup; i++) {
    n = strlen(names[i]) + 1;
    fwrite(&n,sizeof(int),1,fp);
    fwrite(names[i],sizeof(char),n,fp);
  }
}

/* ----------------------------------------------------------------------
   read group info from a restart file
   proc 0 reads, bcast to all procs
------------------------------------------------------------------------- */

void Group::read_restart(FILE *fp)
{
  int i,n;

  for (i = 0; i < ngroup; i++) memory->sfree(names[i]);

  if (me == 0) fread(&ngroup,sizeof(int),1,fp);
  MPI_Bcast(&ngroup,1,MPI_INT,0,world);
  
  for (i = 0; i < ngroup; i++) {
    if (me == 0) fread(&n,sizeof(int),1,fp);
    MPI_Bcast(&n,1,MPI_INT,0,world);
    names[i] = (char *) memory->smalloc(n*sizeof(char),"group:names[]");
    if (me == 0) fread(names[i],sizeof(char),n,fp);
    MPI_Bcast(names[i],n,MPI_CHAR,0,world);
  }
}
