/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "mpi.h"
#include "math.h"
#include "stdlib.h"
#include "string.h"
#include "neighbor.h"
#include "atom.h"
#include "force.h"
#include "pair.h"
#include "domain.h"
#include "group.h"
#include "modify.h"
#include "fix.h"
#include "update.h"
#include "respa.h"
#include "output.h"
#include "memory.h"
#include "error.h"

#define PGDELTA 1
#define LB_FACTOR 1.5
#define SMALL 1.0e-6
#define EXDELTA 1

#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

/* ---------------------------------------------------------------------- */

Neighbor::Neighbor()
{
  MPI_Comm_rank(world,&me);
  MPI_Comm_size(world,&nprocs);

  style = 1;
  every = 1;
  delay = 10;
  dist_check = 1;
  pgsize = 10000;
  oneatom = 2000;
    
  cutneighsq = NULL;

  max_totalneigh = 0;

  maxlocal = 0;
  numneigh = NULL;
  firstneigh = NULL;

  maxpage = 0;
  pages = NULL;

  maxhead = 0;
  binhead = NULL;

  maxbin = 0;
  bins = NULL;

  maxstencil = 0;
  stencil = NULL;

  maxhold = 0;
  xhold = NULL;

  // pair exclusion list info

  nex_type = maxex_type = 0;
  ex1_type = ex2_type = NULL;
  ex_type = NULL;

  nex_group = maxex_group = 0;
  ex1_group = ex2_group = ex1_bit = ex2_bit = NULL;

  nex_mol = maxex_mol = 0;
  ex_mol_group = ex_mol_bit = NULL;

  // shear history neighbor list info

  history = -1;
  firsttouch = NULL;
  firstshear = NULL;
  pages_touch = NULL;
  pages_shear = NULL;

  // multiple respa neighbor list info

  respa = 0;
  numneigh_inner = NULL;
  firstneigh_inner = NULL;
  maxpage_inner = 0;
  pages_inner = NULL;
  numneigh_middle = NULL;
  firstneigh_middle = NULL;
  maxpage_middle = 0;
  pages_middle = NULL;

  // full neighbor list info

  full = 0;
  maxfull = 0;
  numneigh_full = NULL;
  firstneigh_full = NULL;
  maxpage_full = 0;
  pages_full = NULL;

  maxstencil_full = 0;
  stencil_full = NULL;

  // bond topology list info

  bondmax = 0;
  bondlist = NULL;
  anglemax = 0;
  anglelist = NULL;
  dihedralmax = 0;
  dihedrallist = NULL;
  impropermax = 0;
  improperlist = NULL;
}

/* ---------------------------------------------------------------------- */

Neighbor::~Neighbor()
{
  memory->destroy_2d_double_array(cutneighsq);

  memory->sfree(numneigh);
  memory->sfree(firstneigh);
  for (int i = 0; i < maxpage; i++) memory->sfree(pages[i]);
  memory->sfree(pages);

  memory->sfree(binhead);
  memory->sfree(bins);
  memory->sfree(stencil);
  memory->destroy_2d_double_array(xhold);

  memory->sfree(ex1_type);
  memory->sfree(ex2_type);
  memory->destroy_2d_int_array(ex_type);

  memory->sfree(ex1_group);
  memory->sfree(ex2_group);
  delete [] ex1_bit;
  delete [] ex2_bit;

  memory->sfree(ex_mol_group);
  delete [] ex_mol_bit;

  memory->destroy_2d_int_array(bondlist);
  memory->destroy_2d_int_array(anglelist);
  memory->destroy_2d_int_array(dihedrallist);
  memory->destroy_2d_int_array(improperlist);

  if (history >= 0) {
    memory->sfree(firsttouch);
    memory->sfree(firstshear);
    for (int i = 0; i < maxpage; i++) memory->sfree(pages_touch[i]);
    for (int i = 0; i < maxpage; i++) memory->sfree(pages_shear[i]);
    memory->sfree(pages_touch);
    memory->sfree(pages_shear);
  }

  if (respa) {
    memory->sfree(numneigh_inner);
    memory->sfree(firstneigh_inner);
    for (int i = 0; i < maxpage_inner; i++) memory->sfree(pages_inner[i]);
    memory->sfree(pages_inner);

    memory->sfree(numneigh_middle);
    memory->sfree(firstneigh_middle);
    for (int i = 0; i < maxpage_middle; i++) memory->sfree(pages_middle[i]);
    memory->sfree(pages_middle);
  }

  if (full == 2) {
    memory->sfree(numneigh_full);
    memory->sfree(firstneigh_full);
    for (int i = 0; i < maxpage_full; i++) memory->sfree(pages_full[i]);
    memory->sfree(pages_full);
    memory->sfree(stencil_full);
  }
}

/* ---------------------------------------------------------------------- */

void Neighbor::init()
{
  int i,j,m,n;

  ncalls = 0;
  ndanger = 0;

  if (force->pair) cutneigh = force->pair->cutforce + skin;
  else cutneigh = skin;
  triggersq = 0.25*skin*skin;

  // which other classes must be checked for reneighboring in decide()

  must_check = 0;
  restart_check = 0;
  fix_check = 0;

  if (output->restart_every) restart_check = 1;

  for (i = 0; i < modify->nfix; i++)
    if (strcmp(modify->fix[i]->style,"insert") == 0) {
      fix_check = 1;
      fixwhich = i;
    }

  if (restart_check || fix_check) must_check = 1;
  
  // memory allocation for standard pair list and molecular lists
  // if 1st run, these arrays have previously not been allocated
  // if neighbor parameters have changed, free unneeded arrays

  if (cutneighsq == NULL)
    cutneighsq = memory->create_2d_double_array(atom->ntypes+1,atom->ntypes+1,
						"neighbor:cutneighsq");

  if (atom->molecular && atom->nbonds && bondmax == 0) {
    if (nprocs == 1) bondmax = atom->nbonds;
    else bondmax = static_cast<int> (LB_FACTOR * atom->nbonds / nprocs);
    bondlist = memory->create_2d_int_array(bondmax,3,"neighbor:bondlist");
  }

  if (atom->molecular && atom->nangles && anglemax == 0) {
    if (nprocs == 1) anglemax = atom->nangles;
    else anglemax = static_cast<int> (LB_FACTOR * atom->nangles / nprocs);
    anglelist =  memory->create_2d_int_array(anglemax,4,"neighbor:anglelist");
  }

  if (atom->molecular && atom->ndihedrals && dihedralmax == 0) {
    if (nprocs == 1) dihedralmax = atom->ndihedrals;
    else dihedralmax = static_cast<int> 
	   (LB_FACTOR * atom->ndihedrals / nprocs);
    dihedrallist = 
      memory->create_2d_int_array(dihedralmax,5,"neighbor:dihedrallist");
  }

  if (atom->molecular && atom->nimpropers && impropermax == 0) {
    if (nprocs == 1) impropermax = atom->nimpropers;
    else impropermax = static_cast<int>
	   (LB_FACTOR * atom->nimpropers / nprocs);
    improperlist = 
      memory->create_2d_int_array(impropermax,5,"neighbor:improperlist");
  }

  if (maxlocal == 0) {
    maxlocal = atom->nlocal;
    numneigh = (int *) memory->smalloc(maxlocal*sizeof(int),"numneigh");
    firstneigh = (int **) memory->smalloc(maxlocal*sizeof(int *),"firstneigh");
    add_pages(0);
  }

  if (style == 0) {
    memory->sfree(bins);
    memory->sfree(binhead);
    memory->sfree(stencil);
    memory->sfree(stencil_full);
    maxbin = maxhead = maxstencil = maxstencil_full = 0;
  }

  if (style == 1) {
    if (maxbin == 0) {
      maxbin = atom->nmax;
      bins = (int *) memory->smalloc(maxbin*sizeof(int),"bins");
    }
  }

  if (dist_check == 0) {
    memory->destroy_2d_double_array(xhold);
    maxhold = 0;
  }
    
  if (dist_check) {
    if (maxhold == 0) {
      maxhold = atom->nlocal;
      xhold = memory->create_2d_double_array(maxhold,3,"neighbor:xhold");
    }
  }

  // exclusion lists for type, group, molecule settings from neigh_modify

  n = atom->ntypes;

  if (nex_type == 0 && nex_group == 0 && nex_mol == 0) exclude = 0;
  else exclude = 1;

  if (nex_type) {
    memory->destroy_2d_int_array(ex_type);
    ex_type = (int **) memory->create_2d_int_array(n+1,n+1,"neighbor:ex_type");

    for (i = 1; i <= n; i++)
      for (j = 1; j <= n; j++)
	ex_type[i][j] = 0;

    for (i = 0; i < nex_type; i++) {
      if (ex1_type[i] <= 0 || ex1_type[i] > n || 
	  ex2_type[i] <= 0 || ex2_type[i] > n)
	error->all("Illegal atom type in neighbor exclusion list");
      ex_type[ex1_type[i]][ex2_type[i]] = 1;
      ex_type[ex2_type[i]][ex1_type[i]] = 1;
    }
  }

  if (nex_group) {
    delete [] ex1_bit;
    delete [] ex2_bit;
    ex1_bit = new int[nex_group];
    ex2_bit = new int[nex_group];

    for (i = 0; i < nex_group; i++) {
      ex1_bit[i] = group->bitmask[ex1_group[i]];
      ex2_bit[i] = group->bitmask[ex2_group[i]];
    }
  }

  if (nex_mol) {
    delete [] ex_mol_bit;
    ex_mol_bit = new int[nex_mol];

    for (i = 0; i < nex_mol; i++)
      ex_mol_bit[i] = group->bitmask[ex_mol_group[i]];
  }

  // set neighbor cutoffs with skin included

  if (force->pair) {
    double cutoff;
    double **cutsq = force->pair->cutsq;
    for (i = 1; i <= n; i++)
      for (j = i; j <= n; j++) {
	cutoff = sqrt(cutsq[i][j]);
	cutneighsq[i][j] = (cutoff+skin) * (cutoff+skin);
	cutneighsq[j][i] = cutneighsq[i][j];
      }
  } else {
    for (i = 1; i <= n; i++)
      for (j = i; j <= n; j++) {
	cutneighsq[i][j] = skin*skin;
	cutneighsq[j][i] = cutneighsq[i][j];
      }
  }

  // set special_flag for 1-2, 1-3, 1-4 neighbors
  // flag[0] is not used, flag[1] = 1-2, flag[2] = 1-3, flag[3] = 1-4
  // flag = 0 if both LJ/Coulomb special values are 0.0
  // flag = 1 if both LJ/Coulomb special values are 1.0
  // flag = 2 otherwise or if KSpace solver is enabled
  // (pairwise portion of KSpace solver uses all 1-2,1-3,1-4 neighbors)

  if (force->special_lj[1] == 0.0 && force->special_coul[1] == 0.0) 
    special_flag[1] = 0;
  else if (force->special_lj[1] == 1.0 && force->special_coul[1] == 1.0) 
    special_flag[1] = 1;
  else special_flag[1] = 2;

  if (force->special_lj[2] == 0.0 && force->special_coul[2] == 0.0) 
    special_flag[2] = 0;
  else if (force->special_lj[2] == 1.0 && force->special_coul[2] == 1.0) 
    special_flag[2] = 1;
  else special_flag[2] = 2;

  if (force->special_lj[3] == 0.0 && force->special_coul[3] == 0.0) 
    special_flag[3] = 0;
  else if (force->special_lj[3] == 1.0 && force->special_coul[3] == 1.0) 
    special_flag[3] = 1;
  else special_flag[3] = 2;

  if (force->kspace) special_flag[1] = special_flag[2] = special_flag[3] = 2;

  // set bond_off, etc to determine which neighboring routines to use
  // SHAKE sets bonds and angles negative
  // bond_quartic_harmoic sets bonds to 0
  // delete_bonds sets all interactions negative

  bond_off = angle_off = dihedral_off = improper_off = 0;

  for (i = 0; i < modify->nfix; i++)
    if (strcmp(modify->fix[i]->style,"shake") == 0)
      bond_off = angle_off = 1;
  if (force->bond && strcmp(force->bond_style,"quartic/harmonic") == 0)
    bond_off = 1;

  if (atom->bonds_allow) {
    for (i = 0; i < atom->nlocal; i++) {
      if (bond_off) break;
      for (m = 0; m < atom->num_bond[i]; m++)
	if (atom->bond_type[i][m] <= 0) bond_off = 1;
    }
  }
  if (atom->angles_allow) {
    for (i = 0; i < atom->nlocal; i++) {
      if (angle_off) break;
      for (m = 0; m < atom->num_angle[i]; m++)
	if (atom->angle_type[i][m] <= 0) angle_off = 1;
    }
  }
  if (atom->dihedrals_allow) {
    for (i = 0; i < atom->nlocal; i++) {
      if (dihedral_off) break;
      for (m = 0; m < atom->num_dihedral[i]; m++)
	if (atom->dihedral_type[i][m] <= 0) dihedral_off = 1;
    }
  }
  if (atom->impropers_allow) {
    for (i = 0; i < atom->nlocal; i++) {
      if (improper_off) break;
      for (m = 0; m < atom->num_improper[i]; m++)
	if (atom->improper_type[i][m] <= 0) improper_off = 1;
    }
  }

  // history = index of granular shear history fix if it exists
  // setup/delete memory for shear history neighbor lists as appropriate
  
  int history_previous = history;
  history = -1;
  if (force->pair_match("gran/history") || force->pair_match("gran/hertzian"))
    for (i = 0; i < modify->nfix; i++)
      if (strcmp(modify->fix[i]->style,"SHEAR_HISTORY") == 0) history = i;

  if (history == -1 && history_previous >= 0) {
    memory->sfree(firsttouch);
    memory->sfree(firstshear);
    for (int i = 0; i < maxpage; i++) memory->sfree(pages_touch[i]);
    for (int i = 0; i < maxpage; i++) memory->sfree(pages_shear[i]);
    memory->sfree(pages_touch);
    memory->sfree(pages_shear);
    pages_touch = NULL;
    pages_shear = NULL;
  } else if (history >= 0 && history_previous == -1) {
    firsttouch = (int **) memory->smalloc(maxlocal*sizeof(int *),"firsttouch");
    firstshear = (double **)
      memory->smalloc(maxlocal*sizeof(double *),"firstshear");
    add_pages_history(0);
  }

  // check for rRESPA requiring multiple neighbor lists (respa > 0)
  // setup/delete memory for multiple neighbor lists as appropriate
  // set neighbor cutoffs for multiple lists

  int respa_previous = respa;
  respa = 0;
  if (strcmp(update->integrate_style,"respa") == 0) {
    if (((Respa *) update->integrate)->level_inner >= 0) respa = 1;
    if (((Respa *) update->integrate)->level_middle >= 0) respa = 2;
  }

  if (respa == 0 && respa_previous) {
    memory->sfree(numneigh_inner);
    memory->sfree(firstneigh_inner);
    for (int i = 0; i < maxpage; i++) memory->sfree(pages_inner[i]);
    memory->sfree(pages_inner);
    pages_inner = NULL;
    maxpage_inner = 0;
    if (respa_previous == 2) {
      memory->sfree(numneigh_middle);
      memory->sfree(firstneigh_middle);
      for (int i = 0; i < maxpage; i++) memory->sfree(pages_middle[i]);
      memory->sfree(pages_middle);
      pages_middle = NULL;
      maxpage_middle = 0;
    }
  } else if (respa && respa_previous == 0) {
    numneigh_inner = (int *) 
      memory->smalloc(maxlocal*sizeof(int),"numneigh_inner");
    firstneigh_inner = (int **) 
      memory->smalloc(maxlocal*sizeof(int *),"firstneigh_inner");
    add_pages_inner(0);
    if (respa == 2) {
      numneigh_middle = (int *) 
	memory->smalloc(maxlocal*sizeof(int),"numneigh_middle");
      firstneigh_middle = (int **) 
	memory->smalloc(maxlocal*sizeof(int *),"firstneigh_middle");
      add_pages_middle(0);
    }
  }

  if (respa) {
    double *cut_respa = ((Respa *) update->integrate)->cutoff;
    cut_inner_sq = (cut_respa[1] + skin) * (cut_respa[1] + skin);
    cut_middle_sq = (cut_respa[3] + skin) * (cut_respa[3] + skin);
    cut_middle_inside_sq = (cut_respa[0] - skin) * (cut_respa[0] - skin);
  }

  // check for full flag = 2 which requires extra full neighbor list
  // setup/delete memory for extra full list as appropriate
  // use add_pages by swapping values for maxpage,pages

  int full_previous = full;
  full = 0;
  if (force->pair_match("silicon")) full = 1;
  for (int i = 0; i < modify->nfix; i++) {
    if (strcmp(modify->fix[i]->style,"CENTRO") == 0) full = 2;
    if (strcmp(modify->fix[i]->style,"flatgbforce") == 0) full = 2;
  }

  if (full != 2 && full_previous == 2) {
    memory->sfree(numneigh_full);
    memory->sfree(firstneigh_full);
    for (int i = 0; i < maxpage; i++) memory->sfree(pages_full[i]);
    memory->sfree(pages_full);
    pages_full = NULL;
    maxpage_full = 0;
    maxfull = 0;
  } else if (full == 2 && full_previous != 2) {
    maxfull = maxlocal;
    numneigh_full = (int *) 
      memory->smalloc(maxfull*sizeof(int),"numneigh_full");
    firstneigh_full = (int **) 
      memory->smalloc(maxfull*sizeof(int *),"firstneigh_full");

    int **pages_hold = pages;
    int maxpage_hold = maxpage;
    pages = NULL;
    maxpage = 0;
    add_pages(0);
    pages_full = pages;
    maxpage_full = maxpage;
    pages = pages_hold;
    maxpage = maxpage_hold;
  }

  // set ptr to correct pair build function
  // granular and respa (w/ inner/middle/outer) use their own pair functions

  if (strcmp(atom->style,"granular") == 0) {
    if (style == 0) {
      if (force->newton_pair == 0) 
	pair_build = &Neighbor::pair_nsq_no_newton_granular;
      else pair_build = &Neighbor::pair_nsq_newton_granular;
    } else if (style == 1) {
      if (force->newton_pair == 0) 
	pair_build = &Neighbor::pair_bin_no_newton_granular;
      else pair_build = &Neighbor::pair_bin_newton_granular;
    }
  } else if (respa) {
    if (style == 0) {
      if (force->newton_pair == 0) 
	pair_build = &Neighbor::pair_nsq_no_newton_respa;
      else pair_build = &Neighbor::pair_nsq_newton_respa;
    } else if (style == 1) {
      if (force->newton_pair == 0) 
	pair_build = &Neighbor::pair_bin_no_newton_respa;
      else pair_build = &Neighbor::pair_bin_newton_respa;
    }
  } else if (force->pair == NULL) {
    pair_build = &Neighbor::pair_none;
  } else {
    if (style == 0) {
      if (force->newton_pair == 0)
	pair_build = &Neighbor::pair_nsq_no_newton;
      else pair_build = &Neighbor::pair_nsq_newton;
    } else if (style == 1) {
      if (force->newton_pair == 0)
	pair_build = &Neighbor::pair_bin_no_newton;
      else pair_build = &Neighbor::pair_bin_newton;
    }
  }

  // set ptrs to correct intra-molecular build functions

  if (bond_off) bond_build = &Neighbor::bond_build_partial;
  else bond_build = &Neighbor::bond_build_all;

  if (angle_off) angle_build = &Neighbor::angle_build_partial;
  else angle_build = &Neighbor::angle_build_all;

  if (dihedral_off) dihedral_build = &Neighbor::dihedral_build_partial;
  else dihedral_build = &Neighbor::dihedral_build_all;

  if (improper_off) improper_build = &Neighbor::improper_build_partial;
  else improper_build = &Neighbor::improper_build_all;

  // set intra-molecular neighbor list counts to 0
  // in case all are turned off but potential is still defined

  nbondlist = nanglelist = ndihedrallist = nimproperlist = 0;
}

/* ---------------------------------------------------------------------- */

int Neighbor::decide()
{
  if (must_check) {
    int n = update->ntimestep;
    if (restart_check && n == output->next_restart) return 1;
    if (fix_check && n == modify->fix[fixwhich]->next_reneighbor) return 1;
  }

  ago++;
  if (ago >= delay && ago % every == 0) {
    if (dist_check == 0) return 1;
    else return check_distance();
  } else return 0;
}

/* ---------------------------------------------------------------------- */

int Neighbor::check_distance()
{
  double delx,dely,delz,rsq;

  int nlocal = atom->nlocal;
  double **x = atom->x;
  int flag = 0;
  for (int i = 0; i < nlocal; i++) {
    delx = x[i][0] - xhold[i][0];
    dely = x[i][1] - xhold[i][1];
    delz = x[i][2] - xhold[i][2];
    rsq = delx*delx + dely*dely + delz*delz;
    if (rsq > triggersq) flag = 1;
  }

  int flagall;
  MPI_Allreduce(&flag,&flagall,1,MPI_INT,MPI_MAX,world);
  if (flagall && ago == MAX(every,delay)) ndanger++;
  return flagall;
}

/* ---------------------------------------------------------------------- */

void Neighbor::build()
{
  ago = 0;
  ncalls++;

  // store current nlocal used on this build

  nlocal_neighbor = atom->nlocal;

  // store current atom positions

  if (dist_check) {
    double **x = atom->x;
    int nlocal = atom->nlocal;
    if (nlocal > maxhold) {
      maxhold = nlocal;
      memory->destroy_2d_double_array(xhold);
      xhold = memory->create_2d_double_array(maxhold,3,"neighbor:xhold");
    }
    for (int i = 0; i < nlocal; i++) {
      xhold[i][0] = x[i][0];
      xhold[i][1] = x[i][1];
      xhold[i][2] = x[i][2];
    }
  }

  // extend atom arrays if necessary

  if (atom->nlocal > maxlocal) {
    maxlocal = atom->nlocal;

    memory->sfree(numneigh);
    memory->sfree(firstneigh);
    numneigh = (int *)
      memory->smalloc(maxlocal*sizeof(int),"neigh:numneigh");
    firstneigh = (int **)
      memory->smalloc(maxlocal*sizeof(int *),"neigh:firstneigh");

    if (history >= 0) {
      memory->sfree(firsttouch);
      memory->sfree(firstshear);
      firsttouch = (int **) 
	memory->smalloc(maxlocal*sizeof(int *),"neigh:firsttouch");
      firstshear = (double **)
	memory->smalloc(maxlocal*sizeof(double *),"neigh:firstshear");
    }

    if (respa) {
      memory->sfree(numneigh_inner);
      memory->sfree(firstneigh_inner);
      numneigh_inner = (int *)
	memory->smalloc(maxlocal*sizeof(int),"neigh:numneigh_inner");
      firstneigh_inner = (int **)
	memory->smalloc(maxlocal*sizeof(int *),"neigh:firstneigh_inner");
      if (respa == 2) {
	memory->sfree(numneigh_middle);
	memory->sfree(firstneigh_middle);
	numneigh_middle = (int *)
	  memory->smalloc(maxlocal*sizeof(int),"neigh:numneigh_middle");
	firstneigh_middle = (int **)
	  memory->smalloc(maxlocal*sizeof(int *),"neigh:firstneigh_middle");
      }
    }
  }

  // extend bin list if necessary

  if (style && atom->nmax > maxbin) {
    maxbin = atom->nmax;
    memory->sfree(bins);
    bins = (int *) memory->smalloc(maxbin*sizeof(int),"bins");
  }

  // list construction for pairs and bonds

  (this->*pair_build)();

  if (atom->molecular) {
    if (atom->nbonds) (this->*bond_build)();
    if (atom->nangles) (this->*angle_build)();
    if (atom->ndihedrals) (this->*dihedral_build)();
    if (atom->nimpropers) (this->*improper_build)();
  }
}

/* ----------------------------------------------------------------------
   build an extra full neighbor list where each atom stores all its neighbors
   done via pair_full routines, but must swap values before and after call
------------------------------------------------------------------------- */

void Neighbor::build_full()
{
  // extend atom arrays if necessary

  if (atom->nlocal > maxfull) {
    maxfull = atom->nlocal;

    memory->sfree(numneigh_full);
    memory->sfree(firstneigh_full);
    numneigh_full = (int *)
      memory->smalloc(maxfull*sizeof(int),"neigh:numneigh_full");
    firstneigh_full = (int **)
      memory->smalloc(maxfull*sizeof(int *),"neigh:firstneigh_full");
  }

  // swap values

  int *numneigh_hold = numneigh;
  int **firstneigh_hold = firstneigh;
  int **pages_hold = pages;
  int maxpage_hold = maxpage;
  int *stencil_hold = stencil;
  int nstencil_hold = nstencil;

  numneigh = numneigh_full;
  firstneigh = firstneigh_full;
  pages = pages_full;
  maxpage = maxpage_full;
  stencil = stencil_full;
  nstencil = nstencil_full;

  // list construction

  if (style == 0) pair_nsq_full();
  else pair_bin_full();

  // unswap values
  // pages, maxpage may have changed due to realloc, so reset full versions
  // numneigh, firstneigh, stencil, nstencil did not change

  pages_full = pages;
  maxpage_full = maxpage;

  numneigh = numneigh_hold;
  firstneigh = firstneigh_hold;
  pages = pages_hold;
  maxpage = maxpage_hold;
  stencil = stencil_hold;
  nstencil = nstencil_hold;
}

/* ----------------------------------------------------------------------
   setup neighbor binning parameters
   bin numbering is global: 0 = 0.0 to binsize
                            1 = binsize to 2*binsize
			    nbin-1 = prd-binsize to binsize
			    nbin = prd to prd+binsize
                          -1 = -binsize to 0.0
   coord = lowest and highest values of ghost atom coords I will have
           add in "small" for round-off safety
   mbinlo = lowest global bin any of my ghost atoms could fall into
   mbinhi = highest global bin any of my ghost atoms could fall into
   mbin = number of bins I need in a dimension
   stencil() = bin offsets in 1d sense for stencil of surrounding bins
   stencil_full() = bin offsets in 1d sense for stencil for full neighbor list
------------------------------------------------------------------------- */

void Neighbor::setup_bins()
{
  // divide box into bins
  // optimal size is roughly 1/2 the cutoff

  nbinx = static_cast<int> (2.0 * domain->xprd / cutneigh);
  nbiny = static_cast<int> (2.0 * domain->yprd / cutneigh);
  if (force->dimension == 3)
    nbinz = static_cast<int> (2.0 * domain->zprd / cutneigh);
  else nbinz = 1;

  if (nbinx == 0) nbinx = 1;
  if (nbiny == 0) nbiny = 1;
  if (nbinz == 0) nbinz = 1;

  binsizex = domain->xprd/nbinx;
  binsizey = domain->yprd/nbiny;
  binsizez = domain->zprd/nbinz;

  bininvx = 1.0 / binsizex;
  bininvy = 1.0 / binsizey;
  bininvz = 1.0 / binsizez;

  // find lowest and highest bins my ghost atoms could be in

  double coord;
  int mbinxhi,mbinyhi,mbinzhi;

  coord = domain->subxlo - cutneigh - SMALL*domain->xprd;
  mbinxlo = static_cast<int> ((coord-domain->boxxlo)*bininvx);
  if (coord < 0.0) mbinxlo = mbinxlo - 1;
  coord = domain->subxhi + cutneigh + SMALL*domain->xprd;
  mbinxhi = static_cast<int> ((coord-domain->boxxlo)*bininvx);

  coord = domain->subylo - cutneigh - SMALL*domain->yprd;
  mbinylo = static_cast<int> ((coord-domain->boxylo)*bininvy);
  if (coord < 0.0) mbinylo = mbinylo - 1;
  coord = domain->subyhi + cutneigh + SMALL*domain->yprd;
  mbinyhi = static_cast<int> ((coord-domain->boxylo)*bininvy);

  coord = domain->subzlo - cutneigh - SMALL*domain->zprd;
  mbinzlo = static_cast<int> ((coord-domain->boxzlo)*bininvz);
  if (coord < 0.0) mbinzlo = mbinzlo - 1;
  coord = domain->subzhi + cutneigh + SMALL*domain->zprd;
  mbinzhi = static_cast<int> ((coord-domain->boxzlo)*bininvz);

  // extend bins by 1 to insure stencil extent is included

  mbinxlo = mbinxlo - 1;
  mbinxhi = mbinxhi + 1;
  mbinx = mbinxhi - mbinxlo + 1;

  mbinylo = mbinylo - 1;
  mbinyhi = mbinyhi + 1;
  mbiny = mbinyhi - mbinylo + 1;

  mbinzlo = mbinzlo - 1;
  mbinzhi = mbinzhi + 1;
  mbinz = mbinzhi - mbinzlo + 1;

  // memory for bin ptrs

  mbins = mbinx*mbiny*mbinz;
  if (mbins > maxhead) {
    maxhead = mbins;
    memory->sfree(binhead);
    binhead = (int *) memory->smalloc(maxhead*sizeof(int),"neighbor:binhead");
  }

  // create stencil of bins whose closest corner to central bin
  //   is within neighbor cutoff
  // next(xyz) = how far the stencil could possibly extend
  // for partial Newton (newton = 0)
  //   stencil is all surrounding bins including self
  // for full Newton (newton = 1)
  //   stencil is bins to the "upper right" of central bin
  //   stencil does NOT include self
  // for full neighbor list
  //   stencil is all surrounding bins including self, regardless of Newton
  //   stored in stencil when full = 1
  //   stored in stencil_full when full = 2
  // 3d creates xyz stencil, 2d is only xy

  int nextx = static_cast<int> (cutneigh*bininvx);
  if (nextx*binsizex < cutneigh) nextx++;
  int nexty = static_cast<int> (cutneigh*bininvy);
  if (nexty*binsizey < cutneigh) nexty++;
  int nextz = static_cast<int> (cutneigh*bininvz);
  if (nextz*binsizez < cutneigh) nextz++;

  int nmax = (2*nextz+1) * (2*nexty+1) * (2*nextx+1);
  if (nmax > maxstencil) {
    maxstencil = nmax;
    memory->sfree(stencil);
    stencil = (int *) memory->smalloc(maxstencil*sizeof(int),
				      "neighbor:stencil");
  }

  int i,j,k;
  nstencil = 0;
  double cutsq = cutneigh*cutneigh;

  if (force->dimension == 3) {
    if (force->newton_pair == 0 || full == 1) {
      for (k = -nextz; k <= nextz; k++)
	for (j = -nexty; j <= nexty; j++)
	  for (i = -nextx; i <= nextx; i++)
	    if (bin_distance(i,j,k) < cutsq)
	      stencil[nstencil++] = k*mbiny*mbinx + j*mbinx + i;
    } else {
      for (k = 0; k <= nextz; k++)
	for (j = -nexty; j <= nexty; j++)
	  for (i = -nextx; i <= nextx; i++)
	    if (k > 0 || j > 0 || (j == 0 && i > 0))
	      if (bin_distance(i,j,k) < cutsq)
		stencil[nstencil++] = k*mbiny*mbinx + j*mbinx + i;
    }
  } else {
    if (force->newton_pair == 0 || full == 1) {
      for (j = -nexty; j <= nexty; j++)
	for (i = -nextx; i <= nextx; i++)
	  if (bin_distance(i,j,0) < cutsq)
	    stencil[nstencil++] = j*mbinx + i;
    } else {
	for (j = 0; j <= nexty; j++)
	  for (i = -nextx; i <= nextx; i++)
	    if (j > 0 || (j == 0 && i > 0))
	      if (bin_distance(i,j,0) < cutsq)
		stencil[nstencil++] = j*mbinx + i;
    }
  }

  if (full == 2) {
    if (nmax > maxstencil_full) {
      maxstencil_full = nmax;
      memory->sfree(stencil_full);
      stencil_full = (int *) memory->smalloc(maxstencil_full*sizeof(int),
					     "neighbor:stencil_full");
    }
    nstencil_full = 0;
    if (force->dimension == 3) {
      for (k = -nextz; k <= nextz; k++)
	for (j = -nexty; j <= nexty; j++)
	  for (i = -nextx; i <= nextx; i++)
	    if (bin_distance(i,j,k) < cutsq)
	      stencil_full[nstencil_full++] = k*mbiny*mbinx + j*mbinx + i;
    } else {
      for (j = -nexty; j <= nexty; j++)
	for (i = -nextx; i <= nextx; i++)
	  if (bin_distance(i,j,0) < cutsq)
	    stencil_full[nstencil_full++] = j*mbinx + i;
    }
  }
}
      
/* ----------------------------------------------------------------------
   compute closest distance between central bin (0,0,0) and bin (i,j,k)
------------------------------------------------------------------------- */

double Neighbor::bin_distance(int i, int j, int k)
{
  double delx,dely,delz;

  if (i > 0)
    delx = (i-1)*binsizex;
  else if (i == 0)
    delx = 0.0;
  else
    delx = (i+1)*binsizex;

  if (j > 0)
    dely = (j-1)*binsizey;
  else if (j == 0)
    dely = 0.0;
  else
    dely = (j+1)*binsizey;

  if (k > 0)
    delz = (k-1)*binsizez;
  else if (k == 0)
    delz = 0.0;
  else
    delz = (k+1)*binsizez;
 
  return (delx*delx + dely*dely + delz*delz);
}

/* ----------------------------------------------------------------------
   modify parameters of the pair-wise neighbor build
------------------------------------------------------------------------- */

void Neighbor::modify_params(int narg, char **arg)
{
  int iarg = 0;
  while (iarg < narg) {
    if (strcmp(arg[iarg],"every") == 0) {
      if (iarg+2 > narg) error->all("Illegal neigh_modify command");
      every = atoi(arg[iarg+1]);
      iarg += 2;
    } else if (strcmp(arg[iarg],"delay") == 0) {
      if (iarg+2 > narg) error->all("Illegal neigh_modify command");
      delay = atoi(arg[iarg+1]);
      iarg += 2;
    } else if (strcmp(arg[iarg],"check") == 0) {
      if (iarg+2 > narg) error->all("Illegal neigh_modify command");
      if (strcmp(arg[iarg+1],"yes") == 0) dist_check = 1;
      else if (strcmp(arg[iarg+1],"no") == 0) dist_check = 0;
      else error->all("Illegal neigh_modify command");
      iarg += 2;
    } else if (strcmp(arg[iarg],"page") == 0) {
      if (iarg+2 > narg) error->all("Illegal neigh_modify command");
      pgsize = atoi(arg[iarg+1]);
      iarg += 2;
    } else if (strcmp(arg[iarg],"one") == 0) {
      if (iarg+2 > narg) error->all("Illegal neigh_modify command");
      oneatom = atoi(arg[iarg+1]);
      iarg += 2;
    } else if (strcmp(arg[iarg],"exclude") == 0) {
      if (iarg+2 > narg) error->all("Illegal neigh_modify command");

      if (strcmp(arg[iarg+1],"type") == 0) {
	if (iarg+4 > narg) error->all("Illegal neigh_modify command");
	if (nex_type == maxex_type) {
	  maxex_type += EXDELTA;
	  ex1_type = (int *) memory->srealloc(ex1_type,maxex_type*sizeof(int),
					      "neighbor:ex1_type");
	  ex2_type = (int *) memory->srealloc(ex2_type,maxex_type*sizeof(int),
					      "neighbor:ex2_type");
	}
	ex1_type[nex_type] = atoi(arg[iarg+2]);
	ex2_type[nex_type] = atoi(arg[iarg+3]);
	nex_type++;
	iarg += 4;

      } else if (strcmp(arg[iarg+1],"group") == 0) {
	if (iarg+4 > narg) error->all("Illegal neigh_modify command");
	if (nex_group == maxex_group) {
	  maxex_group += EXDELTA;
	  ex1_group = 
	    (int *) memory->srealloc(ex1_group,maxex_group*sizeof(int),
				     "neighbor:ex1_group");
	  ex2_group = 
	    (int *) memory->srealloc(ex2_group,maxex_group*sizeof(int),
				     "neighbor:ex2_group");
	}
	ex1_group[nex_group] = group->find(arg[iarg+2]);
	ex2_group[nex_group] = group->find(arg[iarg+3]);
	if (ex1_group[nex_group] == -1 || ex2_group[nex_group] == -1)
	  error->all("Illegal group ID in neigh_modify command");
	nex_group++;
	iarg += 4;

      } else if (strcmp(arg[iarg+1],"molecule") == 0) {
	if (iarg+3 > narg) error->all("Illegal neigh_modify command");
	if (atom->molecular == 0) {
	  char *str = "Must use molecular atom style with neigh_modify exclude molecule";
	  error->all(str);
	}
	if (nex_mol == maxex_mol) {
	  maxex_mol += EXDELTA;
	  ex_mol_group = 
	    (int *) memory->srealloc(ex_mol_group,maxex_mol*sizeof(int),
				     "neighbor:ex_mol_group");
	}
	ex_mol_group[nex_mol] = group->find(arg[iarg+2]);
	if (ex_mol_group[nex_mol] == -1)
	  error->all("Illegal group ID in neigh_modify command");
	nex_mol++;
	iarg += 3;

      } else if (strcmp(arg[iarg+1],"none") == 0) {
	nex_type = nex_group = nex_mol = 0;
	iarg += 2;
      } else error->all("Incorrect args in neigh_modify command");

    } else error->all("Incorrect args in neigh_modify command");
  }
}

/* ----------------------------------------------------------------------
   return # of bytes of allocated memory
------------------------------------------------------------------------- */

int Neighbor::memory_usage()
{
  int bytes = 0;

  bytes += maxlocal * sizeof(int);
  bytes += maxlocal * sizeof(int *);

  bytes += bondmax*3 * sizeof(int);
  bytes += anglemax*4 * sizeof(int);
  bytes += dihedralmax*5 * sizeof(int);
  bytes += impropermax*5 * sizeof(int);

  bytes += maxbin * sizeof(int);
  bytes += maxhead * sizeof(int);
  bytes += maxstencil * sizeof(int);
  bytes += maxhold*3 * sizeof(double);

  bytes += maxpage*pgsize * sizeof(int);

  if (history >= 0) {
    bytes += maxlocal * sizeof(int *);
    bytes += maxlocal * sizeof(double *);
    bytes += maxpage*pgsize * sizeof(int);
    bytes += maxpage*pgsize*3 * sizeof(double);
  }

  if (respa) {
    bytes += maxlocal * sizeof(int);
    bytes += maxlocal * sizeof(int *);
    bytes += maxpage_inner*pgsize * sizeof(int);
    if (respa == 2) {
      bytes += maxlocal * sizeof(int);
      bytes += maxlocal * sizeof(int *);
      bytes += maxpage_middle*pgsize * sizeof(int);
    }
  }

  if (full == 2) {
    bytes += maxfull * sizeof(int);
    bytes += maxfull * sizeof(int *);
    bytes += maxpage_full*pgsize * sizeof(int);
    bytes += maxstencil_full * sizeof(int);
  }

  return bytes;
}

/* ----------------------------------------------------------------------
   add pages to neighbor list, starting at npage
------------------------------------------------------------------------- */

void Neighbor::add_pages(int npage)
{
  maxpage += PGDELTA;
  pages = (int **) 
    memory->srealloc(pages,maxpage*sizeof(int *),"neighbor:pages");
  for (int i = npage; i < maxpage; i++)
    pages[i] = (int *) memory->smalloc(pgsize*sizeof(int),"neighbor:pages[i]");
}

/* ----------------------------------------------------------------------
   add pages to granular neighbor list, starting at npage
------------------------------------------------------------------------- */

void Neighbor::add_pages_history(int npage)
{
  pages_touch = (int **)
    memory->srealloc(pages_touch,maxpage*sizeof(int *),"neighbor:pages_touch");
  pages_shear = (double **)
    memory->srealloc(pages_shear,maxpage*sizeof(double *),
		     "neighbor:pages_shear");
  for (int i = npage; i < maxpage; i++) {
    pages_touch[i] = (int *)
      memory->smalloc(pgsize*sizeof(int),"neighbor:pages_touch[i]");
    pages_shear[i] = (double *)
      memory->smalloc(3*pgsize*sizeof(double),"neighbor:pages_shear[i]");
  }
}

/* ----------------------------------------------------------------------
   add pages to rRESPA inner neighbor list, starting at npage_inner
------------------------------------------------------------------------- */

void Neighbor::add_pages_inner(int npage_inner)
{
  maxpage_inner += PGDELTA;
  pages_inner = (int **) 
    memory->srealloc(pages_inner,maxpage_inner*sizeof(int *),
		     "neighbor:pages_inner");
  for (int i = npage_inner; i < maxpage_inner; i++)
    pages_inner[i] = 
      (int *) memory->smalloc(pgsize*sizeof(int),"neighbor:pages_inner[i]");
}

/* ----------------------------------------------------------------------
   add pages to rRESPA middle neighbor list, starting at npage_middle
------------------------------------------------------------------------- */

void Neighbor::add_pages_middle(int npage_middle)
{
  maxpage_middle += PGDELTA;
  pages_middle = (int **) 
    memory->srealloc(pages_middle,maxpage_middle*sizeof(int *),
		     "neighbor:pages_middle");
  for (int i = npage_middle; i < maxpage_middle; i++)
    pages_middle[i] = 
      (int *) memory->smalloc(pgsize*sizeof(int),"neighbor:pages_middle[i]");
}

/* ----------------------------------------------------------------------
   determine if atom j is in special list of atom i
   if it is not, return 0
   if it is and special flag is 0 (both coeffs are 0.0), return -1
   if it is and special flag is 1 (both coeffs are 1.0), return 0
   if it is and special flag is 2 (otherwise), return 1,2,3
     for which neighbor it is (and which coeff it maps to)
------------------------------------------------------------------------- */

int Neighbor::find_special(int i, int j)
{
  int *list = atom->special[i];
  int n1 = atom->nspecial[i][0];
  int n2 = atom->nspecial[i][1];
  int n3 = atom->nspecial[i][2];
  int tag = atom->tag[j];

  for (int i = 0; i < n3; i++) {
    if (list[i] == tag) {
      if (i < n1) {
	if (special_flag[1] == 0) return -1;
	else if (special_flag[1] == 1) return 0;
	else return 1;
      } else if (i < n2) {
	if (special_flag[2] == 0) return -1;
	else if (special_flag[2] == 1) return 0;
	else return 2;
      } else {
	if (special_flag[3] == 0) return -1;
	else if (special_flag[3] == 1) return 0;
	else return 3;
      }
    }
  }
  return 0;
}

/* ----------------------------------------------------------------------
   bin owned and ghost atoms
------------------------------------------------------------------------- */

void Neighbor::bin_atoms()
{
  int i,ibin,nlocal,nall;
  double **x;

  nlocal = atom->nlocal;
  nall = atom->nlocal + atom->nghost;
  x = atom->x;

  for (i = 0; i < mbins; i++) binhead[i] = -1;

  // bin ghost atoms 1st, so will be at end of linked list

  for (i = nlocal; i < nall; i++) {
    ibin = coord2bin(x[i]);
    bins[i] = binhead[ibin];
    binhead[ibin] = i;
  }

  // bin owned atoms

  for (i = 0; i < nlocal; i++) {
    ibin = coord2bin(x[i]);
    bins[i] = binhead[ibin];
    binhead[ibin] = i;
  }
}

/* ----------------------------------------------------------------------
   convert atom coords into local bin #
   take special care to insure ghost atoms with
   coord >= prd or coord < 0.0 are put in correct bins
------------------------------------------------------------------------- */

int Neighbor::coord2bin(double *x)
{
  int ix,iy,iz;

  if (x[0] >= domain->boxxhi)
    ix = static_cast<int> ((x[0]-domain->boxxhi)*bininvx) + nbinx - mbinxlo;
  else if (x[0] >= domain->boxxlo)
    ix = static_cast<int> ((x[0]-domain->boxxlo)*bininvx) - mbinxlo;
  else
    ix = static_cast<int> ((x[0]-domain->boxxlo)*bininvx) - mbinxlo - 1;
  
  if (x[1] >= domain->boxyhi)
    iy = static_cast<int> ((x[1]-domain->boxyhi)*bininvy) + nbiny - mbinylo;
  else if (x[1] >= domain->boxylo)
    iy = static_cast<int> ((x[1]-domain->boxylo)*bininvy) - mbinylo;
  else
    iy = static_cast<int> ((x[1]-domain->boxylo)*bininvy) - mbinylo - 1;
  
  if (x[2] >= domain->boxzhi)
    iz = static_cast<int> ((x[2]-domain->boxzhi)*bininvz) + nbinz - mbinzlo;
  else if (x[2] >= domain->boxzlo)
    iz = static_cast<int> ((x[2]-domain->boxzlo)*bininvz) - mbinzlo;
  else
    iz = static_cast<int> ((x[2]-domain->boxzlo)*bininvz) - mbinzlo - 1;

  return (iz*mbiny*mbinx + iy*mbinx + ix + 1);
}

/* ----------------------------------------------------------------------
   test if atom pair i,j is excluded from neighbor list
   due to type, group, molecule settings from neigh_modify command
   return 1 if should be excluded, 0 if included
------------------------------------------------------------------------- */

int Neighbor::exclusion(int i, int j, int *type, int *mask, int *molecule)
{
  int m;

  if (nex_type && ex_type[type[i]][type[j]]) return 1;

  if (nex_group) {
    for (m = 0; m < nex_group; m++) {
      if (mask[i] & ex1_bit[m] && mask[j] & ex2_bit[m]) break;
      if (mask[i] & ex2_bit[m] && mask[j] & ex1_bit[m]) break;
    }
    if (m < nex_group) return 1;
  }

  if (nex_mol) {
    for (m = 0; m < nex_mol; m++) {
      if (mask[i] & ex_mol_bit[m] && mask[j] & ex_mol_bit[m] &&
	  molecule[i] == molecule[j]) break;
    }
    if (m < nex_mol) return 1;
  }

  return 0;
}
