/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#ifndef NEIGHBOR_H
#define NEIGHBOR_H

#include "system.h"

class Neighbor : public System {
 public:
  int style;                       // 0 = nsq, 1 = binned
  int every;                       // build every this many steps
  int delay;                       // delay build for this many steps
  int dist_check;                  // 0 = always build, 1 = only if 1/2 dist
  int ago;                         // how many steps ago neighboring occurred
  int pgsize;                      // size of neighbor page
  int oneatom;                     // max # of neighbors for one atom

  int must_check;                  // 1 if must check other classes to reneigh
  int restart_check;               // 1 if restart enabled, 0 if no
  int fix_check;                   // 1 if reneigh fix enabled, 0 if no
  int fixwhich;                    // index of which fix to check

  double skin;                     // skin distance
  double cutneigh;                 // neighbor cutoff
  double **cutneighsq;             // neighbor cutoff squared for type pairs
  double triggersq;                // sq distance to trigger build on

  int nbinx,nbiny,nbinz;           // # of global bins
  int ncalls;                      // # of times build has been called
  int ndanger;                     // # of dangerous builds
  int max_totalneigh;              // largest # of neighbors ever stored
  int nlocal_neighbor;             // nlocal at last build

  int *numneigh;                   // # of pair neighbors for each atom
  int **firstneigh;                // ptr to 1st neighbor of each atom

  int nbondlist;                   // list of bonds to compute
  int bondmax;
  int **bondlist;

  int nanglelist;                  // list of angles to compute
  int anglemax;
  int **anglelist;

  int ndihedrallist;               // list of dihedrals to compute
  int dihedralmax;
  int **dihedrallist;

  int nimproperlist;               // list of impropers to compute
  int impropermax;
  int **improperlist;
                                   // granular neighbor list info
  int **firsttouch;                // ptr to first touch flag for each atom
  double **firstshear;             // ptr to first shear values for each atom

                                   // respa neighbor list info
  int respa;                       // 0 = single neighbor list
                                   // 1 = additional inner list
                                   // 2 = additional inner and middle list

  int *numneigh_inner;             // # of inner pair neighbors for each atom
  int **firstneigh_inner;          // ptr to inner 1st neigh of each atom
  int *numneigh_middle;            // # of middle pair neighbors for each atom
  int **firstneigh_middle;         // ptr to middle 1st neigh of each atom

  double inner[2],middle[2];       // rRESPA cutoffs for extra lists
  double cut_inner_sq;		   // outer cutoff for inner neighbor list
  double cut_middle_sq;            // outer cutoff for middle neighbor list
  double cut_middle_inside_sq;     // inner cutoff for middle neighbor list

                                   // full neighbor list info
  int full;                        // 0 = single normal list
                                   // 1 = single full list
                                   // 2 = extra full list
  int *numneigh_full;              // # of full neighbors for each atom
  int **firstneigh_full;           // ptr to 1st full neighbor of each atom

  Neighbor();
  ~Neighbor();
  void init();
  int decide();               // decide whether to build or not
  int check_distance();       // check max distance moved since last build
  void setup_bins();          // setup bins based on box and cutoff
  void build();               // create pair neighbor list
  void build_full();          // create extra full neighbor list
  void modify_params(int, char**);  // modify parameters of neighbor build
  int memory_usage();         // tally memory usage
  
 private:
  int me,nprocs;

  int npage;                       // current page in page list
  int **pages;                     // neighbor list pages
  int maxpage;                     // # of pages currently allocated

  int maxlocal;                    // size of numneigh, firstneigh arrays

  int *bins;                       // ptr to next atom in each bin
  int maxbin;                      // size of bins array

  int *binhead;                    // ptr to 1st atom in each bin
  int maxhead;                     // size of binhead array

  int nstencil;                    // # of bins in stencil
  int *stencil;                    // stencil list of bin offsets
  int maxstencil;                  // size of stencil

  double **xhold;                  // atom coords at last neighbor build
  int maxhold;                     // size of xhold array

  int mbins;                       // binning parameters
  int mbinx,mbiny,mbinz;
  int mbinxlo,mbinylo,mbinzlo;
  double binsizex,binsizey,binsizez;
  double bininvx,bininvy,bininvz;

                                   // granular neighbor list info
  int history;                     // -1 if history not needed
                                   // else stores fix shear history index
  int **pages_touch;               // pages of touch flags
  double **pages_shear;            // pages of shear values

                                   // respa multiple neighbor list info
  int **pages_inner;               // inner neighbor list pages
  int maxpage_inner;               // # of inner pages currently allocated
  int **pages_middle;              // middle neighbor list pages
  int maxpage_middle;              // # of middle pages currently allocated

                                   // full neighbor info (all atom's neighs)
  int maxfull;                     // size of numneigh_full, firstneigh_full
  int **pages_full;                // full neighbor list pages
  int maxpage_full;                // # of full pages currently allocated

  int nstencil_full;               // # of bins in full neighbor stencil
  int *stencil_full;               // full neighbor stencil list of bin offsets
  int maxstencil_full;             // size of full neighbor stencil

  int special_flag[4];             // flags for 1-2, 1-3, 1-4 neighbors

  int bond_off,angle_off,dihedral_off,improper_off;
                                   // 1 if bond, etc can break or be turned off
                                   // "off" is when type is negative

  int exclude;                    // 0 if no type/group exclusions, 1 if yes

  int nex_type;                   // # of entries in type exclusion list
  int maxex_type;                 // max # in type list
  int *ex1_type,*ex2_type;        // pairs of types to exclude
  int **ex_type;                  // 2d array of excluded type pairs

  int nex_group;                  // # of entries in group exclusion list
  int maxex_group;                // max # in group list
  int *ex1_group,*ex2_group;      // pairs of group #'s to exclude
  int *ex1_bit,*ex2_bit;          // pairs of group bits to exclude

  int nex_mol;                    // # of entries in molecule exclusion list
  int maxex_mol;                  // max # in molecule list
  int *ex_mol_group;              // molecule group #'s to exclude
  int *ex_mol_bit;                // molecule group bits to exclude

  typedef void (Neighbor::*FnPtr)();
  FnPtr pair_build;           // ptr to pair list functions

  void pair_none() {}         // pairwise interactions are turned off
  void pair_nsq_no_newton();  // via Nsq method w/ inter-proc Newton off
  void pair_nsq_newton();     // via Nsq method w/ inter-proc Newton on
  void pair_bin_no_newton();  // via binning w/ inter-proc Newton off
  void pair_bin_newton();     // via binning w/ inter-proc Newton on
  void pair_nsq_no_newton_granular();    // 4 fns for granular systems
  void pair_nsq_newton_granular();
  void pair_bin_no_newton_granular();
  void pair_bin_newton_granular();
  void pair_nsq_no_newton_respa();       // 4 fns for multiple respa lists
  void pair_nsq_newton_respa();
  void pair_bin_no_newton_respa();
  void pair_bin_newton_respa();
  void pair_nsq_full();                  // 2 fns for full neighbor lists
  void pair_bin_full();

  FnPtr bond_build;           // ptr to bond list functions
  void bond_build_all();      // bond list with all bonds
  void bond_build_partial();  // exclude certain bonds

  FnPtr angle_build;          // ptr to angle list functions
  void angle_build_all();     // angle list with all angles
  void angle_build_partial(); // exclude certain angles

  FnPtr dihedral_build;          // ptr to dihedral list functions
  void dihedral_build_all();     // dihedral list with all dihedrals
  void dihedral_build_partial(); // exclude certain dihedrals

  FnPtr improper_build;          // ptr to improper list functions
  void improper_build_all();     // improper list with all impropers
  void improper_build_partial(); // exclude certain impropers

  void add_pages(int);           // add pages to neighbor list
  void add_pages_history(int);   // add pages to granular neighbor list
  void add_pages_inner(int);     // add pages to inner neighbor list (rRESPA)
  void add_pages_middle(int);    // add pages to middle neighbor list (rRESPA)

  void bin_atoms();                     // bin all atoms
  double bin_distance(int, int, int);   // distance between binx
  int coord2bin(double *);              // mapping atom coord to a bin
  int find_special(int, int);           // look for special neighbor
  int exclusion(int, int, int *, int *, int *);  // test for pair exclusion
};

#endif
