/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "stdio.h"
#include "string.h"
#include "output.h"
#include "atom.h"
#include "neighbor.h"
#include "comm.h"
#include "update.h"
#include "group.h"
#include "domain.h"
#include "temperature.h"
#include "pressure.h"
#include "thermo.h"
#include "dump.h"
#include "write_restart.h"
#include "memory.h"
#include "error.h"

#define ThermoInclude
#define DumpInclude
#include "style.h"
#undef ThermoInclude
#undef DumpInclude

#define DELTA 1
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

/* ----------------------------------------------------------------------
   initialize all output 
------------------------------------------------------------------------- */

Output::Output() {
  thermo = new ThermoOne("one");
  thermo_every = 0;

  next_dump_any = 0;
  ndump = 0;
  max_dump = 0;
  next_dump = NULL;
  last_dump = NULL;
  dump_every = NULL;
  dump = NULL;

  restart = NULL;
  restart1 = restart2 = NULL;
  restart_every = 0;
}

/* ----------------------------------------------------------------------
   free all memory 
------------------------------------------------------------------------- */

Output::~Output()
{
  delete thermo;

  memory->sfree(next_dump);
  memory->sfree(last_dump);
  memory->sfree(dump_every);
  for (int i = 0; i < ndump; i++) delete dump[i];
  memory->sfree(dump);

  delete restart;
  delete [] restart1;
  delete [] restart2;
}

/* ---------------------------------------------------------------------- */

void Output::init()
{
  thermo->init();
  for (int i = 0; i < ndump; i++) dump[i]->init();
}

/* ----------------------------------------------------------------------
   perform output for setup of run 
------------------------------------------------------------------------- */

void Output::setup()
{
  int ntimestep = update->ntimestep;

  // print out memory usage

  print_memory_usage();

  // always do thermo at start of run
  // set next thermo timestep
  // insure next thermo is a multiple of every
  // always do thermo on last timestep, even if every = 0

  thermo->header();
  thermo->compute(0);

  next_thermo = ntimestep + thermo_every;
  if (thermo_every == 0) next_thermo = update->laststep;
  if (thermo_every && ntimestep % thermo_every != 0)
    next_thermo = (ntimestep/thermo_every)*thermo_every + thermo_every;
  next_thermo = MIN(next_thermo,update->laststep);

  // dump at start of run only if dump has never been done before
  // set next dump timestep
  // insure next dump is a multiple of every

  int idump;
  for (idump = 0; idump < ndump; idump++)
    if (last_dump[idump] == -1) {
      dump[idump]->write();
      last_dump[idump] = ntimestep;
      next_dump[idump] = ntimestep + dump_every[idump];
      if (ntimestep % dump_every[idump] != 0)
	next_dump[idump] = (ntimestep/dump_every[idump])*dump_every[idump] + 
	  dump_every[idump];
    }

  // do not write a restart file at start of run
  // set next restart timestep
  // if every = 0, set to last + 1
  // insure next restart is a multiple of every

  next_restart = ntimestep + restart_every;
  if (restart_every == 0) next_restart = update->laststep + 1;
  if (restart_every && ntimestep % restart_every != 0)
    next_restart = (ntimestep/restart_every)*restart_every + restart_every;

  // next = next timestep any output will be done

  next = MIN(next_thermo,next_restart);
  for (idump = 0; idump < ndump; idump++) next = MIN(next,next_dump[idump]);
}

/* ----------------------------------------------------------------------
   all output for this timestep 
------------------------------------------------------------------------- */

void Output::write(int ntimestep)
{
  // thermo output and set next thermo timestep
  // always do thermo on last timestep of run

  if (next_thermo == ntimestep) {
    thermo->compute(1);
    next_thermo += thermo_every;
    next_thermo = MIN(next_thermo,update->laststep);
  }

  // dump output and set next dump timestep

  int idump;
  for (idump = 0; idump < ndump; idump++)
    if (next_dump[idump] == ntimestep) {
      dump[idump]->write();
      next_dump[idump] += dump_every[idump];
    }

  // restart output and set next restart timestep

  if (next_restart == ntimestep) {
    if (restart_toggle == 0) {
      char file[128];
      sprintf(file,"%s.%d",restart1,ntimestep);
      restart->write(file);
    } else if (restart_toggle == 1) {
      restart->write(restart1);
      restart_toggle = 2;
    } else if (restart_toggle == 2) {
      restart->write(restart2);
      restart_toggle = 1;
    }
    next_restart += restart_every;
  }

  // next = next timestep any output will be done

  next = MIN(next_thermo,next_restart);
  for (idump = 0; idump < ndump; idump++) next = MIN(next,next_dump[idump]);
}

/* ----------------------------------------------------------------------
   add a Dump to list of Dumps 
------------------------------------------------------------------------- */

void Output::add_dump(int narg, char **arg)
{
  if (narg < 5) error->all("Illegal dump command");

  // error checks

  for (int idump = 0; idump < ndump; idump++)
    if (strcmp(arg[0],dump[idump]->id) == 0) error->all("Reuse of dump ID");
  int igroup = group->find(arg[1]);
  if (igroup == -1) error->all("Could not find dump group ID");
  if (atoi(arg[3]) <= 0) error->all("Illegal dump frequency");

  // extend Dump list if necessary

  if (ndump == max_dump) {
    max_dump += DELTA;
    dump = (Dump **)
      memory->srealloc(dump,max_dump*sizeof(Dump *),"output:dump");
    dump_every = (int *)
      memory->srealloc(dump_every,max_dump*sizeof(int *),"output:dump_every");
    next_dump = (int *)
      memory->srealloc(next_dump,max_dump*sizeof(int *),"output:next_dump");
    last_dump = (int *)
      memory->srealloc(last_dump,max_dump*sizeof(int *),"output:last_dump");
  }

  // create the Dump

  if (strcmp(arg[2],"none") == 0) error->all("Illegal dump style");

#define DumpClass
#define DumpStyle(key,Class) \
  else if (strcmp(arg[2],#key) == 0) dump[ndump] = new Class(narg,arg);
#include "style.h"
#undef DumpClass

  else error->all("Illegal dump style");

  dump_every[ndump] = atoi(arg[3]);
  last_dump[ndump] = -1;
  ndump++;
}

/* ----------------------------------------------------------------------
   modify parameters of a Dump 
------------------------------------------------------------------------- */

void Output::modify_dump(int narg, char **arg)
{
  if (narg < 2) error->all("Illegal dump_modify command");

  // find which dump it is

  int idump;
  for (idump = 0; idump < ndump; idump++)
    if (strcmp(arg[0],dump[idump]->id) == 0) break;
  if (idump == ndump) error->all("Cound not find dump_modify ID");

  dump[idump]->modify_params(narg-1,&arg[1]);
}

/* ----------------------------------------------------------------------
   delete a Dump from list of Dumps 
------------------------------------------------------------------------- */

void Output::delete_dump(char *id)
{
  // find which dump it is and delete it

  int idump;
  for (idump = 0; idump < ndump; idump++)
    if (strcmp(id,dump[idump]->id) == 0) break;
  if (idump == ndump) error->all("Could not find undump ID");

  delete dump[idump];

  // move other dumps down in list one slot

  for (int i = idump+1; i < ndump; i++) {
    dump[i-1] = dump[i];
    dump_every[i-1] = dump_every[i];
    next_dump[i-1] = next_dump[i];
    last_dump[i-1] = last_dump[i];
  }
  ndump--;
}

/* ----------------------------------------------------------------------
   new Thermo style 
------------------------------------------------------------------------- */

void Output::create_thermo(char *style)
{
  // don't allow this so that dipole style can safely allocate inertia vector

  if (domain->box_exist == 0) 
    error->all("Thermo_style command before simulation box is defined");

  delete thermo;

  if (strcmp(style,"none") == 0) error->all("Illegal thermo style");

#define ThermoClass
#define ThermoStyle(key,Class) \
  else if (strcmp(style,#key) == 0) thermo = new Class(style);
#include "style.h"
#undef ThermoClass

  else error->all("Illegal thermo style");
}

/* ----------------------------------------------------------------------
   setup restart capability 
------------------------------------------------------------------------- */

void Output::create_restart(int narg, char **arg)
{
  if (narg < 1) error->all("Illegal restart command");

  if (restart) delete restart;
  delete [] restart1;
  delete [] restart2;

  restart_every = atoi(arg[0]);
  if (restart_every == 0) {
    if (narg != 1) error->all("Illegal restart command");
    return;
  }

  restart = new WriteRestart;

  int n = strlen(arg[1]) + 1;
  restart1 = new char[n];
  strcpy(restart1,arg[1]);

  if (narg == 2) {
    restart2 = NULL;
    restart_toggle = 0;
  } else if (narg == 3) {
    n = strlen(arg[2]) + 1;
    restart2 = new char[n];
    strcpy(restart2,arg[2]);
    restart_toggle = 1;
  } else error->all("Illegal restart command");
}

/* ----------------------------------------------------------------------
   sum and print memory usage 
------------------------------------------------------------------------- */

void Output::print_memory_usage()
{
  int bytes = 0;

  bytes += atom->memory_usage();
  bytes += neighbor->memory_usage();
  bytes += comm->memory_usage();

  double mbytes = bytes/1024.0/1024.0;

  if (comm->me == 0) {
    if (screen)
      fprintf(screen,"Memory usage per processor = %g Mbytes\n",mbytes);
    if (logfile) 
      fprintf(logfile,"Memory usage per processor = %g Mbytes\n",mbytes);
  }
}
