/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "math.h"
#include "string.h"
#include "ctype.h"
#include "pair_hybrid.h"
#include "atom.h"
#include "force.h"
#include "pair.h"
#include "pair_eam.h"
#include "pair_soft.h"
#include "neighbor.h"
#include "update.h"
#include "comm.h"
#include "integrate.h"
#include "memory.h"
#include "error.h"

#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

/* ---------------------------------------------------------------------- */

PairHybrid::PairHybrid()
{
  nstyles = 0;
}

/* ----------------------------------------------------------------------
   free all arrays 
------------------------------------------------------------------------- */

PairHybrid::~PairHybrid()
{
  if (nstyles) {
    for (int i = 0; i < nstyles; i++) delete styles[i];
    delete [] styles;
    for (int i = 0; i < nstyles; i++) delete [] keywords[i];
    delete [] keywords;
  }

  if (allocated) {
    memory->destroy_2d_int_array(setflag);
    memory->destroy_2d_int_array(map);
    memory->destroy_2d_double_array(cutsq);
  }
}

/* ---------------------------------------------------------------------- */

void PairHybrid::compute(int eflag, int vflag)
{
  int i,j,k,numneigh,itype,jtype;
  double xtmp,ytmp,ztmp,delx,dely,delz;
  double rsq,fforce,factor,factor_coul,factor_lj,phi;
  double rhoi,rhoj;
  int *neighs;
  double **f;
  One one;

  eng_vdwl = eng_coul = 0.0;
  if (vflag) for (i = 0; i < 6; i++) virial[i] = 0.0;

  if (vflag == 2) f = update->integrate->f_pair;
  else f = atom->f;
  double **torque = atom->torque;
  double **x = atom->x;
  int *type = atom->type;
  int nlocal = atom->nlocal;
  int nall = atom->nlocal + atom->nghost;
  double *special_coul = force->special_coul;
  double *special_lj = force->special_lj;
  int newton_pair = force->newton_pair;

  // if soft sub-style exists, set prefactor used by single()

  if (softstyle) {
    double **prefactor = softstyle->prefactor;
    double **prestart = softstyle->prestart;
    double **prestop = softstyle->prestop;

    double delta = update->ntimestep - update->firststep;
    if (update->nsteps) delta /= update->nsteps;
    else delta = 0.0;
    int ntypes = atom->ntypes;

    for (i = 1; i <= ntypes; i++)
      for (j = 1; j <= ntypes; j++)
	prefactor[i][j] = prestart[i][j] + 
	  delta * (prestop[i][j] - prestart[i][j]);
  }

  // if EAM sub-style exists, compute pairwise embedding function

  if (eamstyle) {
    double *rho = atom->rho;
    double *fp = atom->fp;

    if (newton_pair) {
      int m = nlocal + atom->nghost;
      for (i = 0; i < m; i++) rho[i] = 0.0;
    } else for (i = 0; i < nlocal; i++) rho[i] = 0.0;
    
    for (i = 0; i < nlocal; i++) {
      xtmp = x[i][0];
      ytmp = x[i][1];
      ztmp = x[i][2];
      itype = type[i];
      neighs = neighbor->firstneigh[i];
      numneigh = neighbor->numneigh[i];
      
      for (k = 0; k < numneigh; k++) {
	j = neighs[k];
	if (j >= nall) j = j % nall;
	
	delx = xtmp - x[j][0];
	dely = ytmp - x[j][1];
	delz = ztmp - x[j][2];
	rsq = delx*delx + dely*dely + delz*delz;
	jtype = type[j];

	if (rsq < cutsq[itype][jtype] && map[itype][jtype] == eammap) {
	  eamstyle->single_rho(itype,jtype,rsq,rhoi,rhoj);
	  rho[i] += rhoi;
	  if (newton_pair || j < nlocal) rho[j] += rhoj;
	}
      }
    }

    if (newton_pair) comm->reverse_comm_pair(eamstyle);

    for (i = 0; i < nlocal; i++) {
      eamstyle->single_embed(i,itype,fp[i],eflag,phi);
      if (eflag) eng_vdwl += phi;
    }

    comm->comm_pair(eamstyle);
  }

  // loop over neighbors of my atoms
  // call to low-level pair style's fsingle,esingle

  for (i = 0; i < nlocal; i++) {
    xtmp = x[i][0];
    ytmp = x[i][1];
    ztmp = x[i][2];
    itype = type[i];
    neighs = neighbor->firstneigh[i];
    numneigh = neighbor->numneigh[i];

    for (k = 0; k < numneigh; k++) {
      j = neighs[k];

      if (j < nall) factor_coul = factor_lj = 1.0;
      else {
	factor_coul = special_coul[j/nall];
	factor_lj = special_lj[j/nall];
	j = j % nall;
      }

      delx = xtmp - x[j][0];
      dely = ytmp - x[j][1];
      delz = ztmp - x[j][2];
      rsq = delx*delx + dely*dely + delz*delz;
      jtype = type[j];

      if (rsq < cutsq[itype][jtype]) {

	styles[map[itype][jtype]]->
	  single(i,j,itype,jtype,rsq,factor_coul,factor_lj,eflag,one);

	if (!dipoleflag) {
	  fforce = one.fforce;
	  f[i][0] += delx*fforce;
	  f[i][1] += dely*fforce;
	  f[i][2] += delz*fforce;
	  if (newton_pair || j < nlocal) {
	    f[j][0] -= delx*fforce;
	    f[j][1] -= dely*fforce;
	    f[j][2] -= delz*fforce;
	  }
	} else {
	  f[i][0] += one.fx;
	  f[i][1] += one.fy;
	  f[i][2] += one.fz;
	  torque[i][0] += one.tix;
	  torque[i][1] += one.tiy;
	  torque[i][2] += one.tiz;
	  if (newton_pair || j < nlocal) {
	    f[j][0] -= one.fx;
	    f[j][1] -= one.fy;
	    f[j][2] -= one.fz;
	    torque[j][0] += one.tjx;
	    torque[j][1] += one.tjy;
	    torque[j][2] += one.tjz;
	  }
	}

	if (eflag) {
	  if (newton_pair || j < nlocal) factor = 1.0;
	  else factor = 0.5;
	  eng_coul += factor*factor_coul * one.eng_coul;
	  eng_vdwl += factor*factor_lj * one.eng_vdwl;
	}

	if (vflag == 1) {
	  if (newton_pair || j < nlocal) {
	    virial[0] += delx*delx*fforce;
	    virial[1] += dely*dely*fforce;
	    virial[2] += delz*delz*fforce;
	    virial[3] += delx*dely*fforce;
	    virial[4] += delx*delz*fforce;
	    virial[5] += dely*delz*fforce;
	  } else {
	    virial[0] += 0.5*delx*delx*fforce;
	    virial[1] += 0.5*dely*dely*fforce;
	    virial[2] += 0.5*delz*delz*fforce;
	    virial[3] += 0.5*delx*dely*fforce;
	    virial[4] += 0.5*delx*delz*fforce;
	    virial[5] += 0.5*dely*delz*fforce;
	  }
	}
      }
    }
  }
  if (vflag == 2) virial_compute();
}

/* ----------------------------------------------------------------------
   allocate all arrays 
------------------------------------------------------------------------- */

void PairHybrid::allocate()
{
  allocated = 1;
  int n = atom->ntypes;

  map = memory->create_2d_int_array(n+1,n+1,"pair:map");

  setflag = memory->create_2d_int_array(n+1,n+1,"pair:setflag");
  for (int i = 1; i <= n; i++)
    for (int j = i; j <= n; j++)
      setflag[i][j] = 0;

  cutsq = memory->create_2d_double_array(n+1,n+1,"pair:cutsq");
}

/* ----------------------------------------------------------------------
   create one pair style for each arg in list
------------------------------------------------------------------------- */

void PairHybrid::settings(int narg, char **arg)
{
  int i,j,istyle;

  if (narg < 1) error->all("Incorrect args in pair_style command");

  // delete old lists, since cannot just change settings

  if (nstyles) {
    for (int i = 0; i < nstyles; i++) delete styles[i];
    delete [] styles;
    for (int i = 0; i < nstyles; i++) delete [] keywords[i];
    delete [] keywords;
  }

  if (allocated) {
    memory->destroy_2d_int_array(setflag);
    memory->destroy_2d_int_array(map);
    memory->destroy_2d_double_array(cutsq);
  }
  allocated = 0;

  // count sub-styles by skipping numeric args
  // one exception is 1st arg of style "table", which is non-numeric word

  nstyles = i = 0;
  while (i < narg) {
    if (strcmp(arg[i],"table") == 0) i++;
    i++;
    while (i < narg && !isalpha(arg[i][0])) i++;
    nstyles++;
  }

  // allocate list of sub-styles

  styles = new Pair*[nstyles];
  keywords = new char*[nstyles];

  // allocate each sub-style and call its settings() with subset of args
  // define subset of sub-styles by skipping numeric args
  // one exception is 1st arg of style "table", which is non-numeric word

  nstyles = i = 0;
  while (i < narg) {
    for (j = 0; j < nstyles; j++)
      if (strcmp(arg[i],keywords[j]) == 0) 
	error->all("Pair style hybrid cannot use same pair style twice");
    if (strcmp(arg[i],"hybrid") == 0) 
      error->all("Pair style hybrid cannot have hybrid as an argument");
    styles[nstyles] = force->new_pair(arg[i]);
    keywords[nstyles] = new char[strlen(arg[i])+1];
    strcpy(keywords[nstyles],arg[i]);
    istyle = i;
    if (strcmp(arg[i],"table") == 0) i++;
    i++;
    while (i < narg && !isalpha(arg[i][0])) i++;
    styles[nstyles]->settings(i-istyle-1,&arg[istyle+1]);
    nstyles++;
  }

  // no more than one "long" sub-style can be used
  // no "granular" sub-style can be used

  int nlong = 0;
  int ngran = 0;
  if (force->pair_match("buck/coul/long")) nlong++;
  if (force->pair_match("lj/cut/coul/long")) nlong++;
  if (force->pair_match("lj/charmm/coul/long")) nlong++;
  if (force->pair_match("dipole/long")) nlong++;
  if (force->pair_match("gran/history")) ngran = 1;
  if (force->pair_match("gran/no_history")) ngran = 1;
  if (force->pair_match("gran/hertzian")) ngran = 1;

  if (nlong > 1) 
    error->all("Cannot use multiple long-range potentials with pair hybrid");
  if (ngran) error->all("Cannot use granular potential with pair hybrid");
}

/* ----------------------------------------------------------------------
   set i,j coeffs for one type pair
------------------------------------------------------------------------- */

void PairHybrid::coeff(int narg, char **arg)
{
  if (narg < 3) error->all("Incorrect args for pair coefficients");
  if (!allocated) allocate();

  int ilo,ihi,jlo,jhi;
  force->bounds(arg[0],atom->ntypes,ilo,ihi);
  force->bounds(arg[1],atom->ntypes,jlo,jhi);

  // 3rd arg = pair style name (lj/cut, buck, etc)

  int k;
  for (k = 0; k < nstyles; k++)
    if (strcmp(arg[2],keywords[k]) == 0) break;
  if (k == nstyles) error->all("Pair coeff for hybrid has invalid style");

  // call low-level coeff() for each atom type pair
  // replace 2nd,3rd arg with i,j
  // call set_coeff with no 1st arg

  int count = 0;
  for (int i = ilo; i <= ihi; i++) {
    for (int j = jlo; j <= jhi; j++) {
      sprintf(arg[1],"%d",i);
      sprintf(arg[2],"%d",j);
      map[i][j] = k;
      styles[k]->coeff(narg-1,&arg[1]);
      setflag[i][j] = 1;
      count++;
    }
  }

  if (count == 0) error->all("Incorrect args for pair coefficients");
}

/* ----------------------------------------------------------------------
   init for one type pair i,j and corresponding j,i
------------------------------------------------------------------------- */

double PairHybrid::init_one(int i, int j)
{
  // if i,j is set explicity, call its sub-style
  // if i,j is not set and i,i sub-style = j,j sub-style
  //   set map[i][j] to this sub-style and let sub-style do coeff mixing

  double cut;
  if (setflag[i][j]) cut = styles[map[i][j]]->init_one(i,j);
  else if (map[i][i] == map[j][j]) {
    map[i][j] = map[i][i];
    cut = styles[map[i][i]]->init_one(i,j);
  } else error->one("All pair coeffs are not set");

  cutsq[i][j] = cut*cut;
  cutsq[j][i] = cutsq[i][j];
  map[j][i] = map[i][j];
  return cut;
}

/* ----------------------------------------------------------------------
   init specific to this pair style
------------------------------------------------------------------------- */

void PairHybrid::init_style()
{
  softstyle = (PairSoft *) force->pair_match("soft");

  eamstyle = (PairEAM *) force->pair_match("eam");
  if (eamstyle)
    for (int i = 0; i < nstyles; i++)
      if (eamstyle == styles[i]) eammap = i;

  dipoleflag = 0;
  if (force->pair_match("dipole/cut")) dipoleflag = 1;
  if (force->pair_match("dipole/long")) dipoleflag = 1;

  for (int i = 0; i < nstyles; i++) styles[i]->init_style();
}

/* ----------------------------------------------------------------------
   proc 0 writes to restart file
------------------------------------------------------------------------- */

void PairHybrid::write_restart(FILE *fp)
{
  fwrite(&nstyles,sizeof(int),1,fp);

  // each sub-style writes its settings

  int n;
  for (int i = 0; i < nstyles; i++) {
    n = strlen(keywords[i]) + 1;
    fwrite(&n,sizeof(int),1,fp);
    fwrite(keywords[i],sizeof(char),n,fp);
    styles[i]->write_restart_settings(fp);
  }
}

/* ----------------------------------------------------------------------
   proc 0 reads from restart file, bcasts
------------------------------------------------------------------------- */

void PairHybrid::read_restart(FILE *fp)
{
  allocate();

  int me = comm->me;
  if (me == 0) fread(&nstyles,sizeof(int),1,fp);
  MPI_Bcast(&nstyles,1,MPI_INT,0,world);

  styles = new Pair*[nstyles];
  keywords = new char*[nstyles];
  
  // each sub-style is created via new_pair() and reads its settings

  int n;
  for (int i = 0; i < nstyles; i++) {
    if (me == 0) fread(&n,sizeof(int),1,fp);
    MPI_Bcast(&n,1,MPI_INT,0,world);
    keywords[i] = new char[n];
    if (me == 0) fread(keywords[i],sizeof(char),n,fp);
    MPI_Bcast(keywords[i],n,MPI_CHAR,0,world);
    styles[i] = force->new_pair(keywords[i]);
    styles[i]->read_restart_settings(fp);
  }
}

/* ---------------------------------------------------------------------- */

void PairHybrid::single(int i, int j, int itype, int jtype,
			double rsq, double factor_coul, double factor_lj,
			int eflag, One &one)
{
  styles[map[itype][jtype]]->
    single(i,j,itype,jtype,rsq,factor_coul,factor_lj,eflag,one);
}
