/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "mpi.h"
#include "string.h"
#include "pressure.h"
#include "output.h"
#include "thermo.h"
#include "temperature.h"
#include "force.h"
#include "modify.h"
#include "fix.h"
#include "fix_shake.h"
#include "fix_rigid.h"
#include "bond.h"
#include "angle.h"
#include "dihedral.h"
#include "improper.h"
#include "pair.h"
#include "kspace.h"
#include "atom.h"
#include "domain.h"

/* ---------------------------------------------------------------------- */

void Pressure::init()
{
  boltz = force->boltz;

  // account for SHAKE and rigid virials

  shakeflag = 0;
  rigidflag = 0;
  for (int i = 0; i < modify->nfix; i++) {
    if (strcmp(modify->fix[i]->style,"shake") == 0) {
      shakeflag = 1;
      shake_virial = ((FixShake *) modify->fix[i])->virial;
    }
    if (strcmp(modify->fix[i]->style,"rigid") == 0) {
      rigidflag = 1;
      rigid_virial = ((FixRigid *) modify->fix[i])->virial;
    }
  }

  dof = output->thermo->temperature->dof;
  nktv2p = force->nktv2p;
}

/* ---------------------------------------------------------------------- */

void Pressure::compute(double temperature)
{
  double v[3];

  // sum force contributions to virial

  v[2] = v[1] = v[0] = 0.0;

  if (force->pair) {
    v[0] += force->pair->virial[0];
    v[1] += force->pair->virial[1];
    v[2] += force->pair->virial[2];
  }

  if (atom->molecular) {
    if (force->bond) {
      v[0] += force->bond->virial[0];
      v[1] += force->bond->virial[1];
      v[2] += force->bond->virial[2];
    }
    if (force->angle) {
      v[0] += force->angle->virial[0];
      v[1] += force->angle->virial[1];
      v[2] += force->angle->virial[2];
    }
    if (force->dihedral) {
      v[0] += force->dihedral->virial[0];
      v[1] += force->dihedral->virial[1];
      v[2] += force->dihedral->virial[2];
    }
    if (force->improper) {
      v[0] += force->improper->virial[0];
      v[1] += force->improper->virial[1];
      v[2] += force->improper->virial[2];
    }
    if (shakeflag) {
      v[0] += shake_virial[0];
      v[1] += shake_virial[1];
      v[2] += shake_virial[2];
    }
  }

  if (rigidflag) {
    v[0] += rigid_virial[0];
    v[1] += rigid_virial[1];
    v[2] += rigid_virial[2];
  }

  // sum virial across procs

  MPI_Allreduce(v,virial,3,MPI_DOUBLE,MPI_SUM,world);

  // KSpace virial contribution is already summed across procs

  if (force->kspace) {
    virial[0] += force->kspace->virial[0];
    virial[1] += force->kspace->virial[1];
    virial[2] += force->kspace->virial[2];
  }

  double kinetic = dof * boltz * temperature / 3.0;
  double volume = domain->xprd * domain->yprd * domain->zprd;
  p_component[0] = (kinetic + virial[0]) / volume * nktv2p;
  p_component[1] = (kinetic + virial[1]) / volume * nktv2p;
  p_component[2] = (kinetic + virial[2]) / volume * nktv2p;
  p_total = (p_component[0] + p_component[1] + p_component[2]) / 3.0;
}
