/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "mpi.h"
#include "math.h"
#include "string.h"
#include "set.h"
#include "atom.h"
#include "domain.h"
#include "group.h"
#include "comm.h"
#include "neighbor.h"
#include "force.h"
#include "pair.h"
#include "random_park.h"
#include "error.h"

#define ATOM     1
#define BOND     2
#define ANGLE    3
#define DIHEDRAL 4
#define IMPROPER 5
#define CHARGE   6
#define DIPOLE   7

#define WARMUP 100

/* ---------------------------------------------------------------------- */

void Set::command(int narg, char **arg)
{
  if (narg != 3) error->all("Illegal set command");

  if (domain->box_exist == 0)
    error->all("Set command before simulation box is defined");
  if (atom->natoms == 0)
    error->all("Set command with no atoms existing");

  if (comm->me == 0 && screen) fprintf(screen,"Setting atom values ...\n");

  // identify group

  int igroup = group->find(arg[0]);
  if (igroup == -1) error->all("Cannot find set command group ID");
  int groupbit = group->bitmask[igroup];
  
  // set style and ivalue

  int style;
  if (strcmp(arg[1],"atom") == 0) style = ATOM;
  else if (strcmp(arg[1],"bond") == 0) style = BOND;
  else if (strcmp(arg[1],"angle") == 0) style = ANGLE;
  else if (strcmp(arg[1],"dihedral") == 0) style = DIHEDRAL;
  else if (strcmp(arg[1],"improper") == 0) style = IMPROPER;
  else if (strcmp(arg[1],"charge") == 0) style = CHARGE;
  else if (strcmp(arg[1],"dipole") == 0) style = DIPOLE;
  else error->all("Illegal set command");

  int ivalue;
  double value;
  if (style == CHARGE) value = atof(arg[2]);
  else ivalue = atoi(arg[2]);

  // consistency checks

  if ((style == BOND && atom->bonds_allow == 0) ||
      (style == ANGLE && atom->angles_allow == 0) ||
      (style == DIHEDRAL && atom->dihedrals_allow == 0) ||
      (style == IMPROPER && atom->impropers_allow == 0) ||
      (style == CHARGE && atom->charge_allow == 0) ||
      (style == DIPOLE && atom->dipole_require == 0))
    error->all("Cannot set these values with this atom style");

  if (style == ATOM && (ivalue < 1 || ivalue > atom->ntypes))
    error->all("Illegal type in set command");
  if (style == BOND && (ivalue < 1 || ivalue > atom->nbondtypes))
    error->all("Illegal type in set command");
  if (style == ANGLE && (ivalue < 1 || ivalue > atom->nangletypes))
    error->all("Illegal type in set command");
  if (style == DIHEDRAL && (ivalue < 1 || ivalue > atom->ndihedraltypes))
    error->all("Illegal type in set command");
  if (style == IMPROPER && (ivalue < 1 || ivalue > atom->nimpropertypes))
    error->all("Illegal type in set command");
  if (style == DIPOLE && ivalue <= 0)
    error->all("Illegal random number seed in set command");

  // border swap to insure type and mask is current for off-proc atoms
  // only needed for BOND, ANGLE, etc types
  // must first set neighbor->cutneigh so comm will work
  // this requires pair->cutforce be set, do it via pair->init
  
  if (style == BOND || style == ANGLE || 
      style == DIHEDRAL || style == IMPROPER) {
    if (force->pair == NULL) neighbor->cutneigh = neighbor->skin;
    else {
      force->pair->init();
      neighbor->cutneigh = force->pair->cutforce + neighbor->skin;
    }
    comm->init();
    comm->setup();
    comm->exchange();
    comm->borders();
  }
  
  // set new values

  int *mask = atom->mask;
  int nlocal = atom->nlocal;
  int m,atom1,atom2,atom3,atom4;

  int count = 0;

  // for style ATOM, atom must be in group

  if (style == ATOM)
    for (int i = 0; i < nlocal; i++)
      if (mask[i] & groupbit) {
	atom->type[i] = ivalue;
	count++;
      }

  // for style BOND, each of 2 atoms must be in group

  if (style == BOND)
    for (int i = 0; i < nlocal; i++)
      for (m = 0; m < atom->num_bond[i]; m++) {
	atom1 = atom->map(atom->bond_atom[i][m]);
	if (atom1 == -1) error->one("Bond atom missing in set command");
	if (mask[i] & groupbit && mask[atom1] & groupbit) {
	  atom->bond_type[i][m] = ivalue;
	  count++;
	}
      }

  // for style ANGLE, each of 3 atoms must be in group

  if (style == ANGLE)
    for (int i = 0; i < nlocal; i++)
      for (m = 0; m < atom->num_angle[i]; m++) {
	atom1 = atom->map(atom->angle_atom1[i][m]);
	atom2 = atom->map(atom->angle_atom2[i][m]);
	atom3 = atom->map(atom->angle_atom3[i][m]);
	if (atom1 == -1 || atom2 == -1 || atom3 == -1)
	  error->one("Angle atom missing in set command");
	if (mask[atom1] & groupbit && mask[atom2] & groupbit &&
	    mask[atom3] & groupbit) {
	  atom->angle_type[i][m] = ivalue;
	  count++;
	}
      }

  // for style DIHEDRAL, each of 4 atoms must be in group

  if (style == DIHEDRAL)
    for (int i = 0; i < nlocal; i++)
      for (m = 0; m < atom->num_dihedral[i]; m++) {
	atom1 = atom->map(atom->dihedral_atom1[i][m]);
	atom2 = atom->map(atom->dihedral_atom2[i][m]);
	atom3 = atom->map(atom->dihedral_atom3[i][m]);
	atom4 = atom->map(atom->dihedral_atom4[i][m]);
	if (atom1 == -1 || atom2 == -1 || atom3 == -1 || atom4 == -1)
	  error->one("Dihedral atom missing in set command");
	if (mask[atom1] & groupbit && mask[atom2] & groupbit &&
	    mask[atom3] & groupbit && mask[atom4] & groupbit) {
	  atom->dihedral_type[i][m] = ivalue;
	  count++;
	}
      }

  // for style IMPROPER, each of 4 atoms must be in group

  if (style == IMPROPER)
    for (int i = 0; i < nlocal; i++)
      for (m = 0; m < atom->num_improper[i]; m++) {
	atom1 = atom->map(atom->improper_atom1[i][m]);
	atom2 = atom->map(atom->improper_atom2[i][m]);
	atom3 = atom->map(atom->improper_atom3[i][m]);
	atom4 = atom->map(atom->improper_atom4[i][m]);
	if (atom1 == -1 || atom2 == -1 || atom3 == -1 || atom4 == -1)
	  error->one("Improper atom missing in set command");
	if (mask[atom1] & groupbit && mask[atom2] & groupbit &&
	    mask[atom3] & groupbit && mask[atom4] & groupbit) {
	  atom->improper_type[i][m] = ivalue;
	  count++;
	}
      }

  // for style CHARGE, set atom charge

  if (style == CHARGE) {
    double *q = atom->q;
    for (int i = 0; i < nlocal; i++)
      if (mask[i] & groupbit) {
	q[i] = value;
	count++;
      }
  }

  // for style DIPOLE, create unique random number stream for each proc
  // set dipole moment of each atom to random orientation
  // dipole length is determined by dipole type array

  if (style == DIPOLE) {
    double msq,scale;
    RanPark *random = new RanPark(ivalue + comm->me);
    for (int i = 0; i < WARMUP; i++) random->uniform();
    int *type = atom->type;
    double *dipole = atom->dipole;
    double **mu = atom->mu;
    for (int i = 0; i < nlocal; i++)
      if (mask[i] & groupbit) {
	mu[i][0] = random->uniform() - 0.5;
	mu[i][1] = random->uniform() - 0.5;
	mu[i][2] = random->uniform() - 0.5;
	msq = mu[i][0]*mu[i][0] + mu[i][1]*mu[i][1] + mu[i][2]*mu[i][2];
	scale = dipole[type[i]]/sqrt(msq);
	mu[i][0] *= scale;
	mu[i][1] *= scale;
	mu[i][2] *= scale;
	count++;
      }
    delete random;
  }

  // statistics

  int allcount;
  MPI_Allreduce(&count,&allcount,1,MPI_INT,MPI_SUM,world);
  
  if (comm->me == 0) {
    if (screen) fprintf(screen,"  %d settings made\n",allcount);
    if (logfile) fprintf(logfile,"  %d settings made\n",allcount);
  }
}
