/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "mpi.h"
#include "string.h"
#include "thermo.h"
#include "force.h"
#include "temperature.h"
#include "update.h"
#include "comm.h"
#include "atom.h"
#include "domain.h"
#include "error.h"

/* ---------------------------------------------------------------------- */

Thermo::Thermo(char *style_in)
{
  int n = strlen(style_in) + 1;
  style = new char[n];
  strcpy(style,style_in);

  normuser = 0;
  lostflag = 2;
  lostbefore = 0;
  flush_flag = 0;

  temperature = force->templist[0];
}

/* ---------------------------------------------------------------------- */

Thermo::~Thermo()
{
  delete [] style;
}

/* ---------------------------------------------------------------------- */

void Thermo::init()
{
  // set normflag to default unless user has specified

  if (normuser == 0) {
    if (strcmp(update->unit_style,"lj") == 0) normflag = 1;
    else normflag = 0;
  } else {
    if (normuser == 1) normflag = 0;
    else if (normuser == 2) normflag = 1;
  }

  // this must come after domain init, so box_change is set

  if (domain->box_change == 0) volflag = 0;
  else volflag = 1;

  // init for a particular style

  init_style();
}

/* ----------------------------------------------------------------------
   check for lost atoms, return current number of atoms
------------------------------------------------------------------------- */

double Thermo::lost_check()
{
  // natoms = current # of atoms

  double natoms;
  double rlocal = atom->nlocal;
  MPI_Allreduce(&rlocal,&natoms,1,MPI_DOUBLE,MPI_SUM,world);
  if (natoms == atom->natoms) return natoms;

  // if not checking or already warned, just return

  if (lostflag == 0) return natoms;
  if (lostflag == 1 && lostbefore == 1) return natoms;

  // error message

  if (lostflag == 2) {
    char str[128];
    sprintf(str,"Lost atoms: original %.15g current %.15g",
	    atom->natoms,natoms);
    error->all(str);
  }

  // warning message

  char str[128];
  sprintf(str,"Lost atoms: original %.15g current %.15g",atom->natoms,natoms);
  int me;
  MPI_Comm_rank(world,&me);
  if (me == 0) error->warning(str);
  lostbefore = 1;
  return natoms;
}

/* ----------------------------------------------------------------------
   modify thermo parameters
------------------------------------------------------------------------- */

void Thermo::modify_params(int narg, char **arg)
{
  if (narg == 0) error->all("Illegal thermo_modify command");

  int iarg = 0;
  while (iarg < narg) {
    if (strcmp(arg[iarg],"temp") == 0) {
      if (iarg+2 > narg) error->all("Illegal thermo_modify command");
      int itemp;
      for (itemp = 0; itemp < force->ntemp; itemp++)
	if (strcmp(arg[iarg+1],force->templist[itemp]->id) == 0) break;
      if (itemp == force->ntemp) 
	error->all("Could not find thermo_modify temp ID");
      temperature = force->templist[itemp];
      iarg += 2;
    } else if (strcmp(arg[iarg],"lost") == 0) {
      if (iarg+2 > narg) error->all("Illegal thermo_modify command");
      if (strcmp(arg[iarg+1],"ignore") == 0) lostflag = 0;
      else if (strcmp(arg[iarg+1],"warn") == 0) lostflag = 1;
      else if (strcmp(arg[iarg+1],"error") == 0) lostflag = 2;
      else error->all("Illegal thermo_modify command");
      iarg += 2;
    } else if (strcmp(arg[iarg],"norm") == 0) {
      if (iarg+2 > narg) error->all("Illegal thermo_modify command");
      if (strcmp(arg[iarg+1],"no") == 0) normuser = 1;
      else if (strcmp(arg[iarg+1],"yes") == 0) normuser = 2;
      else error->all("Illegal thermo_modify command");
      iarg += 2;
    } else if (strcmp(arg[iarg],"flush") == 0) {
      if (iarg+2 > narg) error->all("Illegal thermo_modify command");
      if (strcmp(arg[iarg+1],"no") == 0) flush_flag = 1;
      else if (strcmp(arg[iarg+1],"yes") == 0) flush_flag = 2;
      else error->all("Illegal thermo_modify command");
      iarg += 2;
    } else error->all("Incorrect args in thermo_modify command");
  }
}
