/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "string.h"
#include "stdio.h"
#include "thermo_multi.h"
#include "update.h"
#include "atom.h"
#include "force.h"
#include "pair.h"
#include "bond.h"
#include "angle.h"
#include "dihedral.h"
#include "improper.h"
#include "kspace.h"
#include "temperature.h"
#include "pressure.h"
#include "domain.h"
#include "timer.h"
#include "error.h"

/* ---------------------------------------------------------------------- */

ThermoMulti::ThermoMulti(char *style_in) : Thermo(style_in)
{
  int n;
  char *str;

  MPI_Comm_rank(world,&me);

  str = 
    "--------------- Step %8d ---- CPU = %11.4f (sec) ---------------\n"
    "Total E = %14.4f Tot KE  = %14.4f Temp    = %14.4f\n"
    "Tot PE  = %14.4f E_bond  = %14.4f E_angle = %14.4f\n"
    "E_dihed = %14.4f E_impr  = %14.4f E_vdwl  = %14.4f\n"
    "E_coul  = %14.4f E_long  = %14.4f Press   = %14.4f\n";
  n = strlen(str) + 1;
  format = new char[n];
  strcpy(format,str);

  str = 
    "--------------- Step %8d ---- CPU = %11.4f (sec) ---------------\n"
    "Total E = %14.4f Tot KE  = %14.4f Temp    = %14.4f\n"
    "Tot PE  = %14.4f E_bond  = %14.4f E_angle = %14.4f\n"
    "E_dihed = %14.4f E_impr  = %14.4f E_vdwl  = %14.4f\n"
    "E_coul  = %14.4f E_long  = %14.4f Press   = %14.4f\n"
    "Volume  = %14.4f\n";
  n = strlen(str) + 1;
  format_volume = new char[n];
  strcpy(format_volume,str);

  pressure = force->pressure;
}

/* ---------------------------------------------------------------------- */

ThermoMulti::~ThermoMulti()
{
  delete [] format;
  delete [] format_volume;
}

/* ----------------------------------------------------------------------
   compute thermodynamics
------------------------------------------------------------------------- */

void ThermoMulti::compute(int flag)
{
  double natoms = lost_check();

  // temperature & pressure & volume

  double temp = temperature->compute();
  pressure->compute(temp);
  double press = pressure->p_total;
  double volume;
  if (volflag) volume = domain->xprd * domain->yprd * domain->zprd;

  // energies
  // sum across procs, except e_kspace is already summed
  // include contributions to Vdwl and Coulombic from topology interactions

  double tmp = 0.0;

  double eng_vdwl;
  if (force->pair) tmp += force->pair->eng_vdwl;
  if (force->bond) tmp += force->bond->eng_vdwl;
  if (force->dihedral) tmp += force->dihedral->eng_vdwl;
  MPI_Allreduce(&tmp,&eng_vdwl,1,MPI_DOUBLE,MPI_SUM,world);

  tmp = 0.0;
  double eng_coul;
  if (force->pair) tmp += force->pair->eng_coul;
  if (force->dihedral) tmp += force->dihedral->eng_coul;
  MPI_Allreduce(&tmp,&eng_coul,1,MPI_DOUBLE,MPI_SUM,world);

  double eng_bond = 0.0;
  if (force->bond)
    MPI_Allreduce(&force->bond->energy,&eng_bond,1,MPI_DOUBLE,MPI_SUM,world);

  double eng_angle = 0.0;
  if (force->angle)
    MPI_Allreduce(&force->angle->energy,&eng_angle,1,MPI_DOUBLE,MPI_SUM,world);

  double eng_dihedral = 0.0;
  if (force->dihedral)
    MPI_Allreduce(&force->dihedral->energy,&eng_dihedral,1,
		  MPI_DOUBLE,MPI_SUM,world);

  double eng_improper = 0.0;
  if (force->improper)
    MPI_Allreduce(&force->improper->energy,&eng_improper,1,
		  MPI_DOUBLE,MPI_SUM,world);

  double eng_kspace = 0.0;
  if (force->kspace) eng_kspace = force->kspace->energy;

  // total energy
  // set eng_pe for parent class
  
  eng_pe = eng_vdwl + eng_coul + eng_bond + eng_angle +
    eng_dihedral + eng_improper + eng_kspace;
  double eng_ke = 0.5 * temperature->dof * force->boltz * temp;
  double eng_total = eng_pe + eng_ke;

  // normalization

  if (normflag) {
    double n = natoms;
    if (n == 0) n = 1;
    eng_vdwl /= n;
    eng_coul /= n;
    eng_bond /= n;
    eng_angle /= n;
    eng_dihedral /= n;
    eng_improper /= n;
    eng_kspace /= n;
    eng_ke /= n;
    eng_pe /= n;
    eng_total /= n;
  }

  // output

  double cpu;
  if (flag) cpu = timer->elapsed(TIME_LOOP);
  else cpu = 0.0;

  if (me == 0) {
    if (volflag == 0) {
      if (screen)
	fprintf(screen,format,
		update->ntimestep,cpu,
		eng_total,eng_ke,temp,eng_pe,eng_bond,eng_angle,
		eng_dihedral,eng_improper,eng_vdwl,eng_coul,eng_kspace,press);
      if (logfile) {
	fprintf(logfile,format,
		update->ntimestep,cpu,
		eng_total,eng_ke,temp,eng_pe,eng_bond,eng_angle,
		eng_dihedral,eng_improper,eng_vdwl,eng_coul,eng_kspace,press);
	if (flush_flag) fflush(logfile);
      }
    } else {
      if (screen)
	fprintf(screen,format_volume,
		update->ntimestep,cpu,
		eng_total,eng_ke,temp,eng_pe,eng_bond,eng_angle,
		eng_dihedral,eng_improper,eng_vdwl,eng_coul,eng_kspace,press,
		volume);
      if (logfile) {
	fprintf(logfile,format_volume,
		update->ntimestep,cpu,
		eng_total,eng_ke,temp,eng_pe,eng_bond,eng_angle,
		eng_dihedral,eng_improper,eng_vdwl,eng_coul,eng_kspace,press,
		volume);
	if (flush_flag) fflush(logfile);
      }
    }
  }
}
