/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "mpi.h"
#include "string.h"
#include "stdio.h"
#include "thermo_one.h"
#include "update.h"
#include "atom.h"
#include "force.h"
#include "pair.h"
#include "bond.h"
#include "angle.h"
#include "dihedral.h"
#include "improper.h"
#include "kspace.h"
#include "temperature.h"
#include "pressure.h"
#include "domain.h"
#include "error.h"

/* ---------------------------------------------------------------------- */

ThermoOne::ThermoOne(char *style_in) : Thermo(style_in)
{
  int n;
  char *str;

  MPI_Comm_rank(world,&me);

  str = "%6d %12.8g %12.8g %12.8g %12.8g %12.8g\n";
  n = strlen(str) + 1;
  format = new char[n];
  strcpy(format,str);

  str = "%6d %12.8g %12.8g %12.8g %12.8g %12.8g %12.8g\n";
  n = strlen(str) + 1;
  format_volume = new char[n];
  strcpy(format_volume,str);

  pressure = force->pressure;
}

/* ---------------------------------------------------------------------- */

ThermoOne::~ThermoOne()
{
  delete [] format;
  delete [] format_volume;
}

/* ---------------------------------------------------------------------- */

void ThermoOne::header()
{
  if (me == 0) {
    if (volflag == 0) {
      if (screen)
	fprintf(screen,
		"Step Temperature E_pair E_bond E_total Pressure\n");
      if (logfile)
	fprintf(logfile,
		"Step Temperature E_pair E_bond E_total Pressure\n");
    } else {
      if (screen)
	fprintf(screen,
		"Step Temperature E_pair E_bond E_total Pressure Volume\n");
      if (logfile)
	fprintf(logfile,
		"Step Temperature E_pair E_bond E_total Pressure Volume\n");
    }
  }
}

/* ---------------------------------------------------------------------- */

void ThermoOne::compute(int flag)
{
  double natoms = lost_check();

  // temperature & pressure & volume

  double temp = temperature->compute();
  pressure->compute(temp);
  double press = pressure->p_total;
  double volume;
  if (volflag) volume = domain->xprd * domain->yprd * domain->zprd;

  // pair energy
  // include contributions to Vdwl/Coulombic energy from topology interactions
  // include kspace energy (already summed across procs)

  double tmp = 0.0;

  double eng_pair;
  if (force->pair) tmp += force->pair->eng_vdwl + force->pair->eng_coul;
  if (force->bond) tmp += force->bond->eng_vdwl;
  if (force->dihedral) 
    tmp += force->dihedral->eng_vdwl + force->dihedral->eng_coul;
  MPI_Allreduce(&tmp,&eng_pair,1,MPI_DOUBLE,MPI_SUM,world);
  if (force->kspace) eng_pair += force->kspace->energy;

  // bond energy from all topology interactions

  double eng_bond = 0.0;

  if (atom->molecular) {
    tmp = 0.0;
    if (force->bond) tmp += force->bond->energy;
    if (force->angle) tmp += force->angle->energy;
    if (force->dihedral) tmp += force->dihedral->energy;
    if (force->improper) tmp += force->improper->energy;
    MPI_Allreduce(&tmp,&eng_bond,1,MPI_DOUBLE,MPI_SUM,world);
  }

  // total energy
  // set eng_pe for parent class
  
  eng_pe = eng_pair + eng_bond;
  double eng_ke = 0.5 * temperature->dof * force->boltz * temp;
  double eng_total = eng_pe + eng_ke;

  // normalization

  if (normflag) {
    double n = natoms;
    if (n == 0) n = 1;
    eng_pair /= n;
    eng_bond /= n;
    eng_pe /= n;
    eng_total /= n;
  }

  // output

  if (me == 0) {
    if (volflag == 0) {
      if (screen)
	fprintf(screen,format,update->ntimestep,temp,eng_pair,eng_bond,
		eng_total,press);
      if (logfile) {
	fprintf(logfile,format,update->ntimestep,temp,eng_pair,eng_bond,
		eng_total,press);
	if (flush_flag) fflush(logfile);
      }
    } else {
      if (screen)
	fprintf(screen,format_volume,update->ntimestep,temp,eng_pair,eng_bond,
		eng_total,press,volume);
      if (logfile) {
	fprintf(logfile,format_volume,update->ntimestep,temp,eng_pair,eng_bond,
		eng_total,press,volume);
	if (flush_flag) fflush(logfile);
      }
    }
  }
}
