/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "string.h"
#include "variable.h"
#include "universe.h"
#include "memory.h"
#include "error.h"

#define MAXVAR 26
#define PROC    0
#define INDEX   1

/* ---------------------------------------------------------------------- */

Variable::Variable()
{
  num = new int[MAXVAR];
  type = new int[MAXVAR];
  index = new int[MAXVAR];

  data = (char ***) memory->smalloc(MAXVAR*sizeof(char **),"variable:data");

  for (int i = 0; i < MAXVAR; i++) num[i] = 0;
  for (int i = 0; i < MAXVAR; i++) data[i] = NULL;
}

/* ---------------------------------------------------------------------- */

Variable::~Variable()
{
  int i,j;

  // must use free because memory class has been deleted
  // cannot do memory->sfree()

  for (i = 0; i < MAXVAR; i++) {
    for (j = 0; j < num[i]; j++) delete [] data[i][j];
    if (num[i]) free(data[i]);
  }
  if (data) free(data);

  delete [] num;
  delete [] type;
  delete [] index;
}

/* ----------------------------------------------------------------------
   called by variable command in input script to create INDEX or PROC var
------------------------------------------------------------------------- */

void Variable::set(int narg, char **arg)
{
  if (narg < 3) error->all("Illegal variable command");

  // ivar = which variable, a to z = 0 to 25
  // if num[ivar] then variable already exists, so skip the command

  int ivar = arg[0][0] - 'a';
  if (ivar < 0 || ivar >= MAXVAR) 
    error->all("Illegal variable in variable command");
  if (num[ivar]) return;

  if (strcmp(arg[1],"index") == 0) type[ivar] = INDEX;
  else if (strcmp(arg[1],"proc") == 0) type[ivar] = PROC;
  else error->all("Illegal variable command");

  // if INDEX, set index to 1st value
  // if PROC, set index to the partition I'm in, error check for correct count

  num[ivar] = narg - 2;
  data[ivar] = (char **) 
    memory->smalloc(num[ivar]*sizeof(char *),"variable:data");
  copy(num[ivar],&arg[2],data[ivar]);

  if (type[ivar] == INDEX) index[ivar] = 0;
  if (type[ivar] == PROC) {
    if (num[ivar] != universe->nworlds)
      error->all("Proc variable count doesn't match # of partitions");
    index[ivar] = universe->iworld;
  }
}

/* ----------------------------------------------------------------------
   single-value INDEX var created by command-line argument
------------------------------------------------------------------------- */

void Variable::set(char *name, char *value)
{
  int ivar = name[0] - 'a';
  if (ivar < 0 || ivar >= MAXVAR) 
    error->all("Illegal variable in command-line argument");

  if (num[ivar]) error->all("Command-line variable already exists");
  type[ivar] = INDEX;
  num[ivar] = 1;
  data[ivar] = (char **) 
    memory->smalloc(num[ivar]*sizeof(char *),"variable:data");
  copy(num[ivar],&value,data[ivar]);
  index[ivar] = 0;
}

/* ----------------------------------------------------------------------
   increment appropriate variable(s)
   return 0 if OK, -1 if a variable is exhausted
------------------------------------------------------------------------- */

int Variable::next(int narg, char **arg)
{
  int ivar;

  if (narg == 0) error->all("Illegal next command");

  // test that args are valid

    for (int iarg = 0; iarg < narg; iarg++) {
      ivar = arg[iarg][0] - 'a';
      if (ivar < 0 || ivar >= MAXVAR) 
	error->all("Illegal variable in next command");
      if (num[ivar] == 0) error->all("Illegal variable in next command");
      if (type[ivar] == PROC) error->all("Next command has proc variable");
    }

  // one world, increment each variable in list
  // return -1 when any variable is exhausted

  if (universe->nworlds == 1) {
    for (int iarg = 0; iarg < narg; iarg++) {
      ivar = arg[iarg][0] - 'a';
      index[ivar]++;
      if (index[ivar] == num[ivar]) return -1;
    }

    // make a new 3rd type of var
    // when variable command for these vars is read, need to write master file
    // when created, need to write this file with one line for each num

    // master file format:
    // LAMMPS variable status
    // x N P     x = var name, N = which index for var, P = which proc
    //           one more flag for status (done, running) - or does that matter
    //           might matter for restart
    //           how will restart find this file?
    //           could just make restart reset the variables
    // master file contains all variable info, one line per var per value
    // move master file to proc-specific-name
    // if can open proc-specific-name, then read file
    // else someone else has it, so iterate on move
    // update the var(s) for me, exit if done
    // write out a new master file with updated info
    // delete the proc-specific-name file

    // need fns to read/write the master file
    // print message to master screen/logfile about variable updates?
    // doc that these file(s) exist, don't mess with them
    //   what if LAMMPS stops in middle?
    // can I restart from those files?
    //   would mean don't create them when hit variable command
    // when read file, find the num I currently am and set it to "done"

  } else {
    error->all("Next command for multiple partitions not yet implemented");
  }

  return 0;
}

/* ----------------------------------------------------------------------
   return the text associated with a variable
   return NULL if no variable or index is bad, caller must respond
------------------------------------------------------------------------- */

char *Variable::retrieve(char tag)
{
  int ivar = tag - 'a';
  if (num[ivar] == 0) return NULL;                // no variable exists
  if (index[ivar] >= num[ivar]) return NULL;      // variable exhausted
  return data[ivar][index[ivar]];
}

/* ----------------------------------------------------------------------
   copy narg strings from **from to **to 
------------------------------------------------------------------------- */
  
void Variable::copy(int narg, char **from, char **to)
{
  int n;

  for (int i = 0; i < narg; i++) {
    n = strlen(from[i]) + 1;
    to[i] = new char[n];
    strcpy(to[i],from[i]);
  }
}
